/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
IasOSSrv is a generic, OS independent collection of operating systm services needed by the driver.
Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#ifndef __IasOSSrv_h
#define __IasOSSrv_h

#include "ctcomdef.h"
#include "ctdef.h"

#ifdef DEBUG
HRESULT CtDebugLog( CTCHAR*  szString );
typedef HRESULT ( *PLOGFUNC )( CTCHAR*  szString );
#endif

//***********************
// Forward declare
//***********************
interface IasSystem;
interface IasOSSrv;
interface ICtDevice;

#ifdef __cplusplus
#  define STRUCT
#else
#  define STRUCT struct
#endif

typedef  HRESULT ( *PGETCLASSOBJECTFUNC ) ( REFGUID clsid, CTDWORD ClsContext, CTDWORD* reserved, REFGUID iid, ICtUnknown* p, CTVOID** ppv );
CTBEGINCTYPE
typedef void ( *PFCALLBACK )( CTHANDLE hUser );
typedef CTBOOL ( *PF_ISR_CALLBACK )( CTHANDLE hUser );
CTENDCTYPE

//********************
// CtComServ struct
//********************
typedef struct _ComServ
{
    GUID clsid;
    PGETCLASSOBJECTFUNC pFunc;
} stCtComServ;


//***********************
// Typedefs and defines
//***********************
typedef CTVOID * OSVIRTADDR;       /* Host virtual addresses.  */
typedef CTPTRSIZE OSPHYSADDR;      /* Host physical addresses. */
typedef CTPTRSIZE OSBUSADDR;       /* Host bus addresses. */
typedef CTPTRSIZE OSSAMPADDR;      /* Sample memory addresses. */
typedef CTHANDLE  OSPHYSMEMHDL;    /* opaque handle used to reference a */
                                   /* particular region of allocated host memory.*/
typedef CTHANDLE OSMUTEXHDL;       /* A Mutex handle for the mutex operations.*/
typedef CTHANDLE OSTIMEOUTHANDLE;  /* A handle for a previously scheduled timeout.*/
typedef CTHANDLE OSSPINLOCKHDL;
typedef CTHANDLE OSDPCHANDLE;
typedef CTHANDLE OSSEMAHDL;


//***************
// Guids
//***************
// {23D9A3E1-E11E-11d2-9CB4-006097D1FC31}
DEFINE_GUID( CLSID_CtOSSrv, 0x23d9a3e1, 0xe11e, 0x11d2, 0x9c, 0xb4, 0x0, 0x60, 0x97, 0xd1, 0xfc, 0x31);

// {BEF0FBC1-BAA5-11d2-9CB4-006097D1FC31}
DEFINE_GUID( IID_IasOSSrv, 0xbef0fbc1, 0xbaa5, 0x11d2, 0x9c, 0xb4, 0x0, 0x60, 0x97, 0xd1, 0xfc, 0x31);

//***********************
// File IO enumerations
//***********************
typedef enum {
    CTFILEIO_READBINARY,      // Open for read only, fail if not exist
    CTFILEIO_READASCII,
    CTFILEIO_WRITEBINARY,     // Open for write only, overwrite if exist, create if not exist
    CTFILEIO_WRITEASCII,
    CTFILEIO_APPENDBINARY,    // Open for write only, fail if not exist, append to end
    CTFILEIO_APPENDASCII,
    CTFILEIO_READWRITEBINARY, // Open for read and write, create if not exist, don't overwrite if exist
    CTFILEIO_READWRITEASCII
} CTFILEIO_MODE;

typedef enum {
    CTFILEIO_SEEKSET,
    CTFILEIO_SEEKCUR,
    CTFILEIO_SEEKEND
} CTFILEIO_SEEK;




#undef INTERFACE
#define INTERFACE IasOSSrv
DECLARE_INTERFACE_( IasOSSrv, ICtUnknown )
{
    // IUnknown methods
    IMP_ICTUNKNOWN;

    //************************************************************************
    // Atomic increment and decrement operations
    //************************************************************************
    STDMETHOD_( CTULONG, InterlockedIncrement ) ( THIS_ CTULONG*  ) PURE;
    STDMETHOD_( CTULONG, InterlockedDecrement ) ( THIS_ CTULONG*  ) PURE;

    //************************************************************************
    //                     Spinlock Functions
    //
    // Spinlocks can only be used by pagelocked code. In the case of NT,
    // acquiring a spinlock will automatically raise the IRQL to  xxxxxxx,
    // In the case on Windows 9x, interrupts will be disabled.
    //
    //************************************************************************
    STDMETHOD( CreateSpinLock  ) ( THIS_ OSSPINLOCKHDL* phSpinlock, interface ICtDevice * device ) PURE;
    STDMETHOD( DestroySpinLock ) ( THIS_ OSSPINLOCKHDL  hSpinlock) PURE;
    STDMETHOD( AcquireSpinLock ) ( THIS_ OSSPINLOCKHDL  hSpinlock) PURE;
    STDMETHOD( ReleaseSpinLock ) ( THIS_ OSSPINLOCKHDL  hSpinlock) PURE;

    //**************************************************************************
    //                      Page Allocate Functions
    //
    // AllocPages - Allocates enough OS pages to encompass the number of specified
    //  number of bytes.  The range's size is alwayd rounded up to next highest OS
    //  page if the number of bytes requested isn't an even multiple of the OS page
    //  size.  So, for example, if a user asks for 4096 bytes on a system with 8K
    //  pages, they will get back 8K.  In general, a user can assume only that they
    //  will get back at least as much memory as they request.  These pages are
    //  NOT locked (see LockPages).   If the request fails, 0 is returned.
    //  Can only be called at Passive level.
    //  The following flags are used to control the precise behavior the 
    //  AllocPages call.
    //      The CTOS_PAGEALLOC_DMA_INPUT flag indicates that the pages will be used
    //      as the target of a DMA transfer from the device into system memory.
    //
    //      Conversely, the CTOS_PAGEALLOC_DMA_OUTPUT flag indicates that the pages 
    //      will be used as the source of a DMA transfer from system memory to the 
    //      device.
    //
    // AllocContiguousPages - Allocates LOCKED contiguous pages.  As with
    //  AllocPages, this routine rounds the number of bytes requested up to
    //  the next OS page size.  The pages returned are guaranteed to be both
    //  locked and contiguous.  Also, on systems where this is necessary,
    //  the pages are mapped into the bus address space.  Passive level only.
    //
    // FreePages - Deallocates pages.  This routine should be called to free
    //  memory that was originally allocated by either AllocPages or
    //  AllocContiguousPages.   Passive level only.
    //
    // LockPages - Locks a page range that was previously allocated by
    //  AllocPages.  Locking a page range accomplishes three things: it
    //  insures that physical pages are committed for the virtual range;
    //  it marks the pages as locked in memory; and it insures that
    //  a mapping of the pages exists in the bus's address space.
    //  Calls to LockPages may be nested; if it is called multiple times,
    //  the pages won't actually be unlocked until UnlockPages is called
    //  an equal number of times.  While it is legal to call LockPages and
    //  UnlockPages on Contiguous page ranges, it has no effect since the
    //  pages are locked at allocation time.   Passive level only.
    //
    // UnlockPages - Decrements the lock count for a page range.  When the
    //  lock count drops to zero, the pages are unlocked.  Once a page range
    //  has been unlocked, calls to GetPageBusAddr and GetPagePhysAddr may
    //  return incorrect values.  Passive level only.
    //
    // GetPageVirtAddr - Returns the virtual address of the first page in
    //  the page range.  Although this return may be called at any call level
    //  (passive, dispatch, or ISR) the virtual address may not be used to
    //  reference memory at a non-passive level unless the page range has been
    //  locked.
    //
    // GetPageUserModeVirtAddr - Returns a virtual address which may be used
    //  by an application running in user mode to reference the memory defined
    //  by the memhdl.  This address should remain valid until the memhdl is
    //  destroyed or until the process which originally issued the call shuts
    //  down.  Core code developers should be wary of this call, since it is
    //  difficult to implement it in a general way for all OSes.  In general,
    //  it should only be used in optional components.
    //  On certain OSes, this call may fail if used more than once, especially
    //  if the subsequent call was made from a process different from the original
    //
    // GetPagePhysAddr - Given a byte offset into the page range, this routine
    //  returns the physical address of byte.  For example, let's suppose that
    //   a 16K page range has been allocated, and that the value returned by
    //   GetPageVirtAddr for this range is 0x40000.  If the client calls
    //   GetPagePhysAddr with an offset of 0x6000, this routine should return
    //   the physical address of the byte at virtual address 0x46000.
    //   For our purposes, a physical address is defined as the address which
    //   is stored in the processor's  PTE entry as part of virtual to physical
    //   translation.  This should not be called unless the page range has been
    //   locked with LockPages.  It may be called at any call level.
    //
    // GetPageBusAddr - Given a byte offset into the page range, this routine
    //  returns the bus address of the byte at that offset.  For our purposes,
    //  a bus address is defined as the address which the sound chip must
    //  issue in order to read or write a particular byte of memory.  On most
    //  Intel systems, GetPagePhysAddr and GetPageBusAddr return the same
    //  value.  On systems which support physical addresses greater than 4 Gigs,
    //  however, it is necessary to map portions of the physical address space
    //  into the bus address space (the Alpha and most SGIs require this,
    //  for example).   This routine will return invalid addresses unless the
    //  page range has been locked.  It may be called at any call level.
    //
    // GetPageRangeSize - Returns the actual size of the page range, in bytes.
    //  The OS is free to return either the value originally requested by the client
    //  or the actual size of the page range, whichever is most convenient.
    //  Clients have to be able to handle values which are larger than they
    //   asked for.   This routine may be called at any call level.
    //
    // FlushPageCache - Performs a cache operation to bring the processor
    //  into synchronization with physical memory on systems which aren't I/O
    //  cache coherent.  A cache invalidate (CTOS_CACHE_INVALIDATE) causes data
    //  currently in the cache to be flushed rather than being written back to
    //  memory.  This type of cache flush should be used before the client
    //  attempts to read data which has been DMA'ed into memory.  A cache
    //  writeback causes any cached data to be written back to physical memory
    //  and should be used after the user has modified data which is being
    //  DMA'ed to the device.  FlushPageCache cannot be called at ISR time.
    //
    //**************************************************************************
    STDMETHOD_( OSPHYSMEMHDL, AllocPages )        ( THIS_ CTDWORD dwNumBytes, CTDWORD dwFlags ) PURE;
    STDMETHOD_( OSPHYSMEMHDL, AllocContiguousPages )( THIS_ CTDWORD dwNumBytes, interface ICtDevice *device) PURE;
    STDMETHOD ( FreePages )                       ( THIS_ OSPHYSMEMHDL memhdl ) PURE;

    STDMETHOD ( LockPages )                       ( THIS_ OSPHYSMEMHDL memhdl ) PURE;
    STDMETHOD ( UnlockPages )                     ( THIS_ OSPHYSMEMHDL memhdl ) PURE;

    STDMETHOD_( OSVIRTADDR, GetPageVirtAddr )     ( THIS_ OSPHYSMEMHDL memhdl ) PURE;
    STDMETHOD_( OSVIRTADDR, GetPageUserModeVirtAddr) (THIS_ OSPHYSMEMHDL memhdl ) PURE;
    STDMETHOD_( OSPHYSADDR, GetPagePhysAddr )     ( THIS_ OSPHYSMEMHDL memhdl, CTDWORD dwOffset) PURE;
    STDMETHOD_( OSBUSADDR,  GetPageBusAddr  )     ( THIS_ OSPHYSMEMHDL memhdl, CTDWORD dwOffset, interface ICtDevice *device) PURE;
    STDMETHOD_( CTDWORD, GetPageRangeSize   )     ( THIS_ OSPHYSMEMHDL memhdl) PURE;
    STDMETHOD_( CTVOID, FlushPageCache )          ( THIS_ OSPHYSMEMHDL memhdl, CTDWORD dwMode ) PURE;

#define CTOS_PAGEALLOC_DMA_INPUT     0x1      
#define CTOS_PAGEALLOC_DMA_OUTPUT    0x2


// These flags are used
#define CTOS_CACHE_INVALIDATE  0x1
#define CTOS_CACHE_WRITEBACK   0x2

    //**************************************************************************
    //                      Heap Allocate Functions
    //
    // Allocates the specified number of bytes from the system's heap and
    // These function are OS-independent versions of the standard malloc routine.
    // The memory returned by HeapAlloc is pageable.
    // The memory returned by LockHeapAlloc is locked.
    //
    // Alloc functions returns the virtual address of the first allocated byte.
    //**************************************************************************/
    STDMETHOD_( OSVIRTADDR, HeapAlloc )       ( THIS_ CTDWORD dwNumBytes  ) PURE;
    STDMETHOD_( OSVIRTADDR, LockedHeapAlloc ) ( THIS_ CTDWORD dwNumBytes  ) PURE;
    STDMETHOD ( HeapFree )                    ( THIS_ OSVIRTADDR virtAddr ) PURE;
    STDMETHOD ( LockedHeapFree )              ( THIS_ OSVIRTADDR virtAddr ) PURE;

    //**************************************************************************
    //                       Mutex Functions
    //
    // CreateMutex creates a unique Mutex semaphore and writes its handle into
    // pmutexhdl, a pointer to a CTDWORD that will receive the mutex handle.
    //
    // NOTE: Mutexes may only be acquired and released when running at
    //  Passive level.  They cannot be acquired or released at ISR or DPC
    //  time.
    //
    // AcquireMutex Attempts an instantaneous request to obtain the specified mutex
    // If it is available, the OS gives the semaphore to the
    // process and returns TRUE.  If it is unavailable, FALSE is returned.
    // This function does not wait for it to be available.
    // Recursive mutexes are supported - consecutive calls by the same process
    // on the same mutex will all succeed
    //
    // WaitMutex Waits INDEFINITELY for a semaphore to be released
    // by another process.  Once the semaphore is available, it will be
    // given to the calling process and execution will continue.
    // Recursive mutexes are supported - consecutive calls by the same process
    // on the same mutex will not block
    //
    // ReleaseMutex Releases a semaphore that was granted to a process.
    // The mutex must be released for each time it is acquired by the same process
    // before the mutex can be acquired by another process
    //
    // PollMutex test to see if the mutex is available. it is not given to
    // the calling process, even if it is available.
    // If available, this function returns S_OK, otherwise E_FAIL.
    //
    //**************************************************************************/
    STDMETHOD ( CreateMutex  ) ( THIS_ OSMUTEXHDL *pmutexhdl ) PURE;
    STDMETHOD ( AcquireMutex ) ( THIS_ OSMUTEXHDL mutexhdl   ) PURE;
    STDMETHOD ( ReleaseMutex ) ( THIS_ OSMUTEXHDL mutexhdl   ) PURE;
    STDMETHOD ( WaitMutex    ) ( THIS_ OSMUTEXHDL mutexhdl   ) PURE;
    STDMETHOD ( DeleteMutex  ) ( THIS_ OSMUTEXHDL mutexhdl   ) PURE;

    //**************************************************************************
    //                      Semaphore Functions
    //
    // Semaphores are conceptually similar to mutexes, but they can be
    // used to allow multiple threads of execution to enter a guarded region
    // simultaneously.  Also, unlike mutexes, semaphores may be released
    // from code running at Low DPC level.  The following paragraphs include
    // pseudo-code to illustrate how semaphores should work.  So long as an
    // OS maintains the conventions followed in the pseudo code it is free
    // to implement semaphores in whatever manner best suits that particular OS.
    //
    // CreateSemaphore creates a new semaphore and initializes its
    // count to the value specified.  The count can be thought of
    // as the number of threads a semaphore will allow into the guarded
    // region before blocking.  If one specifies an initial count of
    // 1, the semaphore behaves like a mutex and allows only a single
    // thread into the guarded region.  If one specifies a count of
    // 0, the semaphore will initially allow no threads into the
    // critical section.
    //
    // WaitSemaphore decrements the semaphore's count.  If the count is less than 0,
    // the calling thread blocks.  Otherwise, the thread continues to run.
    // WaitSemaphore will return E_FAIL if the semaphore cannot be acquired
    // or some reason, or if the Semaphore is destroyed while the thread
    // is still waiting on it.  WaitSemaphore may only be called at Passive and
    // Low DPC level.
    //      Pseudocode:
    //
    //              BEGIN ATOMIC REGION
    //                  if (semaphore count == 0) {
    //                      Put thread control block of list of sleeping threads
    //                      Yield processor (releasing atomic lock);
    //                  }
    //                  semaphore count--;
    //              END ATOMIC REGION
    //
    // SignalSemaphore increments the semaphore count and awakens any threads
    // which were blocked waiting for the semaphore to be signalled.  Signal
    // may be called at Passive and Low DPC level.
    //      Pseudocode:
    //              BEGIN ATOMIC REGION
    //                  semaphore count ++;
    //                  if (semaphore count == 1) {
    //                      Dequeue the thread with the highest priority
    //                      Mark the thread as runnable.
    //                  }
    //              END ATOMIC REGION
    //
    // DestroySemaphore releases the storage associated with the semaphore.
    // Ideally, if any threads are blocked on the semaphore when the semaphore
    // is destroyed, they should be awakened in such a manner that they will
    // return E_FAIL from WaitSemaphore.  This is not required, however, as
    // it is a programming error to destroy a semaphore while threads are waiting
    // on it.
    //
    STDMETHOD(CreateSemaphore) (THIS_ OSSEMAHDL *phNewSema, CTDWORD dwInitialCount) PURE;
    STDMETHOD(WaitSemaphore) (THIS_ OSSEMAHDL hSema) PURE;
    STDMETHOD(SignalSemaphore) (THIS_ OSSEMAHDL hSema) PURE;
    STDMETHOD(DestroySemaphore) (THIS_ OSSEMAHDL hSema) PURE;

    //**************************************************************************
    //                        Timeout / DPC functions
    // CreateTimeout- creates a timeout object.
    //
    // DestroyTimeout - Destroys a timeout object
    //
    // DestroyTimeout Stops a previously scheduled timeout.
    //
    // ScheduleTimeout schedules a function to execute in the future.
    // This routine schedules functions with a relatively low
    // priority, so it shouldn't be used for things which must
    // execute at precise intervals.  On the other hand, it
    // runs at callback time, so there are no restrictions on
    // what the callback function can do.
    //***************************************************************************
    STDMETHOD ( CreateTimeout   ) ( THIS_ OSTIMEOUTHANDLE *retHandle ) PURE;
    STDMETHOD ( DestroyTimeout  ) ( THIS_ OSTIMEOUTHANDLE ) PURE;
    STDMETHOD ( ScheduleTimeout ) ( THIS_ OSTIMEOUTHANDLE hTimeout,  PFCALLBACK pfCallback,
                                    CTHANDLE hUser, CTDWORD dwMillisecs ) PURE;
    //**************************************************************************
    //                        High-Priority Timeout
    //
    // ScheduleInterruptTimeout schedules a function to execute in the future.
    // This routine schedules functions with an interrupt-time
    // priority, so it shouldn't be used for things which require extensive
    // processing time or paged code segments.  This method does not have
    // an unregister counterpart. Instead, it must be called again after each
    // timeout.  On shutdown, the code must not be unloaded until the callback
    // has been made.
    //
    //***************************************************************************
    STDMETHOD ( CreateInterruptTimeout   ) ( THIS_  OSTIMEOUTHANDLE *retHandle ) PURE;
    STDMETHOD ( DestroyInterruptTimeout  ) ( THIS_ OSTIMEOUTHANDLE ) PURE;
    STDMETHOD ( ScheduleInterruptTimeout ) ( THIS_ OSTIMEOUTHANDLE hTimeout, PFCALLBACK pfCallback,
                                             CTHANDLE hUser, CTDWORD dwMillisecs ) PURE;


    //**************************************************************************
    // CreateDpc- Creates a Dpc object.
    //
    // DestroyDpc - Destroys a Dpc object
    //
    // ScheduleDpc Requests that the OS call the specified routine as soon
    // as possible in a "normal" non-ISR kernel environment
    // The intent of this routine is that it can be called at ISR-time to
    // schedule lower-priority processing in a less restrictive
    // environment.  OS's with no limitations about when routines can
    // be called may turn around and invoke this routine immediately.
    STDMETHOD ( CreateDpc  )  ( THIS_ OSDPCHANDLE* phDpc ) PURE;
    STDMETHOD ( DestroyDpc )  ( THIS_ OSDPCHANDLE hDpc ) PURE;
    STDMETHOD ( ScheduleDpc ) ( THIS_ OSDPCHANDLE hDpc, PFCALLBACK callback, CTHANDLE hUser ) PURE;


    //**************************************************************************
    // Get the current system time.  The system time is the current
    // number of milliseconds since the system started.  While there
    // is no hard requirement that the value of the tick update every
    // millisecond, this is preferable.
    //**************************************************************************/
    STDMETHOD_( CTDWORD, GetSystemTime ) ( THIS ) PURE;


    //**************************************************************************
    // DisableAllInterrupts disables all interrupts for the current thread
    // on the current processor.  This routine is not intended for synchronization
    // but rather for time critical operations.
    //
    // DisableAllInterrupts will return a handle  When the client calls enable
    // allinterrupts he should pass this handle in as a paramater.
    // Do not call this routine for any reason.
    //
    // EnableAllInterrupts enables interrupts.
    //**************************************************************************/
    STDMETHOD_( CTHANDLE, DisableAllInterrupts ) ( THIS ) PURE;
    STDMETHOD_( CTVOID, EnableAllInterrupts    ) ( THIS_ CTHANDLE h ) PURE;

    //**************************************************************************
    // COM -
    //**************************************************************************
    STDMETHOD ( CtCreateInstance )( THIS_ REFGUID clsid, ICtUnknown* pIUnknownOutter,
                                    CTDWORD dwClsContext, REFIID iid, void** ppv ICTUNK_XTRAARGS) PURE;
    STDMETHOD ( CreateHardwareInstance )( THIS_ REFGUID clsid, interface ICtUnknown *pDevice,
                                          REFIID riid, void **ppv ICTUNK_XTRAARGS) PURE;

    STDMETHOD( RegisterServer   )( THIS_ REFGUID clsid, PGETCLASSOBJECTFUNC ) PURE;
    STDMETHOD( UnregisterServer )( THIS_ REFGUID clsid ) PURE;

	STDMETHOD( RegisterStaticServer   )( REFGUID clsid, PGETCLASSOBJECTFUNC pFunc ) PURE;
    STDMETHOD( UnregisterStaticServer )( REFGUID clsid ) PURE;

    STDMETHOD( RegisterAggregate   )( THIS_ ICtUnknown* p ) PURE;
    STDMETHOD( UnregisterAggregate )( THIS_ ICtUnknown* p ) PURE;

    STDMETHOD( RegisterSystemInterface )( THIS_ interface IasSystem* p ) PURE;

    //**************************************************************************
    // vsprintf-
    //**************************************************************************
    STDMETHOD_(int,vsprintf)(THIS_ CTTCHAR *, const CTTCHAR *, va_list) PURE;

    //**************************************************************************
    // Kernel File IO
    //**************************************************************************
	STDMETHOD(OpenFile)           (THIS_ CTPATH* pFileName, CTFILEIO_MODE enMode, PCTHANDLE pHandle) PURE;
    STDMETHOD(CloseFile)          (THIS_ CTHANDLE hFile) PURE;
	STDMETHOD_(CTLONG,ReadFile)   (THIS_ CTHANDLE hFile, PCTVOID pBuffer, CTLONG lSize, CTLONG lElements) PURE;
	STDMETHOD_(CTLONG,WriteFile)  (THIS_ CTHANDLE hFile, const PCTVOID pBuffer, CTLONG lSize, CTLONG lElements) PURE;
	STDMETHOD(SeekFile)           (THIS_ CTHANDLE hFile, CTLONG lOffset, CTFILEIO_SEEK enSeek) PURE;
	STDMETHOD_(CTLONG,TellFile)   (THIS_ CTHANDLE hFile) PURE;

    //*************************************************************************
    //  Returns whether or not the path denoted in pszString is in fact
	//  a full path to a file. If it is just a file name, a partial path
	//  to a file, or a relative path to a file, this will return FALSE.
	//  Otherwise it returns TRUE.
	//  OSSrv and System
    //*************************************************************************
	STDMETHOD_(CTBOOL,IsFullPath  ) ( THIS_ CTPATH * pszString) PURE;

    //*************************************************************************
    //  Returns whether or not the path denoted in pszString is in fact
	//  a partial path to a file. If it is a partial path
	//  to a file, or a relative path to a file, this will return TRUE.
	//  If it is just a file name or a full path, it returns FALSE.
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTBOOL, IsPartialPath) (THIS_ CTPATH * pszString) PURE;

	//*************************************************************************
    //  Takes the data in 'pszFilePath', which is assumed to be a partial-path
	//  to a file, and properly append it to the end of 'pszFullPath'.
	//  DOES NOT SUPPORT 'pszFileString' being a RELATIVE PATH!
	//  Returns a pointer to the original 'pszFullPath'
	//  string, so it WILL change the data you pass it!
	//  'pPathBufferSize' is the size of the Path string buffer IN BYTES!!!
	//  If that buffer is too small, returns with a null string and fills
	//  pPathBufferSize with the correct value.
	//  DOES NOT SUPPORT RELATIVE PATHS!
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTPATH*,AddPathToPath  ) ( THIS_ CTPATH* pszFullPath, const CTPATH* pszFilePath, PCTDWORD pPathBufferSize) PURE;

	//*************************************************************************
    //  Takes the data in 'pszFileString', which is assumed to be an ASCII string
	//  representing a file, and properly append it to the end of 'pszFullPath'.
	//  DOES NOT SUPPORT 'pszFileString' being a RELATIVE PATH OR A PARTIAL PATH!
	//  Returns a pointer to the original 'pszFullPath'
	//  string, so it WILL change the data you pass it!
	//  'pPathBufferSize' is the size of the Path string buffer IN BYTES!!!
	//  If that buffer is too small, returns with a null string and fills
	//  pPathBufferSize with the correct value.
	//  DOES NOT SUPPORT RELATIVE PATHS!
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTPATH*,AddStringToPath  ) ( THIS_ CTPATH* pszFullPath, const CTCHAR* pszFileString, PCTDWORD pPathBufferSize) PURE;

	//*************************************************************************
    //  Takes the data in 'pszFilePath', which is assumed to be a partial-path
	//  to a file, and properly append it to the end of the driver data path.
	//  That full path is then inserted into the 'pszFullPath' buffer.
	//  'pszFullPath' is assumed to be an empty buffer when passed in.
	//  DOES NOT SUPPORT 'pszFileString' being a RELATIVE PATH!
	//  Returns a pointer to the original 'pszFullPath'
	//  string, so it WILL change the data you pass it!
	//  'pPathBufferSize' is the size of the 'pszFullPath' buffer IN BYTES!!!
	//  If that buffer is too small, returns with a null string and fills
	//  pPathBufferSize with the correct value.
	//  DOES NOT SUPPORT RELATIVE PATHS!
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTPATH*,AddPathToDriverDataPath  ) ( THIS_ CTPATH* pszFullPath, const CTPATH* pszFilePath, PCTDWORD pPathBufferSize) PURE;

	//*************************************************************************
    //  Takes the data in 'pszFilePath', which is assumed to be a partial-path
	//  to a file, and properly append it to the end of the driver data path.
	//  That full path is then inserted into the 'pszFullPath' buffer.
	//  'pszFullPath' is assumed to be an empty buffer when passed in.
	//  DOES NOT SUPPORT 'pszFileString' being a RELATIVE PATH OR A PARTIAL PATH!
	//  Returns a pointer to the original 'pszFullPath'
	//  string, so it WILL change the data you pass it!
	//  'pPathBufferSize' is the size of the 'pszFullPath' buffer IN BYTES!!!
	//  If that buffer is too small, returns with a null string and fills
	//  pPathBufferSize with the correct value.
	//  DOES NOT SUPPORT RELATIVE PATHS!
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTPATH*,AddStringToDriverDataPath  ) ( THIS_ CTPATH* pszFullPath, const CTCHAR* pszFileString, PCTDWORD pPathBufferSize) PURE;

		//*************************************************************************
    //  Takes the data in 'pszFilePath', which is assumed to be a partial-path
	//  to a file, and properly append it to the end of the driver preference path.
	//  That full path is then inserted into the 'pszFullPath' buffer.
	//  'pszFullPath' is assumed to be an empty buffer when passed in.
	//  DOES NOT SUPPORT 'pszFileString' being a RELATIVE PATH!
	//  Returns a pointer to the original 'pszFullPath'
	//  string, so it WILL change the data you pass it!
	//  'pPathBufferSize' is the size of the 'pszFullPath' buffer IN BYTES!!!
	//  If that buffer is too small, returns with a null string and fills
	//  pPathBufferSize with the correct value.
	//  DOES NOT SUPPORT RELATIVE PATHS!
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTPATH*,AddPathToDriverPreferencePath  ) ( THIS_ CTPATH* pszFullPath, const CTPATH* pszFilePath, PCTDWORD pPathBufferSize) PURE;

	//*************************************************************************
    //  Takes the data in 'pszFilePath', which is assumed to be a partial-path
	//  to a file, and properly append it to the end of the driver preference path.
	//  That full path is then inserted into the 'pszFullPath' buffer.
	//  'pszFullPath' is assumed to be an empty buffer when passed in.
	//  DOES NOT SUPPORT 'pszFileString' being a RELATIVE PATH OR A PARTIAL PATH!
	//  Returns a pointer to the original 'pszFullPath'
	//  string, so it WILL change the data you pass it!
	//  'pPathBufferSize' is the size of the 'pszFullPath' buffer IN BYTES!!!
	//  If that buffer is too small, returns with a null string and fills
	//  pPathBufferSize with the correct value.
	//  DOES NOT SUPPORT RELATIVE PATHS!
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTPATH*,AddStringToDriverPreferencePath  ) ( THIS_ CTPATH* pszFullPath, const CTCHAR* pszFileString, PCTDWORD pPathBufferSize) PURE;

	//*************************************************************************
    //  Takes the data in 'pszString', which is assumed to be a path to a file,
	//  and return JUST the file name. Returns a pointer to the original
	//  string, so it WILL change the data you pass it!
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTPATH*,StripPath  ) ( THIS_ CTPATH* pszString) PURE;

    //*************************************************************************
    //  Returns the length of the path, considering WideChar or ASCII
	//  system dependencies. Value returned is in NUMBER OF CHARACTERS.
    //*************************************************************************
	STDMETHOD_(CTULONG, GetPathLength) (THIS_ CTPATH *pszString) PURE;

    //*************************************************************************
    //  Returns whether or not the paths denoted in pszString1 and pszString2
	//  are in fact pointing to the same file. Accounts for case insensitive
	//  file systems and Windows long/short path names in the post-boot INIT
	//  stage.
	//  OSSrv
    //*************************************************************************
	STDMETHOD_(CTBOOL,ArePathsIdentical  ) ( THIS_ CTPATH * pszString1, CTPATH * pszString2) PURE;

	//*************************************************************************
    //  Takes the data in 'pszString', which is assumed to be either an HTML
	//  path or a standard system path to a file, and returns the system
	//  path version of the identical file.
	//  In the case of a case-insensitive operating system, path comes back
	//  in all capital letters. Returns a
	//  pointer to the original string, so it WILL change the data you pass it!
	//  Returns NULL if data in 'pBufSize' is too small to hold the final path,
	//  fills pBufSize with the correct size.
	//  OSSrv and System
    //*************************************************************************
	STDMETHOD_(CTPATH*,GetSystemStylePathName  ) ( THIS_ CTPATH* pszString, PCTDWORD pBufSize) PURE;

    //**************************************************************************
    // External binary interfaces
    //**************************************************************************
    STDMETHOD ( RegisterExternalInterface   ) ( THIS_ REFGUID clsid,  ICtUnknown* p ICTUNK_XTRAARGS) PURE;
    STDMETHOD ( UnregisterExternalInterface ) ( THIS_ REFGUID clsid ICTUNK_XTRAARGS ) PURE;
    STDMETHOD ( QueryExternalInterface      ) ( THIS_ REFGUID clsid,  ICtUnknown** p ICTUNK_XTRAARGS ) PURE;

    //**************************************************************************
    // Don't call these methods
    //**************************************************************************
    STDMETHOD ( Lock   ) ( THIS ) PURE;
    STDMETHOD ( Unlock ) ( THIS ) PURE;

#ifdef DEBUG
    STDMETHOD ( GetDebugLogger ) ( THIS_  PLOGFUNC* pDebugLog, CTBOOL bDriverBootComplete ) PURE;
    STDMETHOD ( CloseDebugLog  ) ( THIS ) PURE;
#endif

};

typedef IasOSSrv * PASOSSRV;

#ifdef __cplusplus
inline int os_sprintf( IasOSSrv* p, CTTCHAR * d, const CTTCHAR * f, ... )
{
    va_list marker;

    va_start( marker, f );

    return p->vsprintf( d, f, marker );

    // return COMCALL( IasOSSrv,p,vsprintf )( p, d, f, marker );
}
#endif

#endif /* __IasOSSrv_h */
