/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/
#ifndef _CTALSA_H_
#   define _CTALSA_H_

#   define __NO_VERSION__
#   include <linux/time.h>
#   include <linux/list.h>

#   include <sound/driver.h>
#   include <sound/core.h>
#   include <sound/initval.h>
#   include <sound/pcm.h>
#   include <sound/seq_oss_legacy.h>
#   include <sound/control.h>
#   include <sound/core.h>

#   include "osutils.h"
#   include "drvutils.h"

#   define USE_SPINLOCKS
#   if defined (USE_SPINLOCKS)
#       define SPIN_LOCK(lock)      { unsigned long flags; spin_lock_irqsave(lock, flags);
#       define SPIN_UNLOCK(lock)    spin_unlock_irqrestore(lock, flags); }
#       if LINUX_VERSION_CODE > KERNEL_VERSION(2, 3, 0)
#           include <linux/spinlock.h>
#       else
#           include <asm/spinlock.h>
#       endif
#   else
#       define SPIN_LOCK(lock)
#       define SPIN_UNLOCK(lock)
#   endif

/* Liu Chun */
/* For comparison of ALSA verion */
#define ALSA_VERSION(a,b,c) (((a) << 16) | ((b) << 8) | (c))
#ifndef ALSA_VERSION_CODE
#	define ALSA_VERSION_CODE 65547 /* Default to ALSA 1.0.11 */
#endif
/* End Liu */

// define the needed uda types here
typedef CTHANDLE  OSPHYSMEMHDL;    /* opaque handle used to reference a */
                                   /* particular region of allocated host memory.*/

/*
 * The following definitions define a bit set which describes
 * what sample format encodings a particular device supports.
 */
#define HASF_SIGNED_4ADPCM          0x0001 /* 4-bit ADPCM-encoded samples */
#define HASF_SIGNED_8ADPCM          0x0002 /* 8-bit ADPCM-encoded samples */
#define HASF_UNSIGNED_8PCM          0x0004 /* Unsigned (offset 128) 8-bit PCM */
#define HASF_SIGNED_8PCM            0x0008 /* Signed 8-bit PCM */
#define HASF_UNSIGNED_16PCM         0x0010 /* Unsigned (offset 32768) 16-bit PCM */
#define HASF_SIGNED_16PCM           0x0020 /* Signed 16-bit PCM */
#define HASF_SIGNED_24PCM           0x0040 /* Signed 24-bit PCM */
#define HASF_SIGNED_32PCM           0x0080 /* Signed 32-bit PCM */
#define HASF_FLOAT_32PCM            0x0100 /* IEEE floating point 32-bit signed */
#define HASF_SIGNED_24PACK32PCM     0x0200 /* Signed 24-bit Packed on 32-Bit PCM */

#define HASF_AC3_ENCODED            0x10000
#define HASF_MP3_ENCODED            0x20000

#define CTALSA_NO_SYNTH_PORT        4 // the number of supported synth ports
#define CTALSA_NO_UART_PORT         2 // the number of supported synth ports
/**
 * The following defines all of the messages wave output clients
 * can receive as part of their callback routines.
 */
typedef enum {
    HAWOD_RESOURCE_STOLEN,
    HAWOD_BUFFERS_COMPLETED,
    HAWOD_NOTIFY,
    HAWOD_VOICE_ZERO,
    HAWOD_END_OF_BUFFER,
    HAWOD_BUFFERS_REVOKED,
    HAWOD_LAST_MESSAGE = HAWOD_BUFFERS_REVOKED
} HAWODMESSAGETYPE;

/****************************************************************************
 * The HAMONOTIFYTYPE structure is passed as an argument to the callback
 * to indicate the reason why the callback is occurring.
 */
typedef enum {
    MONOTIFY_SPACE_AVAILABLE
} HAMONOTIFYTYPE;

///Messages used by HAEASS_CALLBACK
enum HAEASS_TRANSPORTMESSAGE
{
    /** Indicates that one or more buffers have been played. 
     *  hArg is a CTHANDLE pointing to the first buffer on the completed list.
     */ 
    HAEASS_MSG_BUFFER_COMPLETED,
    /** Indicates that one or more buffers have been revoked by 
     *  IhaEAStreamSoundSource::RevokeBuffers().
     *  hArg is a CTHANDLE pointing to the first buffer on the revoked list.
     */
    HAEASS_MSG_BUFFER_REVOKED,
    /** Indicates that the current play position has passed the number of sample frames 
     *  specified by IhaEABufferTransport::SetNotificationFrequency().
     *  hArg is NULL.
     */
    HAEASS_MSG_NOTIFY,
    /** Indicates that a voice or voices have been stolen from the IhaEATransport.
     *  hArg is NULL.
     */
    HAEASS_MSG_RIPOFF
};

#   include "ctdrv.h"
#if ALSA_VERSION_CODE > ALSA_VERSION(1, 0, 13)
#include "typedefs.h"
#endif

typedef struct tKontrolNode {
    enum CTALSA_MIXER_DEV kontrol_dev; // kontrol type
    CTBOOL bVolCtl;
    char *szVolCtlName; // kontrol name
    snd_kcontrol_new_t VolCtl;
    CTBOOL bSwhCtl;
    char *szSwhCtlName;
    snd_kcontrol_new_t SwhCtl;
    CTBOOL bRecCtl;
    char *szRecCtlName;
    snd_kcontrol_new_t RecCtl;
} stKontrolNode;

typedef struct tKontrolList {
    snd_kcontrol_t *pKctl;
    struct list_head list;
} stKontrolList;

typedef int (*FPDEVINIT)(stHwObj *pHwObj, char *szName);
typedef void (*FPDEVEXIT)(stHwObj *pHwObj, int nDev);


/* extern function prototypes */
extern int  wavein_init     (stHwObj *pHwObj, char *szName);
extern int  envaudio_init   (stHwObj *pHwObj, char *szName);
extern int  waveout_init    (stHwObj *pHwObj, char *szName);
extern int  mixer_init      (stHwObj *pHwObj, char *szName);
extern int  uart_init       (stHwObj *pHwObj, char *szName);
extern int  auto_init       (stHwObj *pHwObj, char *szName);
extern int  synth_init      (stHwObj *pHwObj, char *szName);
extern int  datapass_init   (stHwObj *pHwObj, char *szName);
extern int  timer_init      (stHwObj *pHwObj, char *szName);
extern int  ac3mgr_init     (stHwObj *pHwObj, char *szName);

extern void wavein_exit     (stHwObj *pHwObj, int nDev);
extern void envaudio_exit   (stHwObj *pHwObj, int nDev);
extern void waveout_exit    (stHwObj *pHwObj, int nDev);
extern void mixer_exit      (stHwObj *pHwObj, int nDev);
extern void uart_exit       (stHwObj *pHwObj, int nDev);
extern void synth_exit      (stHwObj *pHwObj, int nDev);
extern void datapass_exit   (stHwObj *pHwObj, int nDev);
extern void timer_exit      (stHwObj *pHwObj, int nDev);
extern void ac3mgr_exit     (stHwObj *pHwObj, int nDev);

extern int ctalsa_proc_init(stHwObj *pHwObj);

HRESULT reserve_remap_pages(unsigned long ulUserAddr, stMemHandle *pMem, unsigned long ulSize, pgprot_t prot);

#endif /* _CTALSA_H_ */

