/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
UDA uses COM as a standard way to define module interfaces. Since COM is
a Windows standard, it has no equivalents in non-Windows platforms. Also,
COM is only loosely supported in some Windows kernel environments. For
this reason, UDA uses this header file to define a common COM interface
implementation that is guaranteed to compile and run on all platforms.

$Date: 2008/03/12 09:52:28 $
$Revision: 1.101.18.4 $
 
Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#ifndef __CTCOMDEF_H
#define __CTCOMDEF_H

#ifdef CTCONFIG_H
  /** PreCheck: If for some reason the build environment precludes using 
   *            numerous command-line compile flags in your, the project settings
   *            should include this define that points to a header file
   *            with the rest of the compile flags in it. We assume that the compiler can
   *            handle at least one define as a parameter. That define in this case
   *            should be called CTCONFIG_H and it should define the relative path
   *            to a specific header file. Here is where that gets included.
   */
#  include CTCONFIG_H
#endif

/******************************************************************************
* PreCheck: This provides protection against including this with older
*           versions of the same file that may have different definitions.
*******************************************************************************/

#ifdef __COMstuff_H
#  error CtCOMdef.H: Attempt to use with legacy COMStuff.H
#endif

/******************************************************************************
* Macintosh system file includes
*
*******************************************************************************/
#ifdef __CT_SYS_MACINTOSH
#	include <DriverSynchronization.h>
#endif

#ifdef __CT_SYS_OSX_PPC
#   include "ctdef.h"

#   ifdef __cplusplus
      extern "C" {
#   endif

    extern int memcmp __P((const void *, const void *, size_t));

#   ifdef __cplusplus
	  }
#   endif
#endif

/******************************************************************************
* Windows 95/98 User mode definitions
*
*******************************************************************************/
#if defined(__CT_SYS_WINDOWS_95) && !defined(VTOOLSD) && !defined(CT_CORECODE)

#   pragma warning(disable: 4237 4284 4290 4514 4786)
#   include "ctdef.h"
#   include <memory.h>
#   include <string.h>
#   pragma intrinsic(memcpy,memset,strlen,strcpy,memcmp,strcmp)
#   define COM_NO_WINDOWS_H
#   include <objbase.h>
#   include <cguid.h>

/******************************************************************************
* Windows 95/98 VxD definitions requiring Microsoft header files
*   NOTE: VTOOLSD header files cannot be used with the NEED_MSTYPES option
*         So, ctdef.h must not be included before ctCOMdef.h
*******************************************************************************/
#elif defined(__CT_SYS_WINDOWS_95) && defined(VTOOLSD) && defined(NEED_MSTYPES)

#   ifdef __vtoolscp_h_
#       error You cannot use VTOOLSD header files with the NEED_MSTYPES flag set!
#   endif
#   define COM_NO_WINDOWS       /* Use the microsoft header files */
#   include <objbase.h>
#   define __vtoolscp_h_        /* VTOOLSD just won't play nice with MS */
#   undef VHACK                 /* Can we get rid of this?? */
#   include "ctdef.h"

#else /* Not Windows 95 user or kernel requiring MS types */
#   include "ctdef.h"           /* Include standard datatypes for all other platforms */
#   if (CT_IS_WINDOWS)
        // Disable these annoying benign warnings given by STL
#		pragma intrinsic(memcpy,memset,strlen,strcpy,memcmp,strcmp,strcat)
#       pragma warning(disable: 4237 4284 4290 4514 4786)
#		ifdef __cplusplus
			extern "C" {
#		endif
        #if (!CT_IS_WINDOWS_NT) // NT has these defined already.
        extern CTLONG __stdcall InterlockedIncrement(PCTLONG);
        extern CTLONG __stdcall InterlockedDecrement(PCTLONG);
        #endif
#		ifdef __cplusplus
			}
#		endif
#   endif
/******************************************************************************
* Windows 98 WDM, Win 2K, or WinNT definitions
*******************************************************************************/
#   if (CT_IS_WINDOWS) && !defined(__CT_SYS_WINDOWS_95)
#     if defined(_USRDLL)
#        include <Winerror.h>
#        define E_ASYNC             0x80070009L
#        define CLSCTX_INPROC_SERVER    0x1
#		 define UserMode 1
#     else
#       if defined(__CT_WIN32_DRIVER_MODEL)

#         include <Winerror.h>   /* Include HRESULT codes */
#         define E_ASYNC             0x80070009L

          typedef enum tagCLSCTX
          {   CLSCTX_INPROC_SERVER    = 0x1,
              CLSCTX_INPROC_HANDLER   = 0x2,
              CLSCTX_LOCAL_SERVER     = 0x4,
              CLSCTX_INPROC_SERVER16  = 0x8,
              CLSCTX_REMOTE_SERVER    = 0x10,
              CLSCTX_INPROC_HANDLER16 = 0x20,
              CLSCTX_INPROC_SERVERX86 = 0x40,
              CLSCTX_INPROC_HANDLERX86= 0x80
          } CLSCTX;
#       else
          /* Nothing to do, CTDEF.H includes all COM definitions */
#         include <Winerror.h>   /* Include HRESULT codes */
#         define E_ASYNC             0x80070009L
#       endif
#     endif

/******************************************************************************
* All other platforms including Windows 95/VTOOLSD, Linux, UNIX, and Macintosh
* need COM definitions here
*******************************************************************************/
#   else /* defined(CT_IS_WINDOWS) && !defined(__CT_SYS_WINDOWS_95) */

#       if defined(__CT_SYS_WINDOWS_95) && !defined(CT_CORECODE)
#           include <Winerror.h>   /* Include HRESULT codes */
#       else
            /* Somebody needs to create an HRESULT.H */
#           include "ctcomerr.h"
#       endif
#       if !defined(VTOOLSD) && !defined(__CT_SYS_LINUX) && !defined(CT_IS_OSX)
#           include <memory.h>
#       endif
#       if defined(__CT_SYS_LINUX) && defined(MODULE)
#           include "cttypes.h"
#           include <linux/string.h>
#       endif

#ifndef __LPVOID_DEFINED__
#define __LPVOID_DEFINED__
        typedef PCTVOID LPVOID;    /* This type should be replaced in all files by PCTVOID */
#endif

#       ifdef __cplusplus
#           define EXTERN_C     extern "C"
#           define DEFAULT_VALUE_(p, v)  p=v    
#       else
#           define EXTERN_C     extern
#           define DEFAULT_VALUE_(p, v)  p    
#       endif


#       ifdef __CT_SYS_WINDOWS_95
#           define STDMETHODCALLTYPE   __stdcall
#           define STDAPICALLTYPE      __stdcall
#       elif defined (__CT_SYS_LINUX) || defined (__CT_SYS_MACINTOSH) || (CT_IS_SOLARIS) || (CT_IS_OSX)
#           define STDMETHODCALLTYPE   /*__attribute__ ((stdcall)) */
#           define STDAPICALLTYPE      /* __attribute__ ((stdcall)) */
#       else
#           define STDMETHODCALLTYPE   __cdecl
#           define STDAPICALLTYPE      __cdecl
#       endif

#       if defined (__CT_SYS_LINUX)   || defined (__CT_SYS_MACINTOSH) || (CT_IS_SOLARIS) || (CT_IS_OSX)
#           define STDMETHODVCALLTYPE
#           define STDAPIVCALLTYPE
#       else
#           define STDMETHODVCALLTYPE      __cdecl
#           define STDAPIVCALLTYPE         __cdecl
#       endif

#       if defined (__CT_SYS_LINUX) || (CT_IS_SOLARIS)
#           define STDMETHODMCALLTYPE
#           define STDAPIMCALLTYPE
#       else
#           define STDMETHODMCALLTYPE      __cdecl
#           define STDAPIMCALLTYPE         __cdecl
#       endif

#       define STDAPI                  EXTERN_C HRESULT STDAPICALLTYPE
#       define STDAPI_(type)           EXTERN_C type STDAPICALLTYPE
#       define STDMETHODIMP            HRESULT STDMETHODCALLTYPE
#       define STDMETHODIMP_(type)     type STDMETHODCALLTYPE

        /* The 'V' versions are only for Variable Argument lists in DEBUG mode that translate into
           zero arguments in RELEASE mode.  If more than zero arguments are needed in RELEASE mode,
           you must use the 'M' versions, sorry.  -EL */
#       define STDAPIV                 EXTERN_C HRESULT STDAPIVCALLTYPE
#       define STDAPIV_(type)          EXTERN_C type STDAPIVCALLTYPE
#       define STDMETHODIMPV           HRESULT STDMETHODVCALLTYPE
#       define STDMETHODIMPV_(type)    type STDMETHODVCALLTYPE

#       define STDAPIM                 EXTERN_C HRESULT STDAPIMCALLTYPE
#       define STDAPIM_(type)          EXTERN_C type STDAPIMCALLTYPE
#       define STDMETHODIMPM           HRESULT STDMETHODMCALLTYPE
#       define STDMETHODIMPM_(type)    type STDMETHODMCALLTYPE

        /* These macros build interfaces in either C++ or C */
#       define interface               struct
#       if defined(__cplusplus) && !defined(CINTERFACE)
#           define STDMETHOD(method)       virtual HRESULT STDMETHODCALLTYPE method
#           define STDMETHOD_(type,method) virtual type STDMETHODCALLTYPE method
#           define PURE                    = 0
#           define THIS_
#           define THIS                    void
#           define DECLARE_INTERFACE(iface)    interface iface
#           define DECLARE_INTERFACE_(iface, baseiface)    interface iface : public baseiface
#       else
#           define STDMETHOD(method)       HRESULT (STDMETHODCALLTYPE * method)
#           define STDMETHOD_(type,method) type (STDMETHODCALLTYPE * method)
#           define PURE
#           define THIS_                   CTVOID * This,
#           define THIS                    CTVOID * This
#           ifdef CONST_VTABLE
#               define DECLARE_INTERFACE(iface)    typedef interface iface {                    \
                                                        const struct iface##Vtbl * lpVtbl;      \
                                                   } iface;                                     \
                                                   typedef const struct iface##Vtbl iface##Vtbl;\
                                                   const struct iface##Vtbl
#           else
#               define DECLARE_INTERFACE(iface)    typedef interface iface {                    \
                                                        struct iface##Vtbl * lpVtbl;            \
                                                   } iface;                                     \
                                                   typedef struct iface##Vtbl iface##Vtbl;      \
                                                   struct iface##Vtbl
#           endif
#           define     DECLARE_INTERFACE_(iface, baseiface)    DECLARE_INTERFACE(iface)
#       endif /* defined(__cplusplus) & !defined(CINTERFACE) */

        /* If we are initializing the GUID, create the struct, otherwise, extern it */
#       ifndef INITGUID
#           define DEFINE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
            EXTERN_C const GUID name
#       else
#           define DEFINE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
            EXTERN_C const GUID name; \
            const GUID name \
                = { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }
#       endif /* INITGUID */


        /** GUID (pronounced goo-id or sometimes as gwid) stands for Guaranteed 
		  * Unique Identifier. A GUID is a value that is guaranteed
		  * to be unique across all computer systems and all (practical) time. GUIDs
		  * are 16-byte or 128 bit values, that are encapsulated in this structure.
		  * GUIDs should never be generated manually, they should be generated using
		  * a tool that is guaranteed to honor the GUID standards.
		  */
        typedef struct _GUID {
            CTULONG  Data1;
            CTUSHORT Data2;
            CTUSHORT Data3;
            CTUCHAR  Data4[8];
        } GUID;

		/* IMPORTANT! If you change this, scroll down and change
		   the CT_HRESULT size values! */
        typedef CTLONG HRESULT;

        typedef GUID IID;

#       define IID_NULL            GUID_NULL
#       define IsEqualIID(riid1, riid2) IsEqualGUID(riid1, riid2)

        typedef GUID CLSID;
#       define CLSID_NULL          GUID_NULL
#       define IsEqualCLSID(rclsid1, rclsid2) IsEqualGUID(rclsid1, rclsid2)

        typedef GUID FMTID;
#       define FMTID_NULL          GUID_NULL
#       define IsEqualFMTID(rfmtid1, rfmtid2) IsEqualGUID(rfmtid1, rfmtid2)

#       if defined(__cplusplus)
#           define REFGUID             const GUID &
#           define REFIID              const IID &
#           define REFCLSID            const CLSID &
#           define REFFMTID            const FMTID &
#       else
#           define REFGUID             const GUID * const
#           define REFIID              const IID * const
#           define REFCLSID            const CLSID * const
#           define REFFMTID            const FMTID * const
#       endif /* __cplusplus */

#       if defined(__cplusplus)
            __inline CTBOOL operator==(const GUID& guidOne, const GUID& guidOther)
            {
                return !memcmp(&guidOne,&guidOther,sizeof(GUID));
            }

            __inline CTBOOL operator!=(const GUID& guidOne, const GUID& guidOther)
            {
                return !(guidOne == guidOther);
            }
#       endif

        /* In Windows95, IUnknown, GUID_NULL, and IsEqualGUID are defined in UUID.LIB */
#       if defined(__CT_SYS_WINDOWS_95) && !defined(_CT_WIN95_DONT_USE_UUIDLIB)
            EXTERN_C int IsEqualGUID(REFGUID, REFGUID); /* This is used only by C clients */
#       else
#           define IsEqualGUID( riid1, riid2 ) (!(memcmp( riid1, riid2, sizeof(GUID) )))
#       endif

        EXTERN_C const IID IID_IUnknown;
        EXTERN_C const IID IID_IClassFactory;
        EXTERN_C const GUID GUID_NULL;

#       undef INTERFACE
#       define INTERFACE IUnknown

        /** This defines the IUnknown interface. The IUnknown interface is the one interface
		  * that every COM object must support. The UDA COM implementation supports IUnknown
		  * in a standard manner, so a standard COM reference should adequately explain the
		  * usage and semantics.
		  */
        DECLARE_INTERFACE(IUnknown)
        {
            /* IUnknown methods: */
            STDMETHOD(QueryInterface)			(THIS_ REFIID, PCTVOID *) PURE;
            STDMETHOD_(CTDWORD,AddRef)			(THIS) PURE;
            STDMETHOD_(CTDWORD,Release)			(THIS) PURE;
        };


        typedef enum tagCLSCTX
        {	CLSCTX_INPROC_SERVER	= 0x1,
	        CLSCTX_INPROC_HANDLER	= 0x2,
	        CLSCTX_LOCAL_SERVER	    = 0x4,
	        CLSCTX_INPROC_SERVER16	= 0x8,
	        CLSCTX_REMOTE_SERVER	= 0x10,
	        CLSCTX_INPROC_HANDLER16	= 0x20,
	        CLSCTX_INPROC_SERVERX86	= 0x40,
	        CLSCTX_INPROC_HANDLERX86= 0x80
        }	CLSCTX;



        // IClassFactory interface
#       undef INTERFACE
#       define INTERFACE IClassFactory

        /** This defines the IClassFactory interface. The IClassFactory must be supported
		  * by any object whose purpose is to create COM objects. The UDA COM implementation 
		  * supports IClassFactory in a standard manner, so a standard COM reference should 
		  * adequately explain the usage and semantics.
		  */
        DECLARE_INTERFACE_( IClassFactory, IUnknown )
        {
            // IUnknown methods
            STDMETHOD ( QueryInterface ) ( THIS_ REFIID, PCTVOID * ) PURE;
            STDMETHOD_( CTULONG,AddRef   ) ( THIS ) PURE;
            STDMETHOD_( CTULONG,Release  ) ( THIS ) PURE;

            STDMETHOD ( CreateInstance )( THIS_ IUnknown* pIUnknownOutter, REFGUID iid, void** ppv ) PURE;
            STDMETHOD ( LockServer )    ( THIS_ CTBOOL bLock ) PURE;
        };

#       if defined(__CT_SYS_WINDOWS_95) //&& defined(VTOOLSD)
#           pragma warning(disable: 4237 4284 4290 4514 4786)
#       endif /* if defined(__CT_SYS_WINDOWS_95) && defined(VTOOLSD) */

#   endif /* defined(CT_IS_WINDOWS) && !defined(__CT_SYS_WINDOWS_95) */

#endif /* Not Windows 95 user or kernel requiring MS types  */

/* This won't bloat code like InlineIsEqualGUID(), but it is comparably fast for typical GUID compares */
#if defined(__cplusplus)
__inline CTBOOL FastIsEqualGUID(REFGUID rguid1, REFGUID rguid2)
{
   return ((CTDWORD*)&rguid1)[0] == ((CTDWORD*)&rguid2)[0] && 
          !memcmp(&((CTDWORD*)&rguid1)[1], &((CTDWORD*)&rguid2)[1], sizeof(GUID)-sizeof(CTDWORD));
}
#else
#define FastIsEqualGUID(rguid1, rguid2)     (((CTDWORD*)&rguid1)[0] == ((CTDWORD*)&rguid2)[0] && \
                                            !memcmp(&((CTDWORD*)&rguid1)[1], &((CTDWORD*)&rguid2)[1], sizeof(GUID)-sizeof(CTDWORD)))

#endif

/* This is here because the WDM DDK got IID_IUnknown wrong */
EXTERN_C const IID IID_IRealUnknown;

/* For whatever reason, even Microsoft doesn't define these */
#if defined(__cplusplus) && !defined(CINTERFACE)
#   define STDMETHODV(method)       virtual HRESULT STDMETHODVCALLTYPE method
#   define STDMETHODV_(type,method) virtual type STDMETHODVCALLTYPE method
#else
#   define STDMETHODV(method)       HRESULT (STDMETHODVCALLTYPE * method)
#   define STDMETHODV_(type,method) type (STDMETHODVCALLTYPE * method)
#endif

#define CREATE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
    EXTERN_C const GUID name; \
    const GUID name = { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }

#define CREATE_IID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
    EXTERN_C const IID name; \
    const IID name = { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }

/* Declare ICtUnknown for all platforms */
EXTERN_C const IID IID_ICtUnknown;

/* Declare ICtClassFactory for all platforms */
EXTERN_C const IID IID_ICtClassFactory;

#ifdef DEBUG_MEMORY_OPERATIONS
#   define ICTUNK_IARGLIST   THIS_ const CTCHAR* _file, int _line
#   define ICTUNK_ARGLIST    const CTCHAR* _file, int _line
#   define ICTUNK_QIARGLIST  THIS_ REFIID iid, PCTVOID *ppv, const CTCHAR* _file, int _line
#   define ICTUNK_QARGLIST   REFIID iid, PCTVOID *ppv, const CTCHAR* _file, int _line
#   define ICTUNK_XTRAARGS   ,const CTCHAR* _file, int _line
#else
#   define ICTUNK_IARGLIST   THIS
#   define ICTUNK_ARGLIST
#   define ICTUNK_QIARGLIST  THIS_ REFIID iid, PCTVOID * ppv
#   define ICTUNK_QARGLIST   REFIID iid, PCTVOID * ppv
#   define ICTUNK_XTRAARGS
#endif

#undef INTERFACE
#define INTERFACE ICtUnknown

/** And then there is ICtUnknown, where shall I start? ICtUnknown was implemented as a way
  * to add reference count leak detection instrumentation into the UDA codebase. ICtUnknown
  * honors all of the properties of the standard IUnknown, but each method takes extra parameters
  * indicating something about the calling function for the purposes of the instrumentation.
  * 
  * One such instrumentation technique builds a list of all AddRef and Release usage for all
  * objects in memory, but only in a debug build.
  *
  * Another logs this data directly to a file.
  *
  * As of the date of this documentation, the ICtUnknown interface is used all over the UDA
  * codebase, but its use in terms of instrumentation is quite minimal. Don't be surprised if
  * this whole thing disappears someday.
  */
DECLARE_INTERFACE(ICtUnknown)
{
    /* ICtUnknown methods */
    STDMETHOD(QueryInterface)			(ICTUNK_QIARGLIST) PURE;
    STDMETHOD_(CTDWORD,AddRef)          (ICTUNK_IARGLIST) PURE;
    STDMETHOD_(CTDWORD,Release)         (ICTUNK_IARGLIST) PURE;
};

typedef ICtUnknown * PCTUNKNOWN;

#ifdef DEBUG_MEMORY_OPERATIONS
    typedef enum {
            ICTUNK_AddRef, ICTUNK_Release
    } _CTREFMODE;
    typedef struct _ReferenceTraceTag {
        _CTREFMODE  _action;
        const CTCHAR* _file;
        int         _line;
        int         _Adds;
        int         _Rels;
    } _ReferenceTrace;
#   define CTREFTRACESIZE   256
#   define ICTUNK_MEMVAR    _ReferenceTrace m__astTrace[CTREFTRACESIZE]; CTDWORD m__TraceHead; int m__Adds, m__Rels;
#   define ICTUNK_CONST     m__TraceHead = 0; m__Adds = 0; m__Rels = 0
#   if (CT_IS_WINDOWS)
#       define ICTUNK_TRACEADD  if( _line != 0x7fffffff ) {\
                                    m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._action = ICTUNK_AddRef; \
                                    if( (m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._file = strrchr(_file,'\\')) == NULL) \
                                         m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._file = _file; \
                                    m__astTrace[(m__TraceHead)%CTREFTRACESIZE]._line   = _line;\
                                    m__astTrace[(m__TraceHead)%CTREFTRACESIZE]._Adds = ++m__Adds;\
                                    m__astTrace[(m__TraceHead++)%CTREFTRACESIZE]._Rels = m__Rels;\
                                }
#       define ICTUNK_TRACEREL  if( _line != 0x7fffffff ) {\
                                    m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._action = ICTUNK_Release;\
                                    if( (m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._file = strrchr(_file,'\\')) == NULL) \
                                        m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._file = _file; \
                                    m__astTrace[(m__TraceHead)%CTREFTRACESIZE]._line   = _line;\
                                    m__astTrace[(m__TraceHead)%CTREFTRACESIZE]._Adds = m__Adds;\
                                    m__astTrace[(m__TraceHead++)%CTREFTRACESIZE]._Rels = ++m__Rels;\
                                }
#   else
#       define ICTUNK_TRACEADD  m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._action = ICTUNK_AddRef; \
                                m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._file = _file; \
                                m__astTrace[(m__TraceHead++)%CTREFTRACESIZE]._line   = _line
#       define ICTUNK_TRACEREL  m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._action = ICTUNK_Release;\
                                m__astTrace[ m__TraceHead   %CTREFTRACESIZE]._file = _file; \
                                m__astTrace[(m__TraceHead++)%CTREFTRACESIZE]._line   = _line
#   endif
#   define ICTUNK                   __FILE__, __LINE__
#   define ICTUNK_NOTRACE           __FILE__, 0x7fffffff
#   define ICTUNK_PASSTHRU          _file, _line
#   define QICTUNK_PASSTHRU         ,_file, _line
#   define CICTUNK(thisptr)         thisptr, __FILE__, __LINE__
#   define CICTUNK_NOTRACE(thisptr) thisptr, __FILE__, 0x7fffffff
#   define CICTUNK_PASSTHRU(_this)  _this, _file, _line
#   define QICTUNK                  , ICTUNK
#   define QICTUNK_NOTRACE          , ICTUNK_NOTRACE

#   ifdef _CT_LOG_TO_DISK
#       include "refcountlog.h"
#else
#       define CT_LOG_ADDREF( CLIENT_STRING )
#       define CT_LOG_RELEASE( CLIENT_STRING )
#       define CT_LOG_DESTRUCT()
#   endif

#else
#   define ICTUNK_MEMVAR
#   define ICTUNK_CONST
#   define ICTUNK_TRACEADD
#   define ICTUNK_TRACEREL
#   define ICTUNK
#   define CICTUNK(thisptr) thisptr
#   define ICTUNK_NOTRACE
#   define ICTUNK_PASSTHRU
#   define QICTUNK_PASSTHRU
#   define CICTUNK_NOTRACE(thisptr) thisptr
#   define CICTUNK_PASSTHRU(_this)  _this
#   define QICTUNK
#   define QICTUNK_NOTRACE
#   define CT_LOG_ADDREF( CLIENT_STRING )
#   define CT_LOG_RELEASE( CLIENT_STRING )
#   define CT_LOG_DESTRUCT()

#endif

#define IMP_ICTUNKNOWN \
    STDMETHOD(QueryInterface)  (ICTUNK_QIARGLIST) PURE; \
    STDMETHOD_(CTDWORD,AddRef) (ICTUNK_IARGLIST) PURE; \
    STDMETHOD_(CTDWORD,Release)(ICTUNK_IARGLIST) PURE

#define IMP_DECL_ICTUNK \
    ICTUNK_MEMVAR \
    STDMETHODIMP QueryInterface(ICTUNK_QARGLIST); \
    STDMETHODIMP_(CTDWORD ) AddRef(ICTUNK_ARGLIST); \
    STDMETHODIMP_(CTDWORD ) Release(ICTUNK_ARGLIST)

// ICtClassFactory interface
#undef INTERFACE
#define INTERFACE ICtClassFactory

	
/** And then there is ICtClassFactory, where shall I start? ICtClassFactory was implemented as a way
  * to add reference count leak detection instrumentation into the UDA codebase. ICtClassFactory
  * honors all of the properties of the standard IClassFactory, but each method takes extra parameters
  * indicating something about the calling function for the purposes of the instrumentation.
  * 
  * One such instrumentation technique builds a list of all AddRef and Release usage for all
  * objects in memory, but only in a debug build.
  *
  * Another logs this data directly to a file.
  *
  * As of the date of this documentation, the ICtClassFactory interface is used all over the UDA
  * codebase, but its use in terms of instrumentation is quite minimal. Don't be surprised if
  * this whole thing disappears someday.
  */
DECLARE_INTERFACE_( ICtClassFactory, ICtUnknown )
{
    // IUnknown methods
    IMP_ICTUNKNOWN;

    // ICtClassFactory methods
#ifdef DEBUG_MEMORY_OPERATIONS
    STDMETHOD ( CreateInstance )( THIS_ ICtUnknown* pIUnknownOutter, REFGUID iid, void** ppv ,const CTCHAR* _file, int _line) PURE;
#else
    STDMETHOD ( CreateInstance )( THIS_ ICtUnknown* pIUnknownOutter, REFGUID iid, void** ppv ) PURE;
#endif
    STDMETHOD ( LockServer )    ( THIS_ CTBOOL bLock ) PURE;
};

typedef ICtClassFactory * PCTCLASSFACTORY;

// Standard GetClassObject function prototype
typedef HRESULT (* PCTSTD_GETCLASSOBJECT) (REFCLSID, REFIID, PCTVOID *);

// INondelegatingUnknown Interface
#undef INTERFACE
#define INTERFACE ICtNondelegatingUnknown

/** And then there is ICtNonDelegatingUnknown, where shall I start? ICtNonDelegatingUnknown was implemented as a way
  * to add reference count leak detection instrumentation into the UDA codebase. ICtNonDelegatingUnknown
  * honors all of the properties of the standard IUnknown, with two exceptions. First, this
  * interface is guaranteed never to delegate calls to any other COM object, thus guaranteeing
  * an "end" to any recursive IUnknown calls. Second, each method takes extra parameters
  * indicating something about the calling function for the purposes of the instrumentation.
  * 
  * One such instrumentation technique builds a list of all AddRef and Release usage for all
  * objects in memory, but only in a debug build.
  *
  * Another logs this data directly to a file.
  *
  * As of the date of this documentation, the instrumentation features of the ICtNonDelegatingUnknown 
  * interface is used all over the UDA codebase, but its use in terms of instrumentation is quite 
  * minimal. Don't be surprised if the instrumentation features disappear someday.
  */
DECLARE_INTERFACE( ICtNondelegatingUnknown )
{
    STDMETHOD ( NondelegatingQueryInterface  )  ( ICTUNK_QIARGLIST ) PURE;
    STDMETHOD_( CTULONG,NondelegatingAddRef  ) ( ICTUNK_IARGLIST ) PURE;
    STDMETHOD_( CTULONG,NondelegatingRelease ) ( ICTUNK_IARGLIST ) PURE;
    STDMETHOD ( Init )                          ( THIS_ ICtUnknown* ) PURE;
    IMP_ICTUNKNOWN;
};

typedef ICtNondelegatingUnknown * PCTNONDELEGATINGUNKNOWN;

#define IMP_ICTNONDEL \
    STDMETHOD  (NondelegatingQueryInterface)   (ICTUNK_QIARGLIST) PURE; \
    STDMETHOD_(CTDWORD,NondelegatingAddRef)   (ICTUNK_IARGLIST) PURE; \
    STDMETHOD_(CTDWORD,NondelegatingRelease)  (ICTUNK_IARGLIST) PURE; \
    STDMETHOD  ( Init )                        (THIS_ ICtUnknown* ) PURE; \
    IMP_ICTUNKNOWN

#define IMP_DECL_ICTNONDEL \
    STDMETHODIMP NondelegatingQueryInterface(ICTUNK_QARGLIST);   \
    STDMETHODIMP_(CTDWORD ) NondelegatingAddRef(ICTUNK_ARGLIST);   \
    STDMETHODIMP_(CTDWORD ) NondelegatingRelease(ICTUNK_ARGLIST);  \
    STDMETHODIMP Init( ICtUnknown* pIUnknownOutter );               \
    IMP_DECL_ICTUNK

#define NONDELEGATED_THIS  reinterpret_cast< ICtUnknown* > \
                        (reinterpret_cast< ICtNondelegatingUnknown* >(this))

// Localization definitions

typedef GUID CTLOCSTR;      ///< This refers to a GUID used to indicate a single string located in the string repository
typedef GUID CTLOCUNIT;     ///< This refers to a GUID used to indicate a single unit located in the string repository. A unit defines an ordering and a string.

/** This structure defines a displayable value. 
 *
 * A displayable value has three parts. 
 *
 * First, the format of the value to be displayed. Second, a localizable unit GUID which refers to an
 * appropriate string format given the language. Third, the value that is to be displayed.
 *
 * Example code on using this structure follows
 * 
 @verbatim
  
   CTLOCDISP disp = {0};
   strcpy(disp.szFmt, "%x"); // This means the value will be hexadecimal integer
   disp.dwValue = 4; // The value to be displayed
   disp.locUnit = CTLOCUNIT_DisplayString; // This refers to an entry in the langauge database 
                                           // In English, this entry takes on the format "Value: %s"
                                           // In Spanish, this entry takes on the format "El Numero: %s"
  
  This will result in the user application displaying:

  Value: 4 
    in English

  El Numero: 4
    in Spanish
 
  Note there is no localization in the above code, all localization happens in the string database
  and user application display code.

 @endverbatim
 */
typedef struct _ctlocdisp {
    CTCHAR szFmt[32];           ///< The (sprintf-like) format for LocValue.  Can be %s, or any format of %d, %x, %X, %f, or %g.
    union {
        CTLOCUNIT   locUnit;    ///< If szFmt == numeric formatting (%d, %x, etc), the localizable unit (GUID_NULL for none)
        CTLOCSTR    lstrStr;    ///< If szFmt == "%s", the localizable string
    } LocParam;
    union {
        CTDWORD     dwValue;    ///< If szFmt == "%d", "%x", or "%X", the value
        CTFLOAT     fValue;     ///< If szFmt == "%f" or "%g", the value (Inf, -Inf, and NaN equate to strings)
    } LocValue;
} CTLOCDISP;

#if defined(__cplusplus)
#   define REFCTLOCSTR             const CTLOCSTR &
#   define REFCTLOCUNIT            const CTLOCUNIT &
#   define REFCTLOCDISP            const CTLOCDISP &
#else
#   define REFCTLOCSTR             const CTLOCSTR * const
#   define REFCTLOCUNIT            const CTLOCUNIT * const
#   define REFCTLOCDISP            const CTLOCDISP * const
#endif /* __cplusplus */

#define CTLOCSTR_NULL   GUID_NULL
#define CTLOCUNIT_NULL  GUID_NULL

#define CTSZVERSION(major,minor) ((major)*0x100 + (minor))

#define CT_HRESULT_MINVAL  CT_LONG_MINVAL
#define CT_HRESULT_MAXVAL  CT_LONG_MAXVAL

#if CT_IS_WINDOWS_NT
#define FLATTENGUID(guid)\
    FLAT_##guid
#else
#define FLATTENGUID(x)\
    {x.Data1, x.Data2, x.Data3, \
    {x.Data4[0], x.Data4[1], x.Data4[2], x.Data4[3], \
    x.Data4[4], x.Data4[5], x.Data4[6], x.Data4[7]}}
#endif

#define FLAT_GUID_NULL\
    0x00000000, 0x0000, 0x0000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0
#define FLAT_CTLOCSTR_NULL  FLAT_GUID_NULL
#define FLAT_CTLOCUNIT_NULL FLAT_GUID_NULL

// This verifies that HRESULT and CTSTATUS are binary compatible
#if (CT_HRESULT_MAXVAL != CT_STATUS_MAXVAL)
#  error CtComDef.H: Mismatch between HRESULT and CTSTATUS sizes! See CtDef.h.
#endif

/* Use the COMCALL macro when calling COM methods from C code */
/* Usage: COMCALL(IFoo, pIFoo, foofunc)(pIFoo, param1, param2); */
#if !defined(__cplusplus)
#  if defined(__CT_SYS_LINUX)
#    define COMCALL(intf,ptr,fn) ((struct intf##Vtbl *)(((PCTVOID *) ptr->lpVtbl)))->fn
#  elif (CT_IS_SOLARIS) || (CT_IS_OSX)
#    define COMCALL(intf,ptr,fn) ((struct intf##Vtbl *)(((PCTVOID *) ptr->lpVtbl) + 2))->fn
#  elif defined(__CT_SYS_MACINTOSH)
#    define COMCALL(intf,ptr,fn) ((struct intf##Vtbl *)(((PCTVOID *) ptr->lpVtbl) + 2))->fn
#  else
#    define COMCALL(intf,ptr,fn) ptr->lpVtbl->fn
#  endif
#else
#  define COMCALL(intf,ptr,fn) ptr -> fn
#endif /* __cplusplus */

/* Some stuff specific to the Linux platform */
#ifdef __CT_SYS_LINUX

#define CLASS_E_NOAGGREGATION      0x80040110L
#define CLASS_E_CLASSNOTAVAILABLE  0x80040111L

#ifdef __KERNEL__
CTCTYPE CTLONG InterlockedIncrement(CTLONG *plRef);
CTCTYPE CTLONG InterlockedDecrement(CTLONG *plRef);

#else /* __KERNEL__ */
/* This should really be protected by pthread mutexes */
static inline CTLONG InterlockedIncrement(CTLONG *plRef) { return ++(*plRef); }
static inline CTLONG InterlockedDecrement(CTLONG *plRef) { return --(*plRef); }

HRESULT CoCreateInstance(REFCLSID rclsid, IUnknown *pUnkOuter, CTDWORD dwClsContext, REFIID riid, PCTVOID *ppv);
HRESULT CoInitialize(PCTVOID pvReserved);
void CoUninitialize(void);
#endif /* __KERNEL__ */

#endif /* __CT_SYS_LINUX */

#ifdef __CT_SYS_MACINTOSH
// Some stuff specific to the Macintosh platform

// IncrementAtomic et al perform the requested operation
// but then return the value BEFORE the operation.  This is
// not what  is expected. So we do this the following way:
static inline CTLONG InterlockedIncrement(CTLONG *plRef)	{ IncrementAtomic(plRef); return *plRef ; }
static inline CTLONG InterlockedDecrement(CTLONG *plRef)	{ DecrementAtomic(plRef); return *plRef ; }

#define CLASS_E_NOAGGREGATION      0x80040110L
#define CLASS_E_CLASSNOTAVAILABLE  0x80040111L

#endif //__CT_SYS_MACINTOSH


#endif /* __CTCOMDEF_H */
