/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
UDA uses COM as a standard way to define module interfaces. Since COM is
a Windows standard, it has no equivalents in non-Windows platforms. Also,
COM is only loosely supported in some Windows kernel environments. For
this reason, UDA uses this header file to define a common COM error code
value series that is guaranteed to be consistent on all platforms.

According to COM standards, all error codes are 4-byte signed values. The
value 0 indicates successs. A positive value indicates success but with information
that might be important to the caller. A negative value indicates a failure
with information that might be important to the caller.

$Date: 2007/12/26 08:49:05 $
$Revision: 1.7.56.2 $ 
Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#ifndef __CTCOMERR_H
#define __CTCOMERR_H

/** Success */
#define S_OK                0x00000000

/** Operation not performed, as expected */
#define S_FALSE             0x00000001

/** Catastrophic failure */
#define E_UNEXPECTED        0x8000FFFF

/** Feature not implemented */
#define E_NOTIMPL           0x80004001

/** Ran out of memory */
#define E_OUTOFMEMORY       0x8007000E

/** One or more arguments are invalid */
#define E_INVALIDARG        0x80070057

/** No such interface is supported */
#define E_NOINTERFACE       0x80004002

/** Invalid pointer */
#define E_POINTER           0x80004003

/** Invalid handle */
#define E_HANDLE            0x80070006

/** Operation aborted */
#define E_ABORT             0x80004004

/** Unspecified error */
#define E_FAIL              0x80004005

/** General access denied error */
#define E_ACCESSDENIED      0x80070005

/** Routine cannot run to completion because
 * it cannot run asynchronously (i.e., it
 * can't be called from an interrupt handler,
 * or while the IRQL is not PASSIVE).
 */
#define E_ASYNC             0x80070009

/** Standard macro that is true for all possible successful COM error codes.
 *  Note there is a reason this MACRO is not defined as (_x == 0)!
 */
#define SUCCEEDED(_x) ((HRESULT) (_x) >= 0)

/** Standard macro that is true for all possible failure COM error codes.
 */
#define FAILED(_x) ((HRESULT) (_x) < 0)

#endif /* __CTCOMERR_H */

