/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
This header file is what allows the UDA AS and HA level software
to compile properly under different OS build environments. It contains 
UDA specific overloaded datatypes which are guaranteed to have consistent
behavior and characteristics on any supported build platform. It also
performs a few checks to make sure the build environment is one that has
been proven to be compatible with UDA.

$Date: 2008/03/12 09:52:28 $
$Revision: 1.147.18.8 $

Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#ifndef __CTDEF_H
#define __CTDEF_H

/******************************************************************************
* PreCheck: This provides protection against including this with older
*           versions of the same file that may have different definitions.
*******************************************************************************/

#ifdef __DATATYPE_H
#  error Ctdef.H: Attempt to use with legacy Datatype.H (old E-mu S/W)
#endif

#ifdef _COMMON_H_
#  error Ctdef.H: Attempt to use with legacy Common.H or Haldefs.H (old CTL S/W)
#endif

#ifdef CTCONFIG_H
  /** PreCheck: If for some reason the build environment precludes using 
   *            numerous command-line compile flags in your, the project settings
   *            should include this define that points to a header file
   *            with the rest of the compile flags in it. We assume that the compiler can
   *            handle at least one define as a parameter. That define in this case
   *            should be called CTCONFIG_H and it should define the relative path
   *            to a specific header file. Here is where that gets included.
   */
#  include CTCONFIG_H
#endif

/*****************************************
* Ctdef specific compile error checking
*****************************************/
#ifdef __CT_SYS_OSX_PPC
    #include "/usr/include/limits.h"    /* This is where limits checking values are */

    #include <machine/ansi.h>
    #ifndef NULL
        #define NULL    0
    #endif

#else
    #include <linux/limits.h> 	/* This is where limits checking values are */
#endif

#ifndef NULL
  #ifdef __CT_SYS_LINUX
    #include "ctstddef.h"
  #else
    #include <stddef.h>  /* This is where NULL is defined */
  #endif
#endif

/********************************************************************
* 1. Check system dependent flags
*
* Ctdef.H compensates for system differences w/r/t endian nature, memory
* address sizes and the system itself. Compile flags allow automatic switches
* within the codebase to account for differences between systems in these areas.
* This checks to make sure Ctdef is compiled with the proper compile
* flags set. Unfortunately, there is no universal way to have a compiler
* display a custom warning, so we must trigger an error if the
* proper compile flags are not set.
********************************************************************/

/*  For Aptos systems, make all unions unaligned in hopes of catching
    all bit fields and making them unaligned in order to overcome a
    Metaware compiler bug.                                          */
/*
#if   defined (__CT_SYS_APTOS)
#   define union _Unaligned union
#endif
*/
/* Endian nature considerations */
#if   !defined (__CT_BIG_ENDIAN)         &&  !defined (__CT_LITTLE_ENDIAN)
#  if !defined (__CT_ENDIAN_INDEPENDENT)
/*   If you get an error here, set one of the above defines as a global
     compile flag. These flags indicate ENDIAN NATURE of your system.   */
#    error Ctdef.H: No Endian Nature flag assigned
#  endif
#endif

/* System memory boundary considerations */
#if   !defined (__CT_BOUND_16BIT) && !defined (__CT_BOUND_32BIT)
#  if !defined (__CT_BOUND_64BIT)
  /* If you get an error here, set one of the above defines as a global compile
     flag. These flags indicate the MEMORY BOUNDARY of your system. */
#    error Ctdef.H: No Memory boundary flag assigned
#  endif
#endif

/* System type considerations */
#if       !defined (__CT_SYS_GENERAL) && !defined (__CT_SYS_WINDOWS_95) && !defined(__CT_SYS_APTOS)
#  if      !defined (__CT_SYS_MSDOS)   && !defined (__CT_SYS_MACINTOSH)
#    if     !defined (__CT_SYS_LINUX)  && !defined (__CT_SYS_TMS320C32)
#      if   !defined (__CT_SYS_SGI)    && !defined (__CT_SYS_WINDOWS_NT_IX86)
#        if !defined (__CT_SYS_WINDOWS_NT_AXP64) && !defined(__CT_SYS_WINDOWS_NT_IA64) && !defined(__CT_SYS_WINDOWS_NT_AMD64)
#          if !defined(__CT_SYS_SOLARIS_X86) && !defined(__CT_SYS_SOLARIS_SPARC)
#        if !defined(__CT_SYS_OSX_PPC) && !defined(__CT_SYS_OSX_X86)
               /* If you get an error here, set one of the above defines as a
                  global compile flag. These flags SYSTEM SPECIFIC considerations
                of your system. */
#              error Ctdef.H: No System type flag assigned
#            endif
#          endif
#        endif
#      endif
#    endif
# endif
#endif

/** For characteristics that won't change based on flavor of NT
   (WinNT 4, Win 2k, WDM) */
#if !defined(__CT_SYS_LINUX)
#define CT_IS_WINDOWS_NT    (defined (__CT_SYS_WINDOWS_NT_IX86) || defined (__CT_SYS_WINDOWS_NT_AXP64) || defined (__CT_SYS_WINDOWS_NT_IA64) || defined (__CT_SYS_WINDOWS_NT_AMD64))
#endif


/** For characteristics that won't change based on 32 bit Windows
   operating systems running on Intel platforms.
   (Win95, Win98, WinNT/WDM on Intel) */
#define CT_IS_WINDOWS_32    (defined (__CT_SYS_WINDOWS_95) || defined (__CT_SYS_WINDOWS_NT_IX86))

/** For characteristics that won't change based on 64 bit Windows
   operating systems. */
#define CT_IS_WINDOWS_64    (defined (__CT_SYS_WINDOWS_NT_AXP64) || defined (__CT_SYS_WINDOWS_NT_IA64) || defined (__CT_SYS_WINDOWS_NT_AMD64))

/** For characteristics that won't change based on using Microsoft Windows.
   BE CAREFUL BEFORE USING THIS FLAG! */
#if !defined(__CT_SYS_LINUX)
#define CT_IS_WINDOWS       (CT_IS_WINDOWS_NT || CT_IS_WINDOWS_32 || CT_IS_WINDOWS_64)
#endif

/** For characteristics that won't change based on using LINUX.
   BE CAREFUL BEFORE USING THIS FLAG! */
#define CT_IS_LINUX         (defined (__CT_SYS_LINUX) || defined (__CT_SYS_LINUX_X86) || defined (__CT_SYS_LINUX_AMD64))

/** For characteristics that won't change based on using Sun Solaris.
   BE CAREFUL BEFORE USING THIS FLAG! */
#define CT_IS_SOLARIS       (defined (__CT_SYS_SOLARIS_SPARC) || defined (__CT_SYS_SOLARIS_X86))

/** For characteristics that won't change based on using Apple OS X.
   BE CAREFUL BEFORE USING THIS FLAG! */
#define CT_IS_OSX           (defined (__CT_SYS_OSX_PPC) || defined (__CT_SYS_OSX_X86))

/** For characteristics that won't changed based on x86 platforms */
#define CT_IS_PLATFORM_X86  (defined (__CT_SYS_WINDOWS_95) || defined (__CT_SYS_MSDOS) || defined (__CT_SYS_LINUX) || defined (__CT_SYS_LINUX_X86) || defined (__CT_SYS_WINDOWS_NT_IX86) || defined (__CT_SYS_SOLARIS_X86) || defined (__CT_SYS_OSX_X86))

/** For characteristics that won't changed based on ALPHA platforms */
#define CT_IS_PLATFORM_AXP64 (defined (__CT_SYS_WINDOWS_NT_AXP64))

/** For characteristics that won't changed based on IA64 platforms */
#define CT_IS_PLATFORM_IA64 (defined (__CT_SYS_WINDOWS_NT_IA64))

/** For characteristics that won't changed based on AMD64 platforms */
#define CT_IS_PLATFORM_AMD64 (defined (__CT_SYS_WINDOWS_NT_AMD64) || defined (__CT_SYS_LINUX_AMD64))

/** For characteristics that won't changed based on SPARC platforms */
#define CT_IS_PLATFORM_SPARC (defined (__CT_SYS_SOLARIS_SPARC))

/** For characteristics that won't changed based on PPC platforms */
#define CT_IS_PLATFORM_PPC (defined (__CT_SYS_OSX_PPC))

/* Speed this bad boy up */
#if CT_IS_WINDOWS && !defined(CT_CORECODE)
#  ifndef WIN32_LEAN_AND_MEAN
#    define WIN32_LEAN_AND_MEAN
#  endif
#endif

/********************************************************************
* 2. Check basic datatype limits
*
* Compare <limits.h> values with CTDEF limits.  Fail compile
* if discrepancy. If compile fails due to one of these errors, then
* the UDA software will not work with your system or the state of your
* development environment.
********************************************************************/

#ifdef __CT_SYS_TMS320C32
#  define CTDEF_TEST_32BIT_ONLY
#endif

#if CT_IS_SOLARIS
#  if defined (__CT_BOUND_64BIT)
#    define CTDEF_TEST_SHORT_ONLY
#  endif
#endif

/* Expected data values */
#define CT_CHAR_MINVAL  -128
#define CT_CHAR_MAXVAL   127
#define CT_BYTE_MAXVAL   255
#define CT_SHRT_MINVAL  -32768
#define CT_SHRT_MAXVAL   32767
#ifndef __x86_64__
#  define CT_LONG_MINVAL  (-2147483647L - 1)
#  define CT_LONG_MAXVAL   2147483647L
#else // __x86_64__ is defined
#  define CT_LONG_MINVAL  (-9223372036854775807L - 1L)
#  define CT_LONG_MAXVAL   9223372036854775807L
#endif
#define CT_DWORD_MAXVAL  4294967295UL

#if !defined (__CT_SYS_LINUX)
#ifndef CTDEF_TEST_32BIT_ONLY
#  if (SCHAR_MAX != CT_CHAR_MAXVAL) || (UCHAR_MAX != CT_BYTE_MAXVAL)
#   error Ctdef.H: char is out of range! Expected 8 bit.
#  endif

#  if (SHRT_MAX != CT_SHRT_MAXVAL)
#   error Ctdef.H: short is out of range! Expected 16 bit.
#  endif
#endif

#ifndef CTDEF_TEST_SHORT_ONLY
#  if (LONG_MAX != CT_LONG_MAXVAL)
#   ifndef __x86_64__
#    error Ctdef.H: long is out of range! Expected 32 bit.
#   else
#    error Ctdef.H: long is out of range! Expected 64 bit.
#   endif
#  endif
#endif
#endif // !defined(__CT_SYS_LINUX)

/*
   If you are developing in C, please add this to the prototypes of all API
   calls so that folks who use C++ may be able to use your code without
   writing their code any special way.

   If you are developing in C++, please add this to the prototypes of
   all classic C function type API calls so that folks who use C code may
   be albe to use your code without recompiling your code.

   IE:
   CTCTYPE void objApiCall(USHORT val1, USHORT val2);
   CTCTYPE WORD * objGetPtr(USHORT val);
   etc.

   Or you could use the "series" macros to define a whole list of functions
   in one shot.

   IE:
   CTBEGINCTYPE
   void objApiCal(USHORT val1, USHORT val2);
   WORD * objGetPtr(USHORT val);
   CTENDEMUCTYPE

   These macros act as compile time switchable 'extern "C"' commands
   based on C or C++ compiler.
*/

#ifdef __cplusplus
#  define CTCTYPE       extern "C"
#  define CTBEGINCTYPE  CTCTYPE {
#  define CTENDCTYPE    }
#else
#  define CTCTYPE
#  define CTBEGINCTYPE
#  define CTENDCTYPE
#endif


/********************************************************************
* 3. Include operating system specific global header files
*
* If the specific OS relies on certain header files always being
* included or on specific ordering of standard header files, those
* go here.
********************************************************************/

#if defined (CT_CORECODE)
#  if defined (__CT_WIN32_DRIVER_MODEL)
#     ifdef __cplusplus
#       pragma warning (disable: 4786)
#       include <portcls.h>
#       include "newdelete.h"
#       include <stdunk.h>
#       include <stdarg.h>
#       include <malloc.h>
#     else
#       include <wdm.h>
#       include <windef.h>
#       include <punknown.h>
#       include <stdio.h>
#       include <stdlib.h>
#       include <stdarg.h>
#       include <malloc.h>
#     endif
#  elif (CT_IS_WINDOWS_NT)
#     if defined(_USRDLL) || defined(_USRKOD)
#       ifndef _USE_SDK_INCLUDES
#         include <portcls.h>
#         include <stdarg.h>
#         include <malloc.h>
#       else
#         include <windows.h>
#         include <mmsystem.h>
#         include <stdarg.h>
#         include <malloc.h>
#         include <objbase.h>
#         include <guiddef.h>
          EXTERN_C const GUID GUID_NULL;
#         define _FILETIME_
#         define _SECURITY_ATTRIBUTES_
#         define _OVERLAPPED_
#       endif
#     else
#       ifdef __cplusplus
#         include <portcls.h>
#         include "newdelete.h"
#         include <stdunk.h>
#         include <stdarg.h>
#         include <malloc.h>
#       else
#         include <ntddk.h>
#         include <windef.h>
#         define RPC_NO_WINDOWS_H
          typedef PVOID PALETTEENTRY;
#         include <wtypes.h>
#         undef RPC_NO_WINDOWS_H
#         include <basetyps.h>
#         include <guiddef.h>
#         include <punknown.h>
#         include <stdio.h>
#         include <stdlib.h>
#         include <stdarg.h>
#         include <malloc.h>
#       endif
#     endif
#  elif defined(__CT_SYS_LINUX)
#    include "cttypes.h"
#    include <stdarg.h>
#    ifdef __KERNEL__
#      include <linux/compiler.h>
#      include <linux/string.h>
#      define _alloca __builtin_alloca
#      ifndef _NOTHREADS
#        define _NOTHREADS
#      endif
#      ifdef inline
#        undef inline
#      endif
#    else
#      include <string.h>
#    endif /* __KERNEL__ */
#    ifndef HUGE_VAL
#       if defined(__GNUC__) && (__GNUC__ >= 3)
#           define HUGE_VAL    (__extension__ 0x1.0p2047)
#       else
            static union { 
                unsigned char __c[8];
                double __d;
            } __huge_val = { 0, 0, 0, 0, 0, 0, 0xf0, 0x7f };
#           define HUGE_VAL    (__huge_val.__d)
#       endif
#   endif
        
CTCTYPE unsigned long int strtoul(const char *nptr, char **endptr, int base);

#  elif (CT_IS_SOLARIS)
#    include <sys/types.h>
#    include <stdarg.h>
#    include <string.h>
#    define __inline inline
#    ifdef _KERNEL
#      include <alloca.h>
#      define _alloca alloca
#    endif  /* _KERNEL */
#  elif defined(__CT_SYS_MACINTOSH)
#    include "Timer.h"
#    include <stdlib.h>
#    include <stdio.h>
#    include <stdarg.h>
#    include <string.h>
#    include <Devices.h>
CTCTYPE void * alloca(size_t);
#    define _alloca alloca
#  elif (CT_IS_OSX)
#    include "/usr/include/stdlib.h"
#    include <stdarg.h>
#    include <string.h>
CTCTYPE void * alloca(size_t);
#    define _alloca alloca

#  else
     /*
      * We need to define the macros which transform
      * the various run-time routines into the proper
      * segment versions.  We're assuming that the locked
      * versions of the intrinsics should always be safe to
      * call; this might not be true for non-reentrant
      * functions (like strtok) so be careful when you
      * add new definitions here.
      */
#    if defined(__CT_SYS_WINDOWS_95) && defined(VTOOLSD)
#      ifdef __cplusplus
#          pragma warning( disable: 4075)
#          pragma init_seg( "ICRTTEXT" )
#          pragma warning( default: 4075 )
#          pragma warning( disable: 4786 )
#      endif
#      define strncpy    LOCK_strncpy
#      define strtoul    LOCK_strtoul
#      define strrchr    LOCK_strrchr
#    endif

     /* Bring in the standard, generic functions */
#    include <stdio.h>
#    include <stdlib.h>
#    include <string.h>
#    include <stdarg.h>

      /* Don't be using malloc and free; we only want _alloca */
#    include <malloc.h>
#  endif  /* General CT_CORECODE case */
#endif  /* ifdef CT_CORECODE */

#ifndef CT_CORECODE
#  if CT_IS_WINDOWS
#    if !defined (VTOOLSD)
#      if defined (__CT_WIN32_DRIVER_MODEL)
#        ifdef __cplusplus
#          include <portcls.h>
#          include "newdelete.h"
#          include <stdunk.h>
#          include <stdarg.h>
#        else
#          include <wdm.h>
#          include <windef.h>
#          include <punknown.h>
#          include <stdio.h>
#          include <stdlib.h>
#          include <stdarg.h>
#        endif
#      elif (CT_IS_WINDOWS_NT)
#        if defined(_USRDLL) || defined(_USRKOD)
#          include <windows.h>
#          include <mmsystem.h>
#          include <stdarg.h>
#          include <malloc.h>
#          include <objbase.h>
#          include <guiddef.h>
           EXTERN_C const GUID GUID_NULL;
#          define _FILETIME_
#          define _SECURITY_ATTRIBUTES_
#          define _OVERLAPPED_
#        else
#          ifdef __cplusplus
#            include <portcls.h>
#            include "newdelete.h"
#            include <stdunk.h>
#            include <stdarg.h>
#          else
#            include <ntddk.h>
#            include <windef.h>
#            define RPC_NO_WINDOWS_H
             typedef PVOID PALETTEENTRY;
#            include <wtypes.h>
#            undef RPC_NO_WINDOWS_H
#            include <basetyps.h>
#            include <guiddef.h>
#            include <punknown.h>
#            include <stdio.h>
#            include <stdlib.h>
#            include <stdarg.h>
#          endif
#        endif // _USRDLL || _USRKOD
#      else
#         include <windows.h>
#         include <malloc.h>
#      endif
#    else
#      define _far
#      include <string.h>
#      ifdef __cplusplus
#        include <vtoolscp.h>
#      else
#        include <vtoolsc.h>
#      endif
#    endif
#  endif
#endif

#if defined (__CT_SYS_LINUX) && !defined (CT_CORECODE)
#include "cttypes.h"
#include <stdarg.h>
#ifdef __KERNEL__
#  include <linux/compiler.h>
#  include <linux/string.h>
#  ifndef _NOTHREADS
#    define _NOTHREADS
#  endif
#  ifdef inline
#    undef inline
#  endif
#else
# include <string.h>
# define _alloca __builtin_alloca
#endif
#endif

#if (CT_IS_SOLARIS) && !defined (CT_CORECODE)
#  include <sys/types.h>
#  include <stdarg.h>
#  include <string.h>
#  define __inline inline
#  include <alloca.h>
#  define _alloca alloca
#endif

#if defined (__CT_SYS_MSDOS) && !defined (CT_CORECODE)
#  include <dos.h>
#  include <stdio.h>
#  include <stdlib.h>
#  include <string.h>
#endif

#if defined (__CT_SYS_MACINTOSH) && !defined (CT_CORECODE)
#  include "FixMath.h"
#  include "MacMemory.h"
#  include "Timer.h"
#  include <stdio.h>
#  include <stdlib.h>
#  include <stdarg.h>
#  include <string.h>
//#  include <malloc.h>
CTCTYPE void * alloca(size_t);
#  define _alloca alloca
#endif

#if (CT_IS_OSX) && !defined(CT_CORECODE)
#  include <stdarg.h>
#  include <string.h>

CTCTYPE void * alloca(size_t);
#    define _alloca alloca
#endif

#define CT_LONG_DOUBLES_UNSUPPORTED defined (__CT_SYS_SGI) || defined (CT_IS_OSX)

/********************************************************************
* 4. OK, now onto the datatypes and defines
*
* Here is where UDA general purpose typedefs are declared. Unless
* otherwise noted, all UDA datatypes have the same size and format
* regardless of the platform.
********************************************************************/

/*************
* Typedefs
*************/

/**  8 bit signed value     */
typedef char                CTCHAR,     *PCTCHAR,       **PPCTCHAR;
/**  8 bit unsigned value   */
typedef unsigned char       CTUCHAR,    *PCTUCHAR,      **PPCTUCHAR;
/** 16 bit signed value     */
typedef short               CTSHORT,    *PCTSHORT,      **PPCTSHORT;
/** 16 bit unsigned value   */
typedef unsigned short      CTUSHORT,   *PCTUSHORT,     **PPCTUSHORT;
typedef unsigned short      CTWCHAR,    *PCTWCHAR,      **PPCTWCHAR;
/** 32 bit signed value     */
#if defined(__CT_SYS_LINUX)
#	ifdef __x86_64__
		typedef int	CTLONG,	*PCTLONG,	**PPCTLONG;
#else
		typedef long	CTLONG,	*PCTLONG,	**PPCTLONG;
#endif
#elif defined (__CT_SYS_SOLARIS_SPARC) && defined (__CT_BOUND_64BIT)
typedef int                 CTLONG,     *PCTLONG,       **PPCTLONG;
#else
typedef long                CTLONG, *PCTLONG,       **PPCTLONG;
#endif
/** 32 bit unsigned value   */
#if defined(__CT_SYS_LINUX)
#	ifdef __x86_64__
		typedef unsigned int        CTULONG,    *PCTULONG,      **PPCTULONG;
#	else
		typedef unsigned long       CTULONG,    *PCTULONG,      **PPCTULONG;
#	endif
#else
typedef unsigned long       CTULONG,    *PCTULONG,      **PPCTULONG;
#endif
/** 32 bit floating point value */
typedef float               CTFLOAT,    *PCTFLOAT,      **PPCTFLOAT;
/** 64 bit floating point value */
typedef double              CTDOUBLE,   *PCTDOUBLE,     **PPCTDOUBLE;

/** 64 bit signed and unsigned values */
#if (CT_IS_WINDOWS) || (defined(__CT_SYS_GENERAL) || defined(_MSC_VER))
#  define CT_LONG_LONG_SUPPORTED
#  define CT_LONGLONG_MAXVAL    0x7FFFFFFFFFFFFFFF
#  define CT_ULONGLONG_MAXVAL   0xFFFFFFFFFFFFFFFF

  typedef unsigned __int64    CTULONGLONG, *PCTULONGLONG,  **PPCTULONGLONG;
  typedef __int64 CTLONGLONG, *PCTLONGLONG, **PPCTLONGLONG;
#else

#  if defined(__CT_SYS_LINUX) || defined(__CT_SYS_OSX_PPC) || defined(__CT_SYS_APTOS)
#    define CT_LONG_LONG_SUPPORTED
#    define CT_LONGLONG_MAXVAL    0x7FFFFFFFFFFFFFFF
#    define CT_ULONGLONG_MAXVAL   0xFFFFFFFFFFFFFFFFULL
#	ifdef __x86_64__
     		typedef unsigned long CTULONGLONG, *PCTULONGLONG, **PPCTULONGLONG;
		typedef long CTLONGLONG, *PCTLONGLONG, **PPCTLONGLONG;
#	else
     		typedef unsigned long long CTULONGLONG, *PCTULONGLONG, **PPCTULONGLONG;
		typedef long long CTLONGLONG, *PCTLONGLONG, **PPCTLONGLONG;
#	endif
#  endif
#endif

/** Parameter used for defining the datatype required of a malloc  */
#if (CT_IS_SOLARIS)
#  if defined (__CT_BOUND_64BIT)
     typedef unsigned long      CTSIZE_T;
#  else
     typedef unsigned int       CTSIZE_T;
#  endif
#else
   typedef size_t               CTSIZE_T;
#endif

#if !(CT_LONG_DOUBLES_UNSUPPORTED)
/** 80 bit floating point value */
  typedef long double       CTLDOUBLE,  *PCTLDOUBLE,    **PPCTLDOUBLE;
#else
#  define  CTLDOUBLE    See_CTDef.h
#  define  PCTLDOUBLE   See_CTDef.h
#  define  PPCTLDOUBLE  See_CTDef.h
#endif

#define CT_STATUS_MINVAL  CT_LONG_MINVAL
#define CT_STATUS_MAXVAL  CT_LONG_MAXVAL
#define CTSTATUS CTLONG

/* Alternate definitions */

/* Variation 1: Unsigned preferred */
#define CTBYTE   CTUCHAR
#define CTSBYTE  CTCHAR
#define CTWORD   CTUSHORT
#define CTSWORD  CTSHORT
#define CTDWORD  CTULONG
#define CTSDWORD CTLONG
#define CTQWORD  CTULONGLONG
#define CTSQWORD CTLONGLONG

#define PCTBYTE   PCTUCHAR
#define PCTSBYTE  PCTCHAR
#define PCTWORD   PCTUSHORT
#define PCTSWORD  PCTSHORT
#define PCTDWORD  PCTULONG
#define PCTSDWORD PCTLONG
#define PCTQWORD  PCTULONGLONG
#define PCTSQWORD PCTLONGLONG

#define PPCTBYTE   PPCTUCHAR
#define PPCTSBYTE  PPCTCHAR
#define PPCTWORD   PPCTUSHORT
#define PPCTSWORD  PPCTSHORT
#define PPCTDWORD  PPCTULONG
#define PPCTSDWORD PPCTLONG
#define PPCTQWORD  PPCTULONGLONG
#define PPCTSQWORD PPCTLONGLONG

/* Variation 2: Bitwidth preferred */
#define CTBYTE   CTUCHAR
#define CTS8     CTCHAR
#define CTU8     CTUCHAR
#define CTS16    CTSHORT
#define CTU16    CTUSHORT
#define CTS32    CTLONG
#define CTU32    CTULONG
#define CTF32    CTFLOAT
#define CTF64    CTDOUBLE
#define CTF80    CTLDOUBLE
#define CTS64    CTLONGLONG
#define CTU64    CTULONGLONG

#define PCTBYTE   PCTUCHAR
#define PCTS8     PCTCHAR
#define PCTU8     PCTUCHAR
#define PCTS16    PCTSHORT
#define PCTU16    PCTUSHORT
#define PCTS32    PCTLONG
#define PCTU32    PCTULONG
#define PCTF32    PCTFLOAT
#define PCTF64    PCTDOUBLE
#define PCTF80    PCTLDOUBLE
#define PCTU64    PCTULONGLONG
#define PCTS64    PCTLONGLONG

#define PPCTBYTE   PPCTUCHAR
#define PPCTS8     PPCTCHAR
#define PPCTU8     PPCTUCHAR
#define PPCTS16    PPCTSHORT
#define PPCTU16    PPCTUSHORT
#define PPCTS32    PPCTLONG
#define PPCTU32    PPCTULONG
#define PPCTF32    PPCTFLOAT
#define PPCTF64    PPCTDOUBLE
#define PPCTF80    PPCTLDOUBLE
#define PPCTU64    PPCTULONGLONG
#define PPCTS64    PPCTLONGLONG

/** CTPTRSIZE is defined as being an unsigned integral type which
 *  is large enough to contain a pointer value without truncation.
 *  We use it in places where we need to do bit manipulation on
 *  pointers in a system-independent way.
 */
#if  defined(__CT_BOUND_16BIT)
   typedef CTWORD CTPTRSIZE;
#elif defined(__CT_BOUND_32BIT)
   typedef CTDWORD CTPTRSIZE;
#elif defined(__CT_BOUND_64BIT)
#  if defined(CT_IS_WINDOWS_NT)
     typedef unsigned __int64 CTPTRSIZE;
#  elif defined(__CT_SYS_SOLARIS_SPARC)
     typedef unsigned long CTPTRSIZE;
#  elif defined(__CT_SYS_LINUX)
     typedef unsigned long CTPTRSIZE;
#  else
#    error "I don't know what type to use for 64 bit integral values!"
#  endif
#else
#  error What the hell kind of ptr size do you have?
#endif

/** CTLARGE is defined as the largest signed integral datatype
 *  that a particular platform is capable of supporting.  It
 *  is used in cases where we're afraid that a CTDWORD is
 *  too small (in particular, we use it for sample counters
 *  in the HA wave code).  Size is more important that efficiency
 *  when using a CTLARGE, so if a platform supports a long long
 *  datatype in software it would be appropriate to define
 *  CTLARGE as a long long.
 */
typedef long          CTLARGE;
typedef unsigned long CTULARGE;


/* The CTQWORD and CTLONGLONG */

   /********************************************************************
   * These datatypes are system and memory word size dependent, so they
   * do not belong in any data structures that must be transferred
   * outside of a binary to an unknown operating system or file system.
   * Don't use in file formats or in COM object definitions
   ********************************************************************/

/** Opaque flag can be either TRUE or FALSE */
typedef int                 CTBOOL,     *PCTBOOL,       **PPCTBOOL;


/** Opaque handle must be the size of a memory address */
#ifdef __CT_BOUND_16BIT
   /* Insure that 16 bit code can hold a 32 bit (large) pointer */
   typedef CTULONG              CTHANDLE;
#  define CTHANDLE_NULL     0
#else
#  if (CT_IS_WINDOWS_NT)
     typedef void *              CTHANDLE;
#    define CTHANDLE_NULL     NULL
#  elif defined(__CT_SYS_LINUX)
     typedef void *       CTHANDLE;
#    define CTHANDLE_NULL   NULL
#  elif (CT_IS_SOLARIS)
#    if defined (__CT_BOUND_64BIT)
     typedef unsigned long       CTHANDLE;
#    else
     typedef int                 CTHANDLE;
#    endif
#    define CTHANDLE_NULL     NULL
#  elif defined (__CT_SYS_MACINTOSH)
     typedef int                 CTHANDLE;
#    define CTHANDLE_NULL     0
#  else
     typedef void *                CTHANDLE;
#    define CTHANDLE_NULL   0
#  endif
#endif

/** Opaque handle must be the size of a memory address */
typedef CTHANDLE            *PCTHANDLE, **PPCTHANDLE;

/** Void datatype, for compatibility with COM interfaces */
typedef void                CTVOID,     *PCTVOID,       **PPCTVOID;

/** Datatypes required for CT STL stuff */
#ifdef __CT_SYS_MACINTOSH
typedef CTLONG ptrdiff_t;
typedef CTULONG result_type;
typedef CTULONG argument_type;
typedef CTULONG first_argument_type;
typedef CTULONG second_argument_type;
#endif

/******************************
 * Defines and other switches
 *****************************/

/* Take care of typically used macros, if not defined under
   the various operating systems */

#define CTSTATUS_SUCCESS 0

#ifndef FALSE
#  define FALSE  (CTBOOL)0
#endif

#ifndef TRUE
#  define TRUE   (CTBOOL)1
#endif

#ifndef NULL
#  define NULL   (0)
#endif

#ifndef LOBYTE
#  define LOBYTE(_x)   (CTUCHAR)((CTUSHORT)(_x) & 0x00FF)
#endif

#ifndef HIBYTE
#  define HIBYTE(_x)   (CTUCHAR)(((CTUSHORT)(_x) & 0xFF00) >> 8)
#endif

#ifndef LOWORD
#  define LOWORD(_x)   (CTUSHORT)((CTULONG)(_x) & 0x0000FFFFL)
#endif

#ifndef HIWORD
#  define HIWORD(_x)   (CTUSHORT)(((CTULONG)(_x) & 0xFFFF0000L) >> 16)
#endif

#ifndef MAX_PATH
#  define MAX_PATH 1024
#endif

#if CT_IS_WINDOWS 
#  if _MSC_VER < 1300 
#    define CTNOOP ((void)0) 
#  else 
#    define CTNOOP __noop 
#  endif // _MSC_VER 
#elif defined(__CT_SYS_LINUX)
#  define CTNOOP
#else 
#  define CTNOOP ((void)0) 
#endif 

#ifdef __CT_BOUND_64BIT
#  if CT_IS_PLATFORM_AMD64
#    define CT_MEMORYPAGESIZE   4096
#    define CT_MEMORYPAGESHIFT  12
#  else
#    define CT_MEMORYPAGESIZE   8192
#    define CT_MEMORYPAGESHIFT  13
#  endif
#else
#  define CT_MEMORYPAGESIZE     4096
#  define CT_MEMORYPAGESHIFT    12
#endif

#if defined(UNICODE)
typedef CTWCHAR CTPATH,     *PCTPATH,   **PPCTPATH;
typedef CTWCHAR CTTCHAR,    *PCTTCHAR,  **PPCTTCHAR;
typedef CTWCHAR CTUTCHAR,   *PCTUTCHAR, **PPCTUTCHAR;
#   ifndef TEXT
#   define TEXT(quote) L##quote
#   endif
#   define STRLEN   wcslen
#   define STRCMP   wcscmp
#   define STRNCMP  wcsncmp
#   define STRICMP  _wcsicmp
#   define STRCPY   wcscpy
#   define STRNCPY  wcsncpy
#   define STRCAT   wcscat
#   define STRTOK   wcstok
#   define STRRCHR  wcsrchr
#   define FOPEN	_wfopen
#	define FPRINTF  fwprintf
#   define FREOPEN	_wfreopen
#	define FSCANF   fwscanf
#	define PRINTF   wprintf
#	define SCANF    wscanf
#	define SPRINTF  swprintf
#	define SSCANF   swscanf
#	define VFPRINTF vfwprintf
#	define VPRINTF  vwprintf
#	define VSPRINTF vswprintf
#else
typedef CTCHAR  CTPATH,     *PCTPATH,   **PPCTPATH;
typedef CTCHAR  CTTCHAR,    *PCTTCHAR,  **PPCTTCHAR;
typedef CTUCHAR CTUTCHAR,   *PCTUTCHAR, **PPCTUTCHAR;
#   ifndef TEXT
#   define TEXT(quote) quote
#   endif
#   define STRLEN   strlen
#   define STRCMP   strcmp
#   define STRNCMP  strncmp
#   define STRICMP  _stricmp
#   define STRCPY   strcpy
#   define STRNCPY  strncpy
#   define STRCAT   strcat
#   define STRTOK   lstrtok
#   define STRRCHR  strrchr
#   define FOPEN	fopen
#	define FPRINTF  fprintf
#   define FREOPEN	freopen
#	define FSCANF   fscanf
#	define PRINTF   printf
#	define SCANF    scanf
#	define SPRINTF  sprintf
#	define SSCANF   sscanf
#	define VFPRINTF vfprintf
#	define VPRINTF  vprintf
#	define VSPRINTF vsprintf
#endif // UNICODE

typedef enum
{
    CTPS_D0,   ///< Fully on
    CTPS_D1,   ///<
    CTPS_D2,   ///<
    CTPS_D3    ///< Low Power Mode
}CT_POWER_STATE;

/* Debug assertions */
#if CT_IS_WINDOWS_NT
#  if DBG && !defined(DEBUG)
#    define DEBUG
#  elif (!DBG) && defined(DEBUG)
#    define DBG 1
#  endif
#endif

#ifdef DEBUG
#  if defined(__CT_SYS_WINDOWS_95) || defined(__CT_SYS_MSDOS)
#    ifdef VTOOLSD
#      ifdef __cplusplus
         extern "C" CTVOID __cdecl LOCK_dprintf (const char *, ...);
#      else
         extern CTVOID __cdecl LOCK_dprintf(const char *, ...);
#      endif

#      define CTDPF    LOCK_dprintf
#      define CTASSERT(_x) if (! (_x)) { \
                LOCK_dprintf("Assertion failed in file %s at line %d\n", __FILE__, __LINE__); \
                {_asm _emit 0xcc}  }
#    else
#      include <stdio.h>
#      include <assert.h>
#      define CTDPF  printf
#      define CTASSERT(_x) assert(_x)
#    endif
#  elif (CT_IS_WINDOWS_NT)
#      define CTDEBUGLVL_BLAB       3
#      define CTDEBUGLVL_VERBOSE    2
#      define CTDEBUGLVL_TERSE      1
#      define CTDEBUGLVL_ERROR      0
       /* change CTDEBUG_LEVEL to set the debug level for CTDPF. */
#      define CTDEBUG_LEVEL 3
#      if (CTDEBUG_LEVEL >= CTDEBUGLVL_VERBOSE)
#         if defined(_USRDLL) || defined(_USRKOD)
#             if defined(OVERRIDE_USER_PRINTF)  //for the user driver
#                include <stdio.h>
#                define CTDPF printf
#              else
#                define CTDPF
#             endif //RING3
#         else
#            define CTDPF DbgPrint
#         endif
#      else
         static __inline void CTDPF(char *string, ...) { }
#      endif
#      if defined(_USRDLL) || defined(_USRKOD)
#        include <assert.h>
#        define CTASSERT(_x) assert(_x)
#      else
#        define CTASSERT(_x) ASSERT(_x)
#      endif
#  elif defined(__CT_SYS_LINUX)
#    ifdef __KERNEL__
       CTBEGINCTYPE
#      include <linux/kernel.h> /* Declare "printk" */
       CTENDCTYPE
#      ifdef MONODBG
         CTCTYPE void monoOutValue(char *fmt, ... );
     CTCTYPE int  monoInKey();
         CTCTYPE void monoSetDisplay(int bDoDisplay);
         CTCTYPE void monoDebug();
#        define PDEBUG(fmt, args...) do { monoOutValue(fmt, ## args); } while (0)
#        define CTDPF(fmt, args...)  do { monoOutValue(fmt, ## args); } while (0)
#        define CTASSERT(x) \
           if (!(x)) { \
             monoOutValue("WARNING: ASSERT failed at file %s line %d", __FILE__, __LINE__); \
           }
#        define __THROW_BAD_ALLOC monoOutValue("Error: Out of memory");
#      else /* MONODBG */
#        define PDEBUG(fmt, args...) do { printk("<1>" "(" __FILE__ "): " fmt, ## args); } while (0)
#        define CTDPF(fmt, args...) do { printk("<1>" "(" __FILE__ "): " fmt, ## args); \
                                         printk("\n"); \
                                       } while (0)
#        define CTASSERT(x) \
           if (!(x)) { \
             printk("<1>" "WARNING: ASSERT failed at file %s line %d\n", __FILE__, __LINE__); \
           }
#        define __THROW_BAD_ALLOC printk("<1>" "Error: Out of memory\n");
#      endif /* MONODBG */
#    else /* __KERNEL__ */
#      include <stdio.h>
#      define CTDPF(fmt, args...) fprintf(stderr, fmt, ## args)
#      define CTASSERT(x) \
         if (!(x)) { \
           fprintf(stderr, "WARNING: ASSERT failed at file %s line %d\n", __FILE__, __LINE__); \
           __asm__ ("int3"); \
         }
#    endif /* __KERNEL__ */
#  elif (CT_IS_SOLARIS)
#    ifdef _KERNEL
#      include <sys/debug.h>
#      include <sys/cmn_err.h>
#      define PDEBUG ctdpf_vargs_ptf
#      define CTDPF  ctdpf_vargs_ptf
       static void ctdpf_vargs_ptf(const char *fmt, ...) \
       {\
         va_list ap;\
         char    buf[256];\
         va_start(ap, fmt); \
         vsprintf(buf, fmt, ap);\
         va_end(ap); \
         cmn_err(CE_CONT,"%s", buf);\
       }\

#      define CTASSERT(x) ASSERT(x)
#      define __THROW_BAD_ALLOC cmn_err(CE_WARN," Error: Out of memory\n")
#    else
#      include <assert.h>
#      define PDEBUG      printf
#      define CTDPF       printf
#      define CTASSERT(x) assert(x)
#      define __THROW_BAD_ALLOC printf(" WARNING :: Error: Out of memory\n")
#    endif  /* _KERNEL */
#  elif defined(__CT_SYS_MACINTOSH)
#       include "EmuDriverDebug.h"
#       define CTDPF DebugPrintf
#       define CTASSERT(x) if(!(x)) { DebugAssert(__FILE__,__LINE__); }
#  elif (CT_IS_OSX)
#               include <assert.h>
#       define CTDPF(x) CTNOOP
#       define CTASSERT(x) assert(x)
#  else /* Not PC based, so for now just turn everything off */
#    include <assert.h>
#    define CTASSERT(_x) assert(_x)
#    define CTDPF  printf
#  endif  /* General non-PC case */


#  define CTASSERTONCE(_x)              \
    {                                   \
        static int __bAsserted=FALSE;       \
        if (!(_x) && !__bAsserted)      \
        {                               \
            CTASSERT(0);                \
            __bAsserted = TRUE;         \
        }                               \
    }

#  define CTFAIL(_x)    \
    {                   \
        CTDPF(_x);      \
        CTASSERT(0);    \
    }
#  define CTFAILONCE(_x) \
    {                    \
        CTDPF(_x);       \
        CTASSERTONCE(0); \
    }

/* The following values are system-independent DEBUG definitions */
#  define CTRETURN(_x)                  \
    {                                   \
        if ((_x) != CTSTATUS_SUCCESS)   \
        {                               \
            CTASSERT(0);                \
        }                               \
        return (_x);                    \
    }
#  define CTRETURNONCE(_x)              \
    {                                   \
        if ((_x) != CTSTATUS_SUCCESS)   \
        {                               \
            CTASSERTONCE(0);            \
        }                               \
        return (_x);                    \
    }

#if defined(__CT_SYS_LINUX)
# ifdef __KERNEL__
#   ifdef MONODBG
#     define QDPF(fmt, args...) do { monoOutValue(fmt, ## args); } while (0)
#   else
#     define QDPF(fmt, args...) do { printk("<1>" "(" __FILE__ "): " fmt, ## args); printk("\n"); } while (0)
#   endif /* MONODBG */
# else
#   define QDPF(fmt, args...) fprintf(stderr, fmt, ## args)
# endif /* __KERNEL__ */
#elif (CT_IS_SOLARIS)
#  ifdef _KERNEL
#    define QDPF   ctdpf_vargs_ptf
#  else
#    define QDPF   printf
#  endif  /* _KERNEL */
#elif !defined (__CT_SYS_GENERAL)
    static __inline void QDPF(char *string, ...)
    {
    #ifdef __CT_SYS_MACINTOSH
        #pragma unused(string)
    #endif
    }

#endif /* defined(__CT_SYS_LINUX) */

#else /* !defined(DEBUG) */
#  if CT_IS_WINDOWS
#    ifdef __cplusplus
#      define PDEBUG CTNOOP
#      define CTDPF  CTNOOP
#      define QDPF   CTNOOP
#    else
#      define PDEBUG
#      define CTDPF
#      define QDPF
#    endif // __cplusplus
#  elif defined(__CT_SYS_LINUX)
#   define PDEBUG(fmt, args...)
#   define CTDPF(fmt, args...)
#   define QDPF(fmt, args...)
#  elif (CT_IS_SOLARIS)
#    define    PDEBUG(X)
#    define    CTDPF(X)
#    define    QDPF(X)
#  elif (CT_IS_OSX)
         static __inline void PDEBUG(char *string, ...) { }
         static __inline void CTDPF(char *string, ...) { }
         static __inline void QDPF(char *string, ...) { }
#  else
         static void CTDPF(char *string, ...) { }
         static void QDPF(char *string, ...) { }
#  endif
#  if defined(__CT_SYS_LINUX) || (CT_IS_SOLARIS)
#    define CTASSERT(_x)
#    define CTASSERTONCE(_x)
#    define __THROW_BAD_ALLOC
#  else
#    define CTASSERT(_x) CTNOOP;
#    define CTASSERTONCE(_x) CTNOOP;
#  endif
#  define CTFAIL(_x)
#  define CTFAILONCE(_x)
#  define CTRETURN(_x)   { return (_x); }
#  define CTRETURNONCE(_x)   { return (_x); }

#endif

/* Compile-time assertions */
#define CTCOMPILEASSERT(_x) typedef char _CTCOMPILEASSERT_[(_x)?1:-1]

/* This area for system-specific stuff (regardless of core code/debug status) */
/* Linux-specific stuff */
#ifdef __CT_SYS_LINUX

    /* Kernel/User space stuff */
    #undef  min
    #define min(x,y) ((x) < (y) ? (x) : (y))

    #undef  max
    #define max(x,y) ((x) > (y) ? (x) : (y))

    #ifndef __KERNEL__ /* User space stuff */

        #ifndef __HANDLE_DEFINED__
        #define __HANDLE_DEFINED__
        typedef CTHANDLE HANDLE;
        #endif

        #ifndef __BYTE_DEFINED__
        #define __BYTE_DEFINED__
        typedef CTBYTE BYTE;
        #endif

        #ifndef __BOOL_DEFINED__
        #define __BOOL_DEFINED__
        typedef CTBOOL BOOL;
        #endif

        #ifndef __LPVOID_DEFINED__
        #define __LPVOID_DEFINED__
        typedef PCTVOID LPVOID;
        #endif

        #ifndef __DWORD_DEFINED__
        #define __DWORD_DEFINED__
        typedef CTDWORD DWORD;
        #endif

        #ifndef __ULONG_DEFINED__
        #define __ULONG_DEFINED__
        typedef CTULONG ULONG;
        #endif

        typedef int OVERLAPPED;
        typedef CTDWORD (*FPTHREADSTART)(CTVOID *pArg);

        CTCTYPE CTBOOL DeviceIoControl(CTHANDLE hDevice, CTDWORD dwCode, PCTVOID pInBuf,
                       CTDWORD dwInBufSize, PCTVOID pOutBuf, CTDWORD dwOutBufSize,
                       CTDWORD *pdwBytesRet, OVERLAPPED *pOverlapped);
        CTCTYPE CTHANDLE CreateThread(CTDWORD *pdwAttrib, CTDWORD dwStackSize,
                          FPTHREADSTART pfnStart, CTVOID *pArg,
                          CTDWORD dwFlags, CTDWORD *pdwThreadID);
        CTCTYPE CTBOOL GetExitCodeThread(CTHANDLE hThread, CTDWORD *pdwExitCode);

        CTCTYPE CTBOOL MapPageUserAddr(CTHANDLE hDevice, CTHANDLE hMem, CTDWORD dwSize, PCTVOID *ppUser);
        CTCTYPE CTBOOL UnmapPageUserAddr(PCTVOID pUser, CTDWORD dwSize);

        typedef void* HINSTANCE;
        typedef void* HMODULE;
        CTCTYPE HMODULE LoadLibrary(char *szLibName);
        CTCTYPE void* GetProcAddress(HMODULE hLib, char *szFunction);
        CTCTYPE CTBOOL FreeLibrary(HMODULE hLib);


        #define WINAPI
        #define STILL_ACTIVE 1
        #define FAR

    #endif /* __KERNEL__ */
#endif /* __CT_SYS_LINUX */


#ifdef __CT_SYS_MACINTOSH

    // macintosh specific fakes for windows system routines.
    // function pointer to a thread entry point.
    typedef unsigned long (*FPTHREADSTART)(void *) ;

    // fake out the overlapped type to stop the compiler whining for
    // the proxy builds.
    #define OVERLAPPED  CTHANDLE


    CTWORD DeviceIoControl(
        CTHANDLE    ghDeviceDriver,
        CTDWORD     code,
        PCTVOID     pInBuf,
        CTDWORD     inBufSize,
        PCTVOID     pOutBuf,
        CTDWORD     outBufSize,
        CTDWORD     *dwBytesReturned,
        OVERLAPPED *o) ;

    CTCTYPE CTHANDLE CreateThread(CTDWORD *pdwAttrib, CTDWORD dwStackSize,
                      FPTHREADSTART pfnStart, CTVOID *pArg,
                      CTDWORD dwFlags, CTDWORD *pdwThreadID) ;
    CTCTYPE CTBOOL GetExitCodeThread(CTHANDLE hThread, CTDWORD *pdwExitCode) ;

    #   define WINAPI

        // this is a stuct we use for storing informaiton about a driver connection.

        // this is information we need to connect to and use the driver,
        // it gets setup in in the open driver call for the proxies this probably
        // needs to go somewhere else.  TO DO
        typedef struct {
            DriverRefNum        mRefNum ;
            Str63               mName ;
            DriverEntryPointPtr mEntry ;
        } DriverInfoRecord ;

    #   define _TIME_T
    //#include "regutils.h"

    #   define SZDRVDEV         "/dev/x-fi"
    #   define STILL_ACTIVE     0xFEEDFACE


#endif  // __CT_SYS_MACINTOSH


/* Macros for DeviceIoControl definition */
#ifdef  R3REMOTE
    #  include "ProxyRmt.h"
    //#  define DEVICEIOCONTROL DeviceIoControlSim
    #  define DEVICEIOCONTROL(a,b,c,d,e,f,g,h) DeviceIoControlSim((HANDLE)(a),(b),(c),(d),(e),(f),(g),(OVERLAPPED*)(h))
    //#  define CREATETHREAD    CreateThread
    #  define CREATETHREAD(a,b,c,d,e,f) (CTHANDLE)CreateThread((a),(b),(c),(d),(e),(f))
    //#  define GETEXITCODETHREAD GetExitCodeThread
    #  define GETEXITCODETHREAD(a,b) GetExitCodeThread((HANDLE)(a),(b))
    #elif defined R3NAMED_PIPE_CLIENT
    #   include "R3NamedPipe.h"
    //#  define DEVICEIOCONTROL DeviceIoControlSim
    #   define DEVICEIOCONTROL(a,b,c,d,e,f,g,h) WriteReadNamedPipe((HANDLE)(a),(b),(c),(d),(e),(f),(g),(OVERLAPPED*)(h))
        //#  define CREATETHREAD    CreateThread
    #   define CREATETHREAD(a,b,c,d,e,f) (CTHANDLE)CreateThread((a),(b),(c),(d),(e),(f))
    //#  define GETEXITCODETHREAD GetExitCodeThread
    #   define GETEXITCODETHREAD(a,b) GetExitCodeThread((HANDLE)(a),(b))
#elif (CT_IS_WINDOWS)
#  if (CT_IS_WINDOWS_32)
#     ifdef __THUNK_32_TO_64__
#       define DEVICE_64BIT 0x80000000
#     else
#       define DEVICE_64BIT 0x00000000
#     endif
#  else
#     define DEVICE_64BIT 0x80000000
#  endif
#  define DEVICEIOCONTROL(a,b,c,d,e,f,g,h) DeviceIoControl((HANDLE)(a),(b|DEVICE_64BIT),(c),(d),(e),(f),(g),(h))
#  define CREATETHREAD(a,b,c,d,e,f) (CTHANDLE)CreateThread((a),(b),(c),(d),(e),(f))
#  define GETEXITCODETHREAD(a,b) GetExitCodeThread((HANDLE)(a),(b))
#else
#  define DEVICEIOCONTROL DeviceIoControl
#  define CREATETHREAD CreateThread
#  define GETEXITCODETHREAD GetExitCodeThread
#endif

/** The CTVERSION structure is a simple structure that defines a major and minor version.
  */
typedef struct ctVersionTag
{
    CTUSHORT usMajor;   /**< Major version value */
    CTUSHORT usMinor;   /**< Minor version value */
} CTVERSION, *PCTVERSION, **PPCTVERSION;

#define CTMARKMAGIC 0x10301968

/** The CTFILEVER structure is used as a binary signature for all UDA compiled files that
    is guaranteed to be consistent across all binaries, regardless of the build environment.
 */
typedef struct ctFileVersionTag
{
    CTCHAR   szFileVer[10];   /**< This field must hold the text "CTFILEVER" as the indicator that the file version signature begins here.*/
    CTUSHORT usMajorVerNum;   /**< This field holds the major version number, which is usually consistent for bi-directional compatible binary builds. */
    CTUSHORT usMinorVerNum;   /**< This field holds the minor version number, which combined with usMajorVerNum is usually consistent for bi-directional compatible binary builds which offer similar functionality sets. */
    CTUSHORT usCodeBuildNum;  /**< This field holds the code build number, which signifies a binary whose only difference is a build date or bug fix. */
    CTUSHORT usLangBuildNum;  /**< This field holds the language build number, which signifies a binary build whose only difference is in language support. */
    CTUSHORT usPlatformsSupported; /**< This field holds the platform code number, which signifies any platform restrictions the binary holds. Most UDA binaries are platform independent, and therefore have this value set to 0.*/
    CTUSHORT usLangCode;       /**< This field holds the language code number, which signifies any language restrictions the binary holds. Most UDA binaries are langauge independent, and therefore have this value set to 0.*/
    CTUSHORT usPad;            /**< This field pads the structure to maintain 4-byte alignment */
    CTULONG  ulMarkMagic;      /**< This field marks the location where a post-build binary mark may be placed in the binary. Its value must be 0x10301968. Marking is used when distributing binaries to third parties, to facilitate tracing in the event such binaries leak. */
    CTULONG  ulMarkNum;        /**< This field is the location of the mark. Standard builds must set this value to 0. Marked builds would modify this value, and then update any OS specific binary headers to do checksum fixup. A utility called CTFILEMARK is used to mark Windows binaries. */
} CTFILEVER, *PCTFILEVER, **PPCTFILEVER;

#ifndef __CT_ENDIAN_INDEPENDENT

   /********************************************************************
   * Data structures and methods to handle endian nature diffeneces
   * between archetectures
   ********************************************************************/

/**** For Little Endian Systems ****/

#ifdef __CT_LITTLE_ENDIAN

/** This structure is used in order to allow software to set individual bytes in a multi-byte integer
    without needing to consider endian nature.
 */
typedef struct twoBytesTag
{
  CTUCHAR by0; /**< The Least significant byte */
  CTUCHAR by1; /**< The Most significant byte */
} twoBytes;

/** This structure is used in order to allow software to set individual bytes in a multi-byte integer
    without needing to consider endian nature.
 */
typedef struct fourBytesTag
{
  CTUCHAR by0; /**< The Least significant byte */
  CTUCHAR by1; /**< The Second Least significant byte */
  CTUCHAR by2; /**< The Third Least significant byte */
  CTUCHAR by3; /**< The Most significant byte */
} fourBytes;


/** This structure is used in order to allow software to set individual bytes in a multi-byte integer
    without needing to consider endian nature.
 */
typedef struct twoWordsTag
{
  CTUSHORT w0;
  CTUSHORT w1;
} twoWords;

/**** For Big Endian Systems ****/

#elif defined(__CT_BIG_ENDIAN)

/** This structure is used in order to allow software to set individual bytes in a multi-byte integer
    without needing to consider endian nature.
 */
typedef struct twoBytesTag
{
  CTUCHAR by1; /**< The Most significant byte */
  CTUCHAR by0; /**< The Least significant byte */
} twoBytes;


/** This structure is used in order to allow software to set individual bytes in a multi-byte integer
    without needing to consider endian nature.
 */
typedef struct fourBytesTag
{
  CTUCHAR by3; /**< The Most significant byte */
  CTUCHAR by2; /**< The Third Least significant byte */
  CTUCHAR by1; /**< The Second Least significant byte */
  CTUCHAR by0; /**< The Least significant byte */
} fourBytes;


/** This structure is used in order to allow software to set individual bytes in a multi-byte integer
    without needing to consider endian nature.
 */
typedef struct twoWordsTag
{
  CTUSHORT w1; /**< The Most significant two-byte value */
  CTUSHORT w0; /**< The Least significant two-byte value */
} twoWords;

#endif  /*  BYTE COHERENCY  */

/** This union is used in order to allow software to set individual bytes in a two byte integer
    without needing to consider endian nature.

  @verbatim
  e.g.

  The goal is to write a two byte value in big endian format to a file. The source code should
  remain the same regardless of whether it is written on a big endian or little endian system.

  twoByteUnion tbu;
  tbu.byVals.by0 = 0x4;
  tbu.byVals.by1 = 0x5;

  At this point, tbu.wVal is guaranteed to be 0x0504, regardless of endian nature. We need
  to swap the bytes, but ONLY if the CPU is little endian. So, we use the byte swapping macro:

  SWAP_WORD_BIG_ENDIAN_ONLY (tbu.wVal);

  At this point, tbu.wVal is guaranteed to hold the value 0x0504 in big endian format, 
  regardless of the CPU platform. We may now write that value to the file.

  fwrite(pFile, &tbu.wVal, 2, 1);

  @endverbatim
 */
typedef union twoByteUnionTag
{
  twoBytes  byVals;  /**< Endian nature independent 2-byte structure */
  CTUSHORT  wVal;    /**< Resulting 2-byte value guaranteed to be correct regardless of endian nature */
} twoByteUnion;

/** This union is used in order to allow software to set individual bytes in a four byte integer
    without needing to consider endian nature.
    
  @verbatim
  e.g.

  The goal is to write a four byte value in big endian format to a file. The source code should
  remain the same regardless of whether it is written on a big endian or little endian system.

  SWAP_DWORD_BIG_ENDIAN_DECLARATIONS;

  fourByteUnion fbu;
  fbu.byVals.by0 = 0x4;
  fbu.byVals.by1 = 0x5;
  fbu.byVals.by2 = 0x6;
  fbu.byVals.by3 = 0x7;

  At this point, fbu.dwVal is guaranteed to be 0x07060504, regardless of endian nature. We need
  to swap the bytes, but ONLY if the CPU is little endian. So, we use the byte swapping macro:

  SWAP_DWORD_BIG_ENDIAN_ONLY (fbu.dwVal);

  At this point, fbu.dwVal is guaranteed to hold the value 0x07060504 in big endian format, 
  regardless of the CPU platform.

  fwrite(pFile, &fbu.dwVal, 4, 1);

  @endverbatim
 */
typedef union fourByteUnionTag  
{                               
  fourBytes byVals; /**< Endian nature independent 4-byte structure */
  twoWords  wVals;  /**< Endian nature independent 2-2-byte structure */
  CTULONG   dwVal;  /**< Resulting 4-byte value guaranteed to be correct regardless of endian nature */
} fourByteUnion;

#endif /* __CT_ENDIAN_INDEPENDENT */

   /*********************************
    * Byte swapping inline methods
    *********************************/

/* This chip cannot use unions for byte swapping... */

#ifndef __SYS_TMS320C32

   /************************************************************
    * These are the VARIABLES which are used for byte swapping
    ***********************************************************/

#define SWAP_WORD_DECLARATIONS \
  twoByteUnion tbUnion;        \
  twoByteUnion *ptbUnion;

#define SWAP_DWORD_DECLARATIONS \
  fourByteUnion fbUnion;        \
  fourByteUnion *pfbUnion;

   /************************************************************
    * These are the GENERIC byte software macros. Note they use
    * the variables defined above!
    ***********************************************************/

#define SWAP_WORD(wValue)                               \
{                                                       \
  tbUnion.wVal            = wValue;                     \
  ptbUnion                = (twoByteUnion *)((CTWORD *)(&(wValue)));\
  ptbUnion->byVals.by0    = tbUnion.byVals.by1;         \
  ptbUnion->byVals.by1    = tbUnion.byVals.by0;         \
}

#define SWAP_DWORD(dwValue)                            \
{                                                      \
  fbUnion.dwVal        = dwValue;                      \
  pfbUnion             = (fourByteUnion *)((CTDWORD *)(&(dwValue)));\
  pfbUnion->byVals.by0 = fbUnion.byVals.by3;           \
  pfbUnion->byVals.by1 = fbUnion.byVals.by2;           \
  pfbUnion->byVals.by2 = fbUnion.byVals.by1;           \
  pfbUnion->byVals.by3 = fbUnion.byVals.by0;           \
}


#else  /* For the TMS system, these may need to be
          defined at a later date */

#define SWAP_WORD_DECLARATIONS
#define SWAP_DWORD_DECLARATIONS
#define SWAP_WORD(wValue)
#define SWAP_DWORD(dwValue)

#endif /* Byte swappings based on system */

   /************************************************************
    * These are DEFINES to allow one to have byte swapping
    * routines exist or not exist based on ENDIAN NATURE.
    * (IE byte swap ONLY if LITTLE ENDIAN system).
    ***********************************************************/

   /********************************************************************
    * Byte Coherent (little endian) specific byte swapping methods
    *******************************************************************/
#ifdef __CT_LITTLE_ENDIAN

/* For little endian systems, the BIG ENDIAN methods are NOT defined */
#define SWAP_WORD_BIG_ENDIAN_DECLARATIONS
#define SWAP_DWORD_BIG_ENDIAN_DECLARATIONS
#define SWAP_WORD_BIG_ENDIAN_ONLY(a)
#define SWAP_DWORD_BIG_ENDIAN_ONLY(a)

/* For little endian systems, the LITTLE ENDIAN methods are defined as the
   generic methods */
#define SWAP_WORD_LITTLE_ENDIAN_DECLARATIONS  SWAP_WORD_DECLARATIONS
#define SWAP_DWORD_LITTLE_ENDIAN_DECLARATIONS  SWAP_DWORD_DECLARATIONS
#define SWAP_WORD_LITTLE_ENDIAN_ONLY(a) SWAP_WORD(a)
#define SWAP_DWORD_LITTLE_ENDIAN_ONLY(a) SWAP_DWORD(a)


   /********************************************************************
    * Byte Incoherent (big endian) specific byte swapping methods
    *******************************************************************/
#elif defined (__CT_BIG_ENDIAN)

/* For big endian systems, the BIG ENDIAN methods are defined as the
   generic methods */
#define SWAP_WORD_BIG_ENDIAN_DECLARATIONS SWAP_WORD_DECLARATIONS
#define SWAP_DWORD_BIG_ENDIAN_DECLARATIONS SWAP_DWORD_DECLARATIONS
#define SWAP_WORD_BIG_ENDIAN_ONLY(a) SWAP_WORD(a)
#define SWAP_DWORD_BIG_ENDIAN_ONLY(a) SWAP_DWORD(a)

/* For big endian systems, the LITTLE ENDIAN methods are NOT defined */
#define SWAP_WORD_LITTLE_ENDIAN_DECLARATIONS
#define SWAP_DWORD_LITTLE_ENDIAN_DECLARATIONS
#define SWAP_WORD_LITTLE_ENDIAN_ONLY(a)
#define SWAP_DWORD_LITTLE_ENDIAN_ONLY(a)


   /********************************************************************
    * No endian nature considered: don't define any endian nature
    * switchable byte swapping macros
    *******************************************************************/

#elif defined (__CT_ENDIAN_INDEPENDENT)

#define SWAP_WORD_BIG_ENDIAN_DECLARATIONS
#define SWAP_DWORD_BIG_ENDIAN_DECLARATIONS
#define SWAP_WORD_BIG_ENDIAN_ONLY(a)
#define SWAP_DWORD_BIG_ENDIAN_ONLY(a)

#define SWAP_WORD_LITTLE_ENDIAN_DECLARATIONS
#define SWAP_DWORD_LITTLE_ENDIAN_DECLARATIONS
#define SWAP_WORD_LITTLE_ENDIAN_ONLY(a)
#define SWAP_DWORD_LITTLE_ENDIAN_ONLY(a)

#endif /* Endian nature */

/* This macro should be defined only if necessary for a compiler to export an API
 * function */
#define CTAPIEXPORT

/* These macros are used to properly align the data structures passed in a
   buffer in DeviceIoControl. It is also for thunking 32-bit to 64-bit. */
#if CT_IS_WINDOWS
#if CT_IS_WINDOWS_32
#ifdef __THUNK_32_TO_64__
#ifndef TYPE_ALIGNMENT
#ifdef __cplusplus
#if _MSC_VER >= 1300
#define TYPE_ALIGNMENT( t ) __alignof(t)
#endif
#else
#define TYPE_ALIGNMENT( t ) \
    FIELD_OFFSET( struct { char x; t test; }, test )
#endif
#endif // TYPE_ALIGNMENT
#define R0_POINTER          unsigned __int64
#define ALIGN_POINTER(p, type)  p=(PUCHAR)(((ULONGLONG)(p)+TYPE_ALIGNMENT(type)-1) & ~(TYPE_ALIGNMENT(type)-1))
#define ALIGN_SIZE(size)        ((size+TYPE_ALIGNMENT(void*__ptr64)-1)&(~(TYPE_ALIGNMENT(void*__ptr64)-1)))
#else /* UNDEF __THUNK_32_TO_64__ */
#define R0_POINTER          CTVOID*
#define ALIGN_POINTER(p, type)
#define ALIGN_SIZE(size)    size
#endif /* __THUNK_32_TO_64__ */
#elif CT_IS_WINDOWS_64
#define R0_POINTER          CTVOID*
#define ALIGN_POINTER(p, type)  p=(PUCHAR)(((ULONGLONG)(p)+TYPE_ALIGNMENT(type)-1) & ~(TYPE_ALIGNMENT(type)-1))
#define ALIGN_SIZE(size)        ((size+TYPE_ALIGNMENT(void*)-1)&(~(TYPE_ALIGNMENT(void*)-1)))
#else /* NOT CT_IS_WINDOWS_32 OR CT_IS_WINDOWS_64 */
#define R0_POINTER          CTVOID*
#define ALIGN_POINTER(p, type)
#define ALIGN_SIZE(size)    size
#endif /* CT_IS_WINDOWS_32 */
#else /* NOT CT_IS_WINDOWS */
#define R0_POINTER          CTVOID*
#define ALIGN_POINTER(p, type)
#define ALIGN_SIZE(size)    size
#endif /* CT_IS_WINDOWS */

#endif /* __CTDEF_H */

