/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#ifndef _CTDRV_H_
#   define _CTDRV_H_

#ifdef __cplusplus
extern "C"{
#endif

#include <linux/spinlock.h>
#include <linux/smp.h>

#ifndef PCI_VENDOR_CREATIVE
#   define PCI_VENDOR_CREATIVE              0x1102
#endif

#ifndef PCI_DEVICE_CREATIVE_EMU10K1
#   define PCI_DEVICE_CREATIVE_EMU10K1      0x0002
#endif

#ifndef PCI_DEVICE_CREATIVE_EMU10K2
#   define PCI_DEVICE_CREATIVE_EMU10K2      0x0004
#endif

#ifndef PCI_DEVICE_CREATIVE_TINA
#   define PCI_DEVICE_CREATIVE_TINA         0x0008
#endif

#ifndef PCI_DEVICE_CREATIVE_CTXFI
#   define PCI_DEVICE_CREATIVE_CTXFI        0x0005
#endif

#define PCI_SUBSYSTEM_ID_APS                0x4001

#define EMU10KX_EXTENT                      0x20    // 32 byte I/O space - Is this assumption correct for all hw ?

#define JOY_EXTENT                          0x8     // 8 byte I/O space
#define JOY_FUNCTION                        1       // Currently always 1
#define JOY_BASE                            0x200   // Linux joystick drivers are hardcoded to this

#define MAX_EMU10KX                         3       // Must be 1 or more

#define DEFAULT_DMASIZE                     32      // In KB
#define MIN_DMASIZE                         (PAGE_SIZE/1024)
#define MAX_DMASIZE                         1024

#define EMU10K1_DMA_MASK                    0x7fffffffUL    /* 31bit */
#define AUDIGY_DMA_MASK                     0xffffffffUL    /* 32bit */

// FIXME: This is defined in ICtFeatureSet.h, which is not currently C-compilable
#define MAX_CONFIG_NAME_SIZE                32

#define SZ_DEVNAME_FORMAT                   "%s" // This has to fit the card_table entries

#ifndef DEVICE_COUNT_RESOURCE
#   define DEVICE_COUNT_RESOURCE            6
#endif

typedef CTVOID (*WIB_CALLBACK)(PCTVOID pClient);

typedef CTVOID (*WOB_CALLBACK)(CTVOID);

typedef CTVOID (*MIC_CALLBACK)(PCTVOID pClient, CTDWORD dwNumBytesAvail);

typedef CTVOID (*MOC_CALLBACK)(PCTVOID pClient, HAMONOTIFYTYPE type);

enum UDACARDS
{
    SBLIVE_CARD,
    AUDIGY_CARD,
    CTXFI_CARD,
    NUM_UDA_CARDS   // This should always be last
};

enum UDADEVS // Must match g_aDevTemplate
{
    WAVEIN_DEV,
    WAVEOUT1_DEV,
    WAVEOUT2_DEV,
    BMIXER_DEV,
    UARTIN_DEV,
    UARTOUT_DEV,
    HWSYNTH_DEV,
    AUTOIN_DEV,
    AUTOOUT_DEV,
    TIMER_DEV,
    SBLIVE_DEV,
    AC3MGR_DEV,
    NUM_UDA_DEVS    // This should always be last
};

enum CTALSA_MIXER_DEV
{
    MIXER_MASTER,
    MIXER_FRONT,
    MIXER_REAR,
    MIXER_CENTER,
    MIXER_REARCENTER,
    MIXER_LFE,
    MIXER_SIDE,
    //
    MIXER_LINE,
    MIXER_AUX,
    //
    MIXER_WAVE,
    MIXER_WAVE_FRONT,
    MIXER_WAVE_REAR,
    MIXER_WAVE_CENTER,
    MIXER_WAVE_REARCENTER,
    MIXER_WAVE_LFE,
    MIXER_WAVE_SIDE,
    //    
    MIXER_TONEBASS,
    MIXER_TONETREBLE,
    MIXER_MIC,
    MIXER_TAD,
    MIXER_PCSPEAKER,
    MIXER_CD,
    MIXER_CDDIGITAL,
    MIXER_WATUHEAR,
    MIXER_MICBOOST,
    MIXER_DIGITALOUT,
    MIXER_ANALOGSTMIX,
    MIXER_ANALOGMIX,
    MIXER_DIGITALIO,
    MIXER_VIDEO,
    MIXER_MICFP,
    MIXER_DIGITALIN,
    MIXER_SPDIFIN,
    NUM_CTALSA_MIXERS
};

enum CTALSA_PCM_DEV
{
    PCM_WAVEOUT, /* Set WaveOut device as the default device for playback and capture */
    PCM_ENVAUDIO,
    PCM_AC3MGR,
    PCM_IEC958,
    PCM_WAVEIN, /* Integrated into WaveOut device and no separate device */
    NUM_CTALSA_PCMS
};

typedef struct tMixerObj stMixerObj;

typedef struct tLineNode {
    stMixerObj *pMixerObj;
    PCTVOID pLine;
    CTDWORD dwChannels;
    CTDWORD dwSupportedParams;
    CTBOOL  bRecordable;
    struct tLineNode *pNextLine;
    CTDWORD nActiveChannels;
    CTDWORD dwFirstActive;
} stLineNode;

typedef struct tLineGroup {
    stLineNode *pLineList;
    CTBOOL bRecSelected;    // whether this LineGroup is secected as recording source or not.
    long dwRecVol[2];      
    struct tLineGroup *pNextGroup;
} stLineGroup;

struct tMixerObj {
    PCTVOID pIhaBMixer;
    stLineGroup *pGroupList; // the LineGroups linklist
    stLineGroup *pGroupArray[NUM_CTALSA_MIXERS]; // the LineGroups array, easy to search.
    stLineGroup *pRecGroup; // the research destination LineGroup
    PCTVOID pKontrolList;
    rwlock_t MixerLock;
};

typedef struct tBuffObj {
    PCTVOID pWIB;       // pointer to WaveIn Buffer object
    PCTVOID pWOB;       // pointer to WaveOut Buffer object

    enum CTALSA_PCM_DEV tPCMDev;

    PCTVOID substream;
    stMemHandle mem;    // buffer handle
    CTBYTE  *pVirtAddr;

    CTDWORD dwALSAPos;
    CTDWORD dwCounter;
    CTDWORD dwPosition;
    CTDWORD dwBufferSize;
    CTDWORD dwSampleRate;
    CTDWORD dwSampleFormat;
    CTDWORD dwNumChans;
    CTDWORD dwFreq;
    CTDWORD dwOffset;

    CTBOOL bLoopState;
    CTBOOL b3DAudio;
    CTBOOL bAC3Format;
    CTBOOL bExtFormat;
    CTBOOL bOSSEmul;

    PCTVOID pRuntimeValue;  // must be here, to store some runtime info.
} stBuffObj;

typedef struct tMemWin {
    CTBOOL bValid;
    CTDWORD dwSize;
    CTDWORD dwStart;
    PCTVOID pStart;
} stMemWin;

typedef struct tHwObj {
    struct pci_dev *pPCIDev;
    struct pci_dev *pJoyDev;

    PCTVOID pOSSrv;
    PCTVOID pCtString;
    PCTVOID pAdapterMgr;
    PCTVOID pDispatch;
    PCTVOID pHwMgr;
    PCTVOID pPSM;
    PCTVOID pCard;  // a pointer to struct snd_card;
    
    stMixerObj *pMixerObj;
    int seq_client;

    enum UDACARDS tCard;

    CTWORD wVendorID;
    CTWORD wDeviceID;
    CTWORD wSubsysID;
    CTWORD wRevision;
    
    CTDWORD dwIOBase;
    CTDWORD dwJoyBase;
    CTDWORD dwIRQ;

    char szDevName[64];
    char szCard[32];
    char szVendor[16];
    char szModel[16];

    PCTVOID apIha[NUM_UDA_DEVS];
    int anDev[NUM_UDA_DEVS];
    stMemWin aMemWin[DEVICE_COUNT_RESOURCE];
} stHwObj;

// dispatch related
extern HRESULT ctdrv_dispatch(PCTDIOCPARAMETERS kdioc);

// HwObj related
extern HRESULT ctdrv_HwObj_ctor(stHwObj *pHwObj);

extern HRESULT ctdrv_HwObj_dtor(stHwObj *pHwObj);

// pcm related
// WIB ops
extern HRESULT ctdrv_WID_CreateBuffer(CTHANDLE pWID, WIB_CALLBACK pCallback, stBuffObj *pBuffObj);

extern HRESULT ctdrv_WIB_Release(stBuffObj *pBuffObj);

extern HRESULT ctdrv_WIB_Record(stBuffObj *pBuffObj);

extern HRESULT ctdrv_WIB_Stop(stBuffObj *pBuffObj);

extern HRESULT ctdrv_WIB_Pause(stBuffObj *pBuffObj);

extern CTDWORD ctdrv_WIB_GetRecordPosition(stBuffObj *pBuffObj);

extern CTHANDLE ctdrv_WIB_GetUserData(CTHANDLE pWIB);

// WOB ops
extern HRESULT ctdrv_WOD_CreateBuffer(CTHANDLE pWOD, WOB_CALLBACK pCallback, stBuffObj *pBuffObj);

extern HRESULT ctdrv_WOB_Release(stBuffObj *pBuffObj);

extern HRESULT ctdrv_WOB_Play(stBuffObj *pBuffObj);

extern HRESULT ctdrv_WOB_Stop(stBuffObj *pBuffObj);

extern HRESULT ctdrv_WOB_Pause(stBuffObj *pBuffObj);

extern HRESULT ctdrv_WOB_UpdateBuffer(stBuffObj *pBuffObj, CTDWORD dwALSAPos, CTDWORD dwOffset);

extern CTDWORD ctdrv_WOB_GetPlaybackPosition(stBuffObj *pBuffObj);

extern CTHANDLE ctdrv_WOB_GetUserData(CTHANDLE pWOB);

        
// mixer related
extern HRESULT ctdrv_LineGroup_Build(stMixerObj *pMixerObj);
extern HRESULT ctdrv_LineGroup_Free(stMixerObj *pMixerObj);

extern HRESULT ctdrv_LineGroup_SetMuteStatus(stLineGroup *pLineGroup, CTBOOL bMute);
extern HRESULT ctdrv_LineGroup_GetMuteStatus(stLineGroup *pLineGroup, CTBOOL *pbMute);

extern HRESULT ctdrv_LineGroup_SetMicboostStatus(stLineGroup *pLineGroup, CTBOOL bStatus);
extern HRESULT ctdrv_LineGroup_GetMicboostStatus(stLineGroup *pLineGroup, CTBOOL *pbStatus);

extern HRESULT ctdrv_LineGroup_SetDigitaloutStatus(stLineGroup *pLineGroup, CTBOOL bStatus);
extern HRESULT ctdrv_LineGroup_GetDigitaloutStatus(stLineGroup *pLineGroup, CTBOOL *pbStatus);

extern HRESULT ctdrv_LineGroup_SetRecorded(stMixerObj *pMixerObj,stLineGroup *pLineGroup);//updated by shikang
extern HRESULT ctdrv_LineGroup_GetRecorded(stMixerObj *pMixerObj, enum CTALSA_MIXER_DEV *pMixerDev);

extern HRESULT ctdrv_LineGroup_SetTreblelevel(stLineGroup *pLineGroup, long nLevel);
extern HRESULT ctdrv_LineGroup_GetTreblelevel(stLineGroup *pLineGroup, long *pnLevel);

extern HRESULT ctdrv_LineGroup_SetBasslevel(stLineGroup *pLineGroup, long nLevel);
extern HRESULT ctdrv_LineGroup_GetBasslevel(stLineGroup *pLineGroup, long *pnLevel);

extern HRESULT ctdrv_LineGroup_SetVolume(stLineGroup *pLineGroup, long anVol[], enum CTALSA_MIXER_DEV tMixerDev);
extern HRESULT ctdrv_LineGroup_GetVolume(stLineGroup *pLineGroup, long anVol[], enum CTALSA_MIXER_DEV tMixerDev);
extern HRESULT ctdrv_LineGroup_SetVolume_record(stLineGroup *pLineGroup,long anVol[],enum CTALSA_MIXER_DEV tMixerDev);
extern HRESULT ctdrv_LineGroup_GetVolume_record(stLineGroup *pLineGroup,long anVol[],enum CTALSA_MIXER_DEV tMixerDev);
extern HRESULT ctdrv_LineGroup_GetDigitalioStatus(stLineGroup *pLineGroup, CTBOOL *pbStatus);
extern HRESULT ctdrv_LineGroup_SetDigitalioStatus(stLineGroup *pLineGroup, CTBOOL bStatus);
extern HRESULT ctdrv_LineGroup_GetBMixerlineName(stMixerObj *pMixerObj);
extern HRESULT ctdrv_getszRecName(stMixerObj *pMixerObj,int* ctalsa_rec_num,int *num);

// midi related

// sound font init and exit

extern HRESULT ctdrv_SoundFont_init(stHwObj *pHwObj);

extern HRESULT ctdrv_SoundFont_exit(stHwObj *pHwObj);

// output
extern HRESULT ctdrv_MOD_CreateClient(PCTVOID pMOD, MOC_CALLBACK pCallBack, PCTVOID *ppMOC);

extern HRESULT ctdrv_MOC_Release(PCTVOID pMOC);

extern HRESULT ctdrv_MOC_Reset(PCTVOID pMOC);

extern HRESULT ctdrv_MOC_SetUserData(PCTVOID pMOC, CTHANDLE hUserData);

extern CTHANDLE ctdrv_MOC_GetUserData(PCTVOID pMOC);

extern HRESULT ctdrv_MOC_SendBuffer(PCTVOID pMOC, CTUCHAR *pBuf, CTULONG dwBufferSize);

// input
extern HRESULT ctdrv_MID_CreateRawClient(PCTVOID pMID, MIC_CALLBACK pCallBack, PCTVOID *ppMIC);

extern HRESULT ctdrv_MIC_Release(PCTVOID pMIC);

extern HRESULT ctdrv_MIC_Start(PCTVOID pMIC);

extern HRESULT ctdrv_MIC_Stop(PCTVOID pMIC);

extern HRESULT ctdrv_MIC_Reset(PCTVOID pMIC);

extern HRESULT ctdrv_MIC_SetUserData(PCTVOID pMIC, CTHANDLE hUserData);

extern CTHANDLE ctdrv_MIC_GetUserData(PCTVOID pMIC);

extern HRESULT ctdrv_MIC_Read(PCTVOID pMIC, CTVOID *pBuffer, CTDWORD dwBufferSize, CTDWORD *pdwBytesRead);


#ifdef __cplusplus
}
#endif

#endif /* _CTDRV_H_ */

