/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/
#ifndef _OSUTILS_H_
#   define _OSUTILS_H_

#   include <ctcomdef.h>
#   include <linux/version.h>
#   include <linux/unistd.h>
#   include <ctioctl.h>

CTBEGINCTYPE

#   define MIXER_MINOR     0
#   define SEQUENCER_MINOR 1
#   define MIDI_MINOR      2
#   define DSP_MINOR       3
#   define AUDIO_MINOR     4
#   define DSP16_MINOR     5
#   define SNDSTAT_MINOR   6
#   define DATAPASS_MINOR  7
#   define ALTSEQ_MINOR    8
#   define RAWSYNTH_MINOR  9
#   define SFMAN_MINOR     10 
#   define EMUPIA_MINOR    11
#   define ADSP_MINOR      12
#   define AMIDI_MINOR     13
#   define ADMMIDI_MINOR   14
#   define EMU10KX_MINOR   15 
#   define SECURE_MINOR    80 // Hack - If 5+ soundcards are used, we are screwed

#   if defined(__KERNEL__) && !defined(MODULE)
#       define INSERT_VERSION_INFO
#   else
#       define INSERT_VERSION_INFO          \
            extern CTFILEVER gVerInfo;      \
            CTFILEVER *g_pVerInfo = &gVerInfo;
#   endif

// Datatypes required for page alloc/free functions
typedef enum {
    MEMTYPE_PAGE = 1,
    MEMTYPE_KMALLOC,
    MEMTYPE_VMALLOC,
    MEMTYPE_UMALLOC     // to indicate the memory allocated in user space.
} MEMTYPE;

typedef struct {
    MEMTYPE         MemType;
    CTDWORD         dwPageRange;
    CTVOID          *pVirtAddr;
    unsigned long   ulOrder;
    unsigned long   ulUserAddr;
    CTBOOL          bMapped;
    CTBOOL          bReserved;
    unsigned long   pid;
} stMemHandle; 


#   define MAX_PATH_LEN 512
#   define MAX_FILE_LEN 256

// These must match the defs in plugindll.h
typedef void (*PFONDLLCANUNLOAD)(CTHANDLE hInstance, CTHANDLE hUser);
    
typedef HRESULT (STDAPICALLTYPE *PFDLLCANUNLOADNOW)(void);
typedef HRESULT (STDAPICALLTYPE *PFDLLADVISEIFCANUNLOAD)(PFONDLLCANUNLOAD pfHandler, CTHANDLE hUser);
typedef HRESULT (STDAPICALLTYPE *PFDLLGETCLASSOBJECT)(REFCLSID rclsid, REFIID riid, PCTVOID* ppv);
typedef HRESULT (STDAPICALLTYPE *PFDLLENUMCLSIDS)(CTDWORD dwIndex, CLSID* pclsid);
typedef HRESULT (STDAPICALLTYPE *PFDLLENUMDEFAULTSTRINGS)(CTDWORD i, CTLOCSTR* pLocStr, PCTCHAR pszAsciiString);
typedef HRESULT (STDAPICALLTYPE *PFDLLENUMDEFAULTUNITS)(CTDWORD i, CTLOCUNIT* pLocUnit, PCTCHAR pszAsciiString);

typedef CTDWORD (*PFPLUGINADDREF)(void);
typedef CTDWORD (*PFPLUGINRELEASE)(void);

typedef struct tPluginNode {
    char                    szName[MAX_FILE_LEN];
    char                    szPath[MAX_PATH_LEN];
    PFDLLCANUNLOADNOW       pfnDllCanUnloadNow;
    PFDLLADVISEIFCANUNLOAD  pfnDllAdviseIfCanUnload;
    PFDLLGETCLASSOBJECT     pfnDllGetClassObject;
    PFDLLENUMCLSIDS         pfnDllEnumCLSIDs;
    PFDLLENUMDEFAULTSTRINGS pfnDllEnumDefaultStrings;
    PFDLLENUMDEFAULTUNITS   pfnDllEnumDefaultUnits;
    PFPLUGINADDREF          pfnPluginAddRef;
    PFPLUGINRELEASE         pfnPluginRelease;
    struct tPluginNode     *pNext;
} PLUGINNODE, *PPLUGINNODE;


// These version boundaries are not exactly precise
// Early 2.3.x kernels may require some tweaking here to compile correctly
#   if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
#       define RSRCISIOREGION(dev,num) ((dev)->resource[(num)].start != 0 &&    \
            ((dev)->resource[(num)].flags & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_IO)
#       define RSRCISMEMORYREGION(dev,num) ((dev)->resource[(num)].start != 0 && \
            ((dev)->resource[(num)].flags & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_MEMORY)
#       define RSRCADDRESS(dev,num) ((dev)->resource[(num)].start)
#       define CTWAITQUEUE wait_queue_head_t 
#       define VM_OFFSET(vma) (vma->vm_pgoff)
#       define init_waitqueue init_waitqueue_head
#       define PCI_ENABLE_DEVICE(pci_dev) pci_enable_device(pci_dev)
#       define PCI_SET_MASTER(pci_dev) pci_set_master(pci_dev)
#   else
#       define RSRCISIOREGION(dev,num) ((dev)->base_address[(num)] != 0 &&      \
            ((dev)->base_address[(num)] & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_IO)
#       define RSRCISMEMORYREGION(dev,num) ((dev)->base_address[(num)] != 0 && \
            ((dev)->base_address[(num)] & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_MEMORY)
#       define RSRCADDRESS(dev,num) (dev)->base_address[(num)]
#       define CTWAITQUEUE  struct wait_queue *
#       define VM_OFFSET(vma) (vma->vm_offset)
#       define init_MUTEX(pSem) *(pSem) = MUTEX
#       define init_MUTEX_LOCKED(pSem) *(pSem) = MUTEX_LOCKED
#       define page_address(x) (x)
#       define PCI_ENABLE_DEVICE(pci_dev) 0
#       define PCI_SET_MASTER(pci_dev) 
#   endif // LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)


#   if defined(DEBUG) && !defined(MODULE)
#       define ALLOC_ARGLIST size_t size, CTBOOL bZero
#       define ALLOCARG(size,zero) size, zero
#       define CT_MOD_INC_USE_COUNT 
#       define CT_MOD_DEC_USE_COUNT
#       define CT_EXPORT_NO_SYMBOLS
#       define CT_EXPORT_SYMBOL_NOVERS(x)
#   else
#       define ALLOC_ARGLIST size_t size, CTBOOL bZero
#       define ALLOCARG(size,zero) size, zero
#       define CT_MOD_INC_USE_COUNT MOD_INC_USE_COUNT
#       define CT_MOD_DEC_USE_COUNT MOD_DEC_USE_COUNT
#       define CT_EXPORT_NO_SYMBOLS EXPORT_NO_SYMBOLS
#       if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 9)
#           define CT_EXPORT_SYMBOL_NOVERS(x) EXPORT_SYMBOL_NOVERS(x)
#       else
#           define CT_EXPORT_SYMBOL_NOVERS(x) EXPORT_SYMBOL(x);
#       endif
#   endif

// Exported functions
CTHANDLE get_ossrv(void);
CTHANDLE get_devreg(void);
CTHANDLE get_sfman(void);
CTHANDLE get_emupia(void);
CTHANDLE get_ha(void);
CTHANDLE get_secure(void);
HRESULT register_plugin(PLUGINNODE *pPluginNode);
HRESULT unregister_plugin(char *szName);
void unload_all_plugins(void);
void *heap_alloc(ALLOC_ARGLIST);
void heap_free(void *pv);

// Utility functions
void __do_global_ctors_aux(void);
void __do_global_dtors_aux(void);

//HRESULT reserve_remap_pages(unsigned long ulUserAddr, stMemHandle *pMem, unsigned long ulSize, pgprot_t prot)
HRESULT unreserve_pages(stMemHandle *pMem, unsigned long ulSize);
unsigned long bytes_to_shift(CTDWORD dwBytes);
unsigned long bytes_to_order(CTDWORD dwBytes);
typedef HRESULT (*FPDIOCDISPATCH)(PCTDIOCPARAMETERS);
int ioctl_dispatch(unsigned long ulArg, FPDIOCDISPATCH pfnDiocDispatch);

CTENDCTYPE

#endif // _OSUTILS_H_
