/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
UDA PIA Plugin DLL Export Function Table
Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/
#ifndef _PluginDLL_h
#define _PluginDLL_h

#include "ctcomdef.h"

typedef void (__cdecl *FPONDLLCANUNLOAD)(HINSTANCE hInstance, HANDLE hUser);
typedef HRESULT (STDAPICALLTYPE *FPDLLCANUNLOADNOW)(void);
typedef HRESULT (STDAPICALLTYPE *FPDLLGETCLASSOBJECT)(REFCLSID rclsid, REFIID riid, LPVOID* ppv);
typedef HRESULT (STDAPICALLTYPE *FPDLLENUMCLSIDS)(DWORD dwIndex, CLSID* pclsid);
typedef HRESULT (STDAPICALLTYPE *FPDLLENUMDEFAULTSTRINGS)(CTDWORD i, CTLOCSTR* pLocStr, PCTCHAR pszAsciiString);
typedef HRESULT (STDAPICALLTYPE *FPDLLENUMDEFAULTUNITS)(CTDWORD i, CTLOCUNIT* pLocUnit, PCTCHAR pszAsciiString);


#if defined(__CT_SYS_LINUX)
CTCTYPE BOOL DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID /*lpReserved*/);
#else
BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID /*lpReserved*/);
#endif

// Returns either S_OK = Can be unloaded, or S_FALSE = Cannot be unloaded;
STDAPI DllCanUnloadNow(void);

// Pass in a function pointer to register for a callback.  Pass in NULL to unregister.
STDAPI DllAdviseIfCanUnload(FPONDLLCANUNLOAD pfHandler, HANDLE hUser);

// Returns a class factory to create an object of the requested type
//  REFCLSID - CLSID for the object
//  REFIID   - IID Reference to the identifier of the interface that communicates with the class object
//  LPVOID*  - Indirect pointer to the interface
STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID* ppv);

// Returns a CLSID for a given index (0, N).  Returns NULL if the index is invalid.
STDAPI DllEnumCLSIDs(DWORD dwIndex, CLSID* pclsid);

STDAPI DllEnumDefaultStrings(CTDWORD i, CTLOCSTR* pLocStr, PCTCHAR pszAsciiString);
STDAPI DllEnumDefaultUnits(CTDWORD i, CTLOCUNIT* pLocUnit, PCTCHAR pszAsciiString);
STDAPI DllEnumDefaultNonLocalizableStrings(CTDWORD i, CTLOCSTR* pLocStr, PCTCHAR pszAsciiString); // Non-Localizable
STDAPI DllEnumDefaultNonLocalizableUnits(CTDWORD i, CTLOCUNIT* pLocUnit, PCTCHAR pszAsciiString); // Non-Localizable


#ifdef __cplusplus

template<class T>
class CGlobalLockObject : public T
{
	long*				m_pcLock;
	FPONDLLCANUNLOAD	m_pfOnCanUnload;
    HANDLE              m_hOnCanUnloadUserParam;
	HINSTANCE			m_hInstance;

public:
	CGlobalLockObject(long* pcLock, FPONDLLCANUNLOAD pfOnCanUnload,
                        HANDLE hOnCanUnloadUserParam, HINSTANCE hInstance) :
		m_pcLock(pcLock), m_pfOnCanUnload(pfOnCanUnload),
        m_hOnCanUnloadUserParam(hOnCanUnloadUserParam), m_hInstance(hInstance)
	{
		(*m_pcLock)++;
	}
	~CGlobalLockObject()
	{
		if((--(*m_pcLock)) == 0 && m_pfOnCanUnload != NULL)
		{
            (*m_pfOnCanUnload)(m_hInstance, m_hOnCanUnloadUserParam);
		}
	}
};

//
// Here's an alternative way to do the DLL lock object, since the VC++
// compiler sometimes has problems with objects that incorporate too many 
// levels of templatization.
//
// Rather than parameterizing the template class CGlobalLockObject with the 
// plugin class type, we define the following base class from which
// plugins can inherit.  To make sure its destructor gets called at the 
// correct time (ie, LAST), this DllLockBaseClass must be the FIRST class  
// among any other multiply-inherited classes used by the plugin.
//
// The DLL itself is responsible for defining and initializing the static 
// variables used by this class for reference counting, etc.  You'll get a 
// link time error if you don't include something the following in your
// .cpp file:
/*
long	    			DllLockBaseClass::g_cLock ;
FPONDLLCANUNLOAD	DllLockBaseClass::g_pfOnCanUnload ;
HANDLE            DllLockBaseClass::g_hOnCanUnloadUserParam ;
HINSTANCE			DllLockBaseClass::g_hInstance ;
*/

class DllLockBaseClass {
public:
   static long	    			g_cLock ;
   static FPONDLLCANUNLOAD	g_pfOnCanUnload ;
   static HANDLE           g_hOnCanUnloadUserParam ;
   static HINSTANCE			g_hInstance ;

   DllLockBaseClass() {
		g_cLock++;  // Count instances of plugins built from this DLL
	}

	~DllLockBaseClass()
	{
		if( --g_cLock == 0 && g_pfOnCanUnload != NULL)
		{
            (*g_pfOnCanUnload)(g_hInstance, g_hOnCanUnloadUserParam);
		}
   } ;
} ;


#endif // __cplusplus

#define CreateClassObject(classname)\
	CGlobalLockObject<classname>* pUnknown = \
		new CGlobalLockObject<classname>(&g_cLock, g_pfOnCanUnload, \
                                        g_hOnCanUnloadUserParam, \
										g_hInstance);\
	hr = pUnknown->QueryInterface(riid, ppv);\
	if(FAILED(hr))\
		delete pUnknown;

#define EnumDefaultString(classname)\
	*pLocStr = *classname::m_DefaultStrings[i].plocStr;\
	for(j=0; j<128; j++)\
		pszAsciiString[j] = classname::m_DefaultStrings[i].szString[j];

#define EnumDefaultUnit(classname)\
	*pLocUnit = *classname::m_DefaultUnits[i].plocUnit;\
	for(j=0; j<128; j++)\
		pszAsciiString[j] = classname::m_DefaultUnits[i].szString[j];

// Non-Localizable
#define EnumDefaultNonLocalizableString(classname)\
	*pLocStr = *classname::m_DefaultNonLocalizableStrings[i].plocStr;\
	for(j=0; j<128; j++)\
		pszAsciiString[j] = classname::m_DefaultNonLocalizableStrings[i].szString[j];

#define EnumDefaultNonLocalizableUnit(classname)\
	*pLocUnit = *classname::m_DefaultNonLocalizableUnits[i].plocUnit;\
	for(j=0; j<128; j++)\
		pszAsciiString[j] = classname::m_DefaultNonLocalizableUnits[i].szString[j];


extern int cfx_sprintf ( char * buffer, const char * format, ... ); 

#endif // _PluginDLL_h
