/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
Copyright (C) 2007, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/
#ifndef _WINTYPES_H_
#define _WINTYPES_H_

#include "ctdef.h"

#ifndef _FILETIME_
#define _FILETIME_
typedef CTVOID *          PFILETIME;
#endif

#ifndef _SECURITY_ATTRIBUTES_
#define _SECURITY_ATTRIBUTES_
typedef CTVOID *          LPSECURITY_ATTRIBUTES;
#endif

#ifndef _OVERLAPPED_
#define _OVERLAPPED_
typedef CTVOID *          LPOVERLAPPED;
#endif

// Additional Windows Defined Data Types
typedef CTDWORD           REGSAM;

// Windows Registry defined constants
typedef CTVOID *          __HKEY;
typedef CTVOID **         __PHKEY;

#define ERROR_SUCCESS                   0L
#define ERROR_FILE_NOT_FOUND            2L
#define ERROR_MORE_DATA                 234L    // dderror
#define ERROR_NO_MORE_ITEMS             259L

#define REG_NONE                        ( 0 )   // No value type
#define REG_SZ                          ( 1 )   // Unicode nul terminated string
#define REG_EXPAND_SZ                   ( 2 )   // Unicode nul terminated string
                                                // (with environment variable references)
#define REG_BINARY                      ( 3 )   // Free form binary
#define REG_DWORD                       ( 4 )   // 32-bit number
#define REG_DWORD_LITTLE_ENDIAN         ( 4 )   // 32-bit number (same as REG_DWORD)
#define REG_DWORD_BIG_ENDIAN            ( 5 )   // 32-bit number
#define REG_LINK                        ( 6 )   // Symbolic Link (unicode)
#define REG_MULTI_SZ                    ( 7 )   // Multiple Unicode strings
#define REG_RESOURCE_LIST               ( 8 )   // Resource list in the resource map
#define REG_FULL_RESOURCE_DESCRIPTOR    ( 9 )   // Resource list in the hardware description
#define REG_RESOURCE_REQUIREMENTS_LIST  ( 10 )


/*

  Introduction
  ============
  The Creative Header Format is a special mechanism to break away from format dependencies when
  the algorithms used for performing checksum, encryption and decryption changes.  It has the added
  robustness to adhere to the mode of operation specified by the file attributes.

  Mode of Operation
  =================
  REGISTRY_READ      - Read only mode.  Fails if file does not exists.  The file attribute of the
                       opened file must be either in READ_MODE or READWRITE_MODE.
  REGISTRY_WRITE     - Read and write mode.  Overwrites if file exists and creates if it doesn't.
                       This is a privileged mode which creates READ only databases.
  REGISTRY_READWRITE - Read and write mode.  Open if exists and creates if it doesn't.  The file
                       attribute of the opened file must be READWRITE_MODE.  If file creation
                       occurs, the file attribute is set to READWRITE_MODE.

  Encryption Status
  =================
  ENABLE             - Enable encryption.  The encryptors used are specified by the corresponding
                       encryptors GUID.  Works only with REGISTRY_WRITE mode.
  DISABLE            - Disable encryption.  The encryptors are ignored.  Works only with
                       REGISTRY_WRITE mode.
  OVERWRITE_ENABLE   - The encryptors specified must match those embedded in the file that was
                       opened using REGSITRY_READ and REGISTRY_READWRITE mode.  If a file is
                       created via REGISTRY_READWRITE mode, encryption is enabled via the
                       corresponding encryptors.
  OVERWRITE_DISABLE  - The encryptors specified must match those embedded in the file that was
                       opened using REGSITRY_READ and REGISTRY_READWRITE mode.  If a file is
                       created via REGISTRY_READWRITE mode, encryption is disabled and all
                       encryptors are ignored.
*/


// Creative Registry Header Format (Version 1.00)
typedef struct CreativeHeaderTag
{
    CTULONG Version;                  // Version number of Creative Header Format
    CTULONG FileAttributes;           // File Attributes
    CTULONG EncryptionStatus;         // Encryption Status

    GUID KeyNameEncryptorGUID;      // GUID of Encryptor
    GUID ValueNameEncryptorGUID;    // GUID of Encryptor
    GUID DataEncryptorGUID;         // GUID of Encryptor

}   CREATIVEHDR, *PCREATIVEHDR;


#define CREATIVE_HEADER_VER_1   0x00010000

#define REGISTRY_READ           0x00000001
#define REGISTRY_WRITE          0x00000002
#define REGISTRY_READWRITE      0x00000003

#define ENABLE                  0x00000001
#define DISABLE                 0x00000002
#define OVERWRITE_ENABLE        0x00000003
#define OVERWRITE_DISABLE       0x00000004


#endif // _WINTYPES_H_
