/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
ALSA Mixer implementation
Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#include "ctalsa.h"
#include "ctdrv.h"
static int nFirst = 0;
static int ctalsa_rec_num[NUM_CTALSA_MIXERS];
static int nNumOfRecGroupFound = 0;
// macros - to initialize the different controls
#define INIT_VOLCTL(szName, tMixer) .VolCtl = { \
                                        .iface = SNDRV_CTL_ELEM_IFACE_MIXER, \
                                        .name = szName, \
                                        .index = 0, \
                                        .access = SNDRV_CTL_ELEM_ACCESS_READWRITE, \
                                        .info = snd_ctalsa_info_volume, \
                                        .get = snd_ctalsa_get_volume, \
                                        .put = snd_ctalsa_put_volume, \
                                        .private_value = (unsigned long)tMixer \
                                    }

#define INIT_SWHCTL(szName, tMixer) .SwhCtl = { \
                                        .iface = SNDRV_CTL_ELEM_IFACE_MIXER, \
                                        .name = szName, \
                                        .index = 0, \
                                        .access = SNDRV_CTL_ELEM_ACCESS_READWRITE, \
                                        .info = snd_ctalsa_info_switch, \
                                        .get = snd_ctalsa_get_switch, \
                                        .put = snd_ctalsa_put_switch, \
                                        .private_value = (unsigned long)tMixer \
                                    }

#define INIT_RECCTL(szName, tMixer) .RecCtl = { \
                                        .iface = SNDRV_CTL_ELEM_IFACE_MIXER, \
                                        .name = szName, \
                                        .index = 0, \
                                        .access = SNDRV_CTL_ELEM_ACCESS_READWRITE, \
                                        .info = snd_ctalsa_info_capture_volume, \
                                        .get = snd_ctalsa_get_capture_volume, \
                                        .put = snd_ctalsa_put_capture_volume, \
                                        .private_value = (unsigned long)tMixer \
                                    }


// volume ops
static int
snd_ctalsa_info_volume(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;

//    CTDPF("snd_ctalsa_info_volume is called, mixer type = %d", tMixer);

    uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
    uinfo->value.integer.min = 0;

    switch (tMixer) {
        case MIXER_WAVE_CENTER:
        case MIXER_WAVE_REARCENTER:
        case MIXER_WAVE_LFE:
        case MIXER_TONEBASS:
        case MIXER_TONETREBLE:
            uinfo->count = 1;
            uinfo->value.integer.max = 65535;
            break;
        case MIXER_FRONT:
        case MIXER_REAR:
        case MIXER_SIDE:
            uinfo->count = 2;
            uinfo->value.integer.max = 100;
            break;
        case MIXER_CENTER:
        case MIXER_REARCENTER:
        case MIXER_LFE:
            uinfo->count = 1;
            uinfo->value.integer.max = 100;
            break;
        default:
            uinfo->count = 2;
            uinfo->value.integer.max = 65535;
            break;
    }

    return 0;
}

static int
snd_ctalsa_get_volume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    stHwObj *pHwObj = (stHwObj *)snd_kcontrol_chip(kcontrol);
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;
    stLineGroup *pLineGroup = pMixerObj->pGroupArray[tMixer];
    unsigned long flags;
    HRESULT hr;

//    CTDPF("snd_ctalsa_get_volume is called, mixer type = %d", tMixer);

    read_lock_irqsave(&pMixerObj->MixerLock, flags);
    
    switch (tMixer) {
        case MIXER_TONEBASS:
            hr = ctdrv_LineGroup_GetBasslevel(pLineGroup, ucontrol->value.integer.value);
            break;
        case MIXER_TONETREBLE:
            hr = ctdrv_LineGroup_GetTreblelevel(pLineGroup, ucontrol->value.integer.value);
            break;
        default:
            hr = ctdrv_LineGroup_GetVolume(pLineGroup, ucontrol->value.integer.value, tMixer);
            break;
    }
    if (FAILED(hr))
        memset(ucontrol->value.integer.value, 0, kcontrol->count*sizeof(long));
    read_unlock_irqrestore(&pMixerObj->MixerLock, flags);
    
    return 0;
}

static int
snd_ctalsa_put_volume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    stHwObj *pHwObj = (stHwObj *)snd_kcontrol_chip(kcontrol);
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;
    stLineGroup *pLineGroup = pMixerObj->pGroupArray[tMixer];
    unsigned long flags;
    int changed = 0;
    HRESULT hr;

//    CTDPF("snd_ctalsa_put_volume is called, mixer type = %d", tMixer);
    write_lock_irqsave(&pMixerObj->MixerLock, flags);
    
    switch (tMixer) {
        case MIXER_TONEBASS:
            hr = ctdrv_LineGroup_SetBasslevel(pLineGroup, ucontrol->value.integer.value[0]);
            break;
        case MIXER_TONETREBLE:
            hr = ctdrv_LineGroup_SetTreblelevel(pLineGroup, ucontrol->value.integer.value[0]);
            break;
        default:
            hr = ctdrv_LineGroup_SetVolume(pLineGroup, ucontrol->value.integer.value, tMixer);
            break;
    }
    write_unlock_irqrestore(&pMixerObj->MixerLock, flags);
    
    if (SUCCEEDED(hr)) changed = 1;

    return changed;
}

// switch ops
static int
snd_ctalsa_info_switch(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
#if DEBUG
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;
#endif

//    CTDPF("snd_ctalsa_info_switch is called, mixer type = %d", tMixer);

    uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
    uinfo->count = 1;
    uinfo->value.integer.min = 0;
    uinfo->value.integer.max = 1;

    return 0;
}
static int
snd_ctalsa_get_switch(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    stHwObj *pHwObj = (stHwObj *)snd_kcontrol_chip(kcontrol);
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;
    stLineGroup *pLineGroup = pMixerObj->pGroupArray[tMixer];
    CTBOOL bStaus = FALSE;
    HRESULT hr = E_FAIL;

//    CTDPF("snd_ctalsa_get_switch is called, mixer type = %d", tMixer);

    switch (tMixer) {
        case MIXER_MICBOOST:
            hr = ctdrv_LineGroup_GetMicboostStatus(pLineGroup, &bStaus);
            break;
        case MIXER_DIGITALOUT:
            hr = ctdrv_LineGroup_GetDigitaloutStatus(pLineGroup, &bStaus);
            break;
	case MIXER_DIGITALIO:
	    hr = ctdrv_LineGroup_GetDigitalioStatus(pLineGroup, &bStaus);
            break;
        default:
            hr = ctdrv_LineGroup_GetMuteStatus(pLineGroup, &bStaus);
            break;
    }
    
    if (S_OK == hr) {
        ucontrol->value.integer.value[0] = bStaus;
    } else {
        ucontrol->value.integer.value[0] = FALSE;
    }

    return 0;
}
static int
snd_ctalsa_put_switch(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    stHwObj *pHwObj = (stHwObj *)snd_kcontrol_chip(kcontrol);
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;
    stLineGroup *pLineGroup = pMixerObj->pGroupArray[tMixer];
    int changed = 0;
    HRESULT hr;

//    CTDPF("snd_ctalsa_put_switch is called, mixer type = %d", tMixer);

    switch (tMixer) {
        case MIXER_MICBOOST:
            hr = ctdrv_LineGroup_SetMicboostStatus(pLineGroup, ucontrol->value.integer.value[0]);
            break;
        case MIXER_DIGITALOUT:
            hr = ctdrv_LineGroup_SetDigitaloutStatus(pLineGroup, ucontrol->value.integer.value[0]);
            break;
	case MIXER_DIGITALIO:
	    hr = ctdrv_LineGroup_SetDigitalioStatus(pLineGroup, ucontrol->value.integer.value[0]);
            break;
        default:
            hr = ctdrv_LineGroup_SetMuteStatus(pLineGroup, ucontrol->value.integer.value[0]);
            break;
    }

    if (SUCCEEDED(hr))
        changed = 1;

    return changed;
}


// capture ops
static int
snd_ctalsa_info_capture_volume(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
//#if DEBUG
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;
//#endif
//    CTDPF("snd_ctalsa_info_capture_volume is called, mixer type = %d", tMixer);

    uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
    uinfo->count = 2;
    uinfo->value.integer.min = 0;
    uinfo->value.integer.max = 65535;
    return 0;
}
static int
snd_ctalsa_get_capture_volume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    stHwObj *pHwObj = (stHwObj *)snd_kcontrol_chip(kcontrol);
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;
    stLineGroup *pLineGroup = pMixerObj->pGroupArray[tMixer];
    unsigned long flags;

    HRESULT hr = E_FAIL;//added by shikang 070108
    int changed = 0;//added by shikang 070108

//    CTDPF("snd_ctalsa_get_capture_volume is called, mixer type = %d", tMixer);

    read_lock_irqsave(&pMixerObj->MixerLock, flags);

    //added by shikang
    pLineGroup->bRecSelected = TRUE;
    if (nFirst){
        // get the volume level directly from the current LineGroup
        ucontrol->value.integer.value[0] = pLineGroup->dwRecVol[0];
        ucontrol->value.integer.value[1] = pLineGroup->dwRecVol[1];
        read_unlock_irqrestore(&pMixerObj->MixerLock, flags);
        return changed;	
    }
    pLineGroup->dwRecVol[0] = 49810;
    pLineGroup->dwRecVol[1] = 49810;
    ucontrol->value.integer.value[0] = pLineGroup->dwRecVol[0];
    ucontrol->value.integer.value[1] = pLineGroup->dwRecVol[1];
//    hr = ctdrv_LineGroup_GetVolume_record(pMixerObj->pRecGroup,ucontrol->value.integer.value,tMixer); 
    //ended by shikang

//    if (SUCCEEDED(hr)){
	changed = 1;
//    }

    read_unlock_irqrestore(&pMixerObj->MixerLock, flags);
    
    // return 0;
    return changed;
}                                                                                                                                                                                                       static int
snd_ctalsa_put_capture_volume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    stHwObj *pHwObj = (stHwObj *)snd_kcontrol_chip(kcontrol);
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    enum CTALSA_MIXER_DEV tMixer = kcontrol->private_value;
    stLineGroup *pLineGroup = pMixerObj->pGroupArray[tMixer];
    unsigned long flags;
    int changed = 0;
    HRESULT hr = E_FAIL;
 
//    CTDPF("snd_ctalsa_put_capture_volume is called, mixer type = %d", tMixer);

    write_lock_irqsave(&pMixerObj->MixerLock, flags);

    // put the volume level directly into the current LineGroup
    pLineGroup->dwRecVol[0] = ucontrol->value.integer.value[0];
    pLineGroup->dwRecVol[1] = ucontrol->value.integer.value[1];
    
    // if this LineGroup is selected as a recording source, then set the pRecGroup right away
    if (TRUE == pLineGroup->bRecSelected) {
//	hr = ctdrv_LineGroup_SetVolume(pMixerObj->pRecGroup,pLineGroup->dwRecVol,tMixer);
	hr = ctdrv_LineGroup_SetVolume_record(pMixerObj->pRecGroup,pLineGroup->dwRecVol,tMixer);//updated by shikang 070108
    }
    nFirst = 1;
    write_unlock_irqrestore(&pMixerObj->MixerLock, flags);
    
    if (SUCCEEDED(hr))
        changed = 1;

    return changed;
}


// the kontrol table for creative alsa driver mixers
static stKontrolNode ctalsa_kontrol_table[NUM_CTALSA_MIXERS] =
{
    [MIXER_MASTER] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Master Playback Volume", MIXER_MASTER),
        .bSwhCtl = FALSE,
	//.bSwhCtl = TRUE,
        INIT_SWHCTL("Master Playback Switch", MIXER_MASTER),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_FRONT] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Front Playback Volume", MIXER_FRONT),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("Front Playback Switch", MIXER_FRONT),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_REAR] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Rear Playback Volume", MIXER_REAR),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("Rear Playback Switch", MIXER_REAR),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_CENTER] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Center Playback Volume", MIXER_CENTER),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("Center Playback Switch", MIXER_CENTER),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_REARCENTER] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Rear Center Playback Volume", MIXER_REARCENTER),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("Rear Center Playback Switch", MIXER_REARCENTER),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_LFE] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("LFE Playback Volume", MIXER_LFE),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("LFE Playback Switch", MIXER_LFE),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_SIDE] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Side Playback Volume", MIXER_SIDE),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("Side Playback Switch", MIXER_SIDE),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_LINE] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Line Playback Volume", MIXER_LINE),
        .bSwhCtl = TRUE,
        INIT_SWHCTL("Line Playback Switch", MIXER_LINE),
        .bRecCtl = TRUE,
//        .szRecCtlName = "Line Capture",
	.szRecCtlName = "Line",//updated by shikang
        INIT_RECCTL("Line Capture Volume", MIXER_LINE),
    },
    [MIXER_AUX] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Aux Playback Volume", MIXER_AUX),
        .bSwhCtl = TRUE,
        INIT_SWHCTL("Aux Playback Switch", MIXER_AUX),
        .bRecCtl = TRUE,
//        .szRecCtlName = "Aux Capture",
	.szRecCtlName = "Aux",//updated by shikang
        INIT_RECCTL("Aux Capture Volume", MIXER_AUX)
    },
    [MIXER_WAVE] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("PCM Playback Volume", MIXER_WAVE),
//	.bSwhCtl = FALSE,
        .bSwhCtl = TRUE,
        INIT_SWHCTL("PCM Playback Switch", MIXER_WAVE),
        .bRecCtl = TRUE,
//        .szRecCtlName = "PCM Capture",
	.szRecCtlName = "PCM",//updated by shikang
        INIT_RECCTL("PCM Capture Volume", MIXER_WAVE)
    },
    [MIXER_WAVE_FRONT] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("PCM Front Playback Volume", MIXER_WAVE_FRONT),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("PCM Front Playback Switch", MIXER_WAVE_FRONT),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_WAVE_REAR] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("PCM Rear Playback Volume", MIXER_WAVE_REAR),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("PCM Rear Playback Switch", MIXER_WAVE_REAR),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_WAVE_CENTER] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("PCM Center Playback Volume", MIXER_WAVE_CENTER),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("PCM Center Playback Switch", MIXER_WAVE_CENTER),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_WAVE_REARCENTER] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("PCM Rear Center Playback Volume", MIXER_WAVE_REARCENTER),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("PCM Rear Center Playback Switch", MIXER_WAVE_REARCENTER),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_WAVE_LFE] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("PCM LFE Playback Volume", MIXER_WAVE_LFE),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("PCM LFE Playback Switch", MIXER_WAVE_LFE),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_WAVE_SIDE] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("PCM Side Playback Volume", MIXER_WAVE_SIDE),
        .bSwhCtl = FALSE,
        INIT_SWHCTL("PCM Side Playback Switch", MIXER_WAVE_SIDE),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },    
    [MIXER_TONEBASS] = {
        .bVolCtl = TRUE,
//        INIT_VOLCTL("Tone Control - Bass", MIXER_TONEBASS),
	INIT_VOLCTL("Tone Bass Playback Volume",MIXER_TONEBASS),//modified by shikang
        .bSwhCtl = FALSE,
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_TONETREBLE] = {
        .bVolCtl = TRUE,
//        INIT_VOLCTL("Tone Control - Treble", MIXER_TONETREBLE),
	INIT_VOLCTL("Tone Treble Playback Volume",MIXER_TONETREBLE),//modified by shikang
        .bSwhCtl = FALSE,
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_MIC] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("Mic Playback Volume", MIXER_MIC),
        .bSwhCtl = TRUE,
        INIT_SWHCTL("Mic Playback Switch", MIXER_MIC),
        .bRecCtl = TRUE,
//        .szRecCtlName = "Mic Capture",
	.szRecCtlName = "Mic",//updated by shikang
        INIT_RECCTL("Mic Capture Volume", MIXER_MIC),
    },
    [MIXER_TAD] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("TAD Playback Volume", MIXER_TAD),
        .bSwhCtl = TRUE,
        INIT_SWHCTL("TAD Playback Switch", MIXER_TAD),
        .bRecCtl = TRUE,
	.szRecCtlName = "TAD",
        INIT_RECCTL("TAD Capture Volume", MIXER_TAD)
    },
    [MIXER_PCSPEAKER] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("PC Speaker Playback Volume", MIXER_PCSPEAKER),
        .bSwhCtl = TRUE,
        INIT_SWHCTL("PC Speaker Playback Switch", MIXER_PCSPEAKER),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_CD] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("CD Playback Volume", MIXER_CD),
        .bSwhCtl = TRUE,
        INIT_SWHCTL("CD Playback Switch", MIXER_CD),
        .bRecCtl = TRUE,
	.szRecCtlName = "CD",
        INIT_RECCTL("CD Capture Volume", MIXER_CD)
    },
    [MIXER_CDDIGITAL] = {
        .bVolCtl = TRUE,
        INIT_VOLCTL("CD Digital Playback Volume", MIXER_CDDIGITAL),
        .bSwhCtl = TRUE,
        INIT_SWHCTL("CD Digital Playback Switch", MIXER_CDDIGITAL),
        .bRecCtl = TRUE,
	.szRecCtlName = "CD Digital",
        INIT_RECCTL("CD Digital Capture Volume", MIXER_CDDIGITAL)
    },
    [MIXER_WATUHEAR] = {
        .bVolCtl = FALSE,
        .bSwhCtl = FALSE,
        .bRecCtl = TRUE,
        .szRecCtlName = "What U Hear",
        INIT_RECCTL("What U Hear Capture Volume", MIXER_WATUHEAR)
    },
    [MIXER_MICBOOST] = {
        .bVolCtl = FALSE,
        .bSwhCtl = TRUE,
        INIT_SWHCTL("Mic Boost (+20dB)", MIXER_MICBOOST),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_DIGITALOUT] = {
        .bVolCtl = FALSE,
        .bSwhCtl = TRUE,
//        INIT_SWHCTL("Digital Output Only Playback Switch", MIXER_DIGITALOUT),
	INIT_SWHCTL("DigitalOut Playback Switch", MIXER_DIGITALOUT),
        .bRecCtl = FALSE,
	.szRecCtlName = "" //added by shikang
    },
    [MIXER_ANALOGSTMIX] = {
        .bVolCtl = FALSE,
        .bSwhCtl = FALSE,
        .bRecCtl = TRUE,
        .szRecCtlName = "Analog Stereo Mix",
        INIT_RECCTL("Analog Stereo Mix Capture Volume", MIXER_ANALOGSTMIX)
    },
    [MIXER_ANALOGMIX] = {
        .bVolCtl = FALSE,
        .bSwhCtl = FALSE,
        .bRecCtl = TRUE,
        .szRecCtlName = "Analog Mix",
        INIT_RECCTL("Analog Mix Capture Volume", MIXER_ANALOGMIX)
    },
    [MIXER_DIGITALIO] = {
	.bVolCtl = FALSE,
	.bSwhCtl = TRUE,
	INIT_SWHCTL("DigitalIO Playback Switch",MIXER_DIGITALIO),
	.bRecCtl = FALSE,
	.szRecCtlName = ""
    },
    [MIXER_VIDEO] = {
	.bVolCtl = TRUE,
	INIT_VOLCTL("VIDEO Playback Volume", MIXER_VIDEO),
	.bSwhCtl = TRUE,
	INIT_SWHCTL("VIDEO Playback Switch",MIXER_VIDEO),
	.bRecCtl = TRUE,
	.szRecCtlName = "VIDEO",
	INIT_RECCTL("VIDEO Capture Volume", MIXER_VIDEO)
    },
    [MIXER_MICFP] = {
	.bVolCtl = TRUE,
	INIT_VOLCTL("MicFP Playback Volume", MIXER_MICFP),
	.bSwhCtl = TRUE,
	INIT_SWHCTL("MicFP Playback Switch",MIXER_MICFP),
	.bRecCtl = TRUE,
	.szRecCtlName = "MicFP",
	INIT_RECCTL("MicFP Capture Volume", MIXER_MICFP)
    },
    [MIXER_DIGITALIN] = {
	.bVolCtl = TRUE,
	INIT_VOLCTL("DigitalIn Playback Volume", MIXER_DIGITALIN),
	.bSwhCtl = TRUE,
	INIT_SWHCTL("DigitalIn Playback Switch", MIXER_DIGITALIN),
	.bRecCtl = TRUE,
	.szRecCtlName = "DigitalIn",
	INIT_RECCTL("DigitalIn Capture Volume", MIXER_DIGITALIN)
    },
    [MIXER_SPDIFIN] = {
	.bVolCtl = TRUE,
	INIT_VOLCTL("SPDIF-IN Playback Volume", MIXER_SPDIFIN),
	.bSwhCtl = TRUE,
	INIT_SWHCTL("SPDIF-IN Playback Switch", MIXER_SPDIFIN),
	.bRecCtl = TRUE,
	.szRecCtlName = "SPDIF-IN",
	INIT_RECCTL("SPDIF-IN Capture Volume", MIXER_SPDIFIN)
     }
};

//added by shikang
struct rec_sel{
	unsigned short mask;
	char **texts;
};

static struct rec_sel std_enum;
static char **ctalsa_rec_name = NULL;
static int n_rec = 0;

/*
#define XFI_ENUM_SINGLE(xmask, xtexts) \
{.mask = xmask, .texts = xtexts }
                                                                                                                  
static  struct rec_sel std_enum[] = {
XFI_ENUM_SINGLE(n_rec, ctalsa_rec_name),
};
*/

static void get_rec_sel(int *n_rec)
{
	int k,i,j = 0;
	ctalsa_rec_name = (char **)kmalloc(sizeof(char*) * (nNumOfRecGroupFound - 1),GFP_KERNEL);
	
	if (ctalsa_rec_name) {

		memset(ctalsa_rec_name, 0, sizeof(char*)*(nNumOfRecGroupFound - 1));
		for (k = 0;k < nNumOfRecGroupFound; k++) {
			i = ctalsa_rec_num[k];
			if ((ctalsa_kontrol_table[i].bRecCtl == TRUE)&&(i != NUM_CTALSA_MIXERS)) {
				ctalsa_rec_name[j] = ctalsa_kontrol_table[i].szRecCtlName;//get names of capture strips;
				j = j + 1;
			}
		}
		
		*n_rec = j;
		std_enum.mask = j;
		std_enum.texts = ctalsa_rec_name;
		
		for ( i = 0; i < *n_rec; i++){
			CTDPF("ctalsa_rec_name[%d] is %s,ok\n",i,ctalsa_rec_name[i]);
		}
	}
}

static int
snd_ctalsa_info_record_mux(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct rec_sel *e = (struct rec_sel*)kcontrol->private_value;
	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->count = 1;
	uinfo->value.enumerated.items = e->mask;
	if (uinfo->value.enumerated.item > e->mask - 1)
		uinfo->value.enumerated.item = e->mask - 1;
	strcpy(uinfo->value.enumerated.name, e->texts[uinfo->value.enumerated.item]);

	return 0;
}
static int
snd_ctalsa_get_record_mux(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	stHwObj *pHwObj = (stHwObj *)snd_kcontrol_chip(kcontrol);
	stMixerObj *pMixerObj = pHwObj->pMixerObj;
	enum CTALSA_MIXER_DEV tMixer;
	int i;


	if (S_OK == ctdrv_LineGroup_GetRecorded(pMixerObj,&tMixer)) {
		for (i = 0; i < n_rec; i++){
			if (!strcmp(ctalsa_kontrol_table[tMixer].szRecCtlName, ctalsa_rec_name[i])) {
				break;
			}
		}
		if (i < n_rec) {
			ucontrol->value.enumerated.item[0] = i;
		}
	} else {
		ucontrol->value.enumerated.item[0] = 0;
		CTASSERT(0);
	}
	return 0;
}
static int
snd_ctalsa_put_record_mux(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	stHwObj *pHwObj = (stHwObj *)snd_kcontrol_chip(kcontrol);
	stMixerObj *pMixerObj = pHwObj->pMixerObj;
	enum CTALSA_MIXER_DEV tMixer;
	enum CTALSA_MIXER_DEV tMixer1;
	int i=0,index,changed = 0;
	HRESULT hr = E_FAIL;
	stLineGroup *pLineGroup;

	if (FAILED(ctdrv_LineGroup_GetRecorded(pMixerObj,&tMixer))) {
		return 0;
	}

	for (i = 0; i < n_rec; i++){
		if (!strcmp(ctalsa_kontrol_table[tMixer].szRecCtlName, ctalsa_rec_name[i])) {
			break;
		}
	}

	if (i == n_rec) {
		return 0;
	}

	if (i == ucontrol->value.enumerated.item[0]) {
		pLineGroup = pMixerObj->pGroupArray[tMixer];
		if (S_OK == ctdrv_LineGroup_SetRecorded(pMixerObj,pLineGroup)){
			pMixerObj->pGroupArray[tMixer]->bRecSelected = TRUE;
			hr = ctdrv_LineGroup_SetVolume_record(pMixerObj->pRecGroup, pMixerObj->pGroupArray[tMixer]->dwRecVol, tMixer);
		}
		if (SUCCEEDED(hr))
			changed = 1;
		return changed;
	}

	index = ucontrol->value.enumerated.item[0];
	for (tMixer1 = 0; tMixer1 < NUM_CTALSA_MIXERS; tMixer1++) {
		if(!strcmp(ctalsa_rec_name[index],ctalsa_kontrol_table[tMixer1].szRecCtlName))
			break;
	}

	if (NUM_CTALSA_MIXERS == tMixer1) {
		return 0;
	}

	pLineGroup = pMixerObj->pGroupArray[tMixer1];	
	if (S_OK == ctdrv_LineGroup_SetRecorded(pMixerObj,pLineGroup)) {
		// Some mixerline is not exposed in ALSA (e.g. Aux-In). Thus, need to check
		// for validity of pGroupArray[tMixer] first.
		if(pMixerObj->pGroupArray[tMixer]) {
			pMixerObj->pGroupArray[tMixer]->bRecSelected = FALSE;
		}
		pMixerObj->pGroupArray[tMixer1]->bRecSelected = TRUE;
		hr = ctdrv_LineGroup_SetVolume_record(pMixerObj->pRecGroup, pMixerObj->pGroupArray[tMixer1]->dwRecVol, tMixer1);
	}
	if (SUCCEEDED(hr))
		changed = 1;
	
	return changed;
}

static snd_kcontrol_new_t snd_ctalsa_recordstrip_control = {
                                        .iface = SNDRV_CTL_ELEM_IFACE_MIXER,
                                        .name = "Capture Source",
                                        .index = 0,
                                        .access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
                                        .info = snd_ctalsa_info_record_mux,
                                        .get = snd_ctalsa_get_record_mux,
                                        .put = snd_ctalsa_put_record_mux,
                                        .private_value = (unsigned long)&std_enum
                                    };

//ended by shikang

// Capture Source opts
/*static int
snd_ctalsa_info_capture_mux(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
    uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
    uinfo->count = 1;
    
    uinfo->value.enumerated.items = NUM_CTALSA_MIXERS;
    if (uinfo->value.enumerated.item > (NUM_CTALSA_MIXERS-1))
        uinfo->value.enumerated.item = (NUM_CTALSA_MIXERS-1);
    if (TRUE == ctalsa_kontrol_table[uinfo->value.enumerated.item].bRecCtl)
        strcpy(uinfo->value.enumerated.name, ctalsa_kontrol_table[uinfo->value.enumerated.item].szRecCtlName);
    else
        strcpy(uinfo->value.enumerated.name, "not supported");
    return 0;
}

static int
snd_ctalsa_get_capture_mux(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    stHwObj *pHwObj = snd_kcontrol_chip(kcontrol);
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    enum CTALSA_MIXER_DEV tMixer;

    if (S_OK == ctdrv_LineGroup_GetRecorded(pMixerObj, &tMixer)) {
        ucontrol->value.enumerated.item[0] = tMixer;
    } else {
        ucontrol->value.enumerated.item[0] = 0;
        CTASSERT(0);
    }
    CTDPF("snd_ctalsa_get_capture_mux is called,tMixer is %d\n",tMixer);
    CTDPF("ucontrol->value.enumerated.item[0] is %d\n",ucontrol->value.enumerated.item[0]);
    return 0;
}

static int
snd_ctalsa_put_capture_mux(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    stHwObj *pHwObj = snd_kcontrol_chip(kcontrol);
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    enum CTALSA_MIXER_DEV tMixer, index;
    int changed = 0;
    HRESULT hr = E_FAIL;

    ctdrv_LineGroup_GetRecorded(pMixerObj, &tMixer);

    CTDPF("snd_ctalsa_put_capture_mux is called,tMixer is %d\n",tMixer);
    CTDPF("put tMixer is %d\n",tMixer);
  
    if (tMixer == ucontrol->value.enumerated.item[0]) {
        // this is the recording source already
        changed = 1;
        return changed;
    }

    // set the new recording source
    index = ucontrol->value.enumerated.item[0];
    if (S_OK == ctdrv_LineGroup_SetRecorded(pMixerObj->pGroupArray[index])) {
        pMixerObj->pGroupArray[tMixer]->bRecSelected = FALSE;
        pMixerObj->pGroupArray[index]->bRecSelected = TRUE;
        // set the recording volume
        hr = ctdrv_LineGroup_SetVolume(pMixerObj->pRecGroup, pMixerObj->pGroupArray[index]->dwRecVol, tMixer);
    }

    if (SUCCEEDED(hr))
        changed = 1;

    return changed;
}

static snd_kcontrol_new_t snd_ctalsa_capturestrip_control = {
    .iface = SNDRV_CTL_ELEM_IFACE_MIXER,
    .name = "Capture Source",
    .info = snd_ctalsa_info_capture_mux,
    .get = snd_ctalsa_get_capture_mux,
    .put = snd_ctalsa_put_capture_mux
};
*/
#if defined(UDA_ENABLE_IEC958_CONTROL)
static int
snd_ctalsa_spdif_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t * uinfo)
{
    uinfo->type = SNDRV_CTL_ELEM_TYPE_IEC958;
    uinfo->count = 1;

    return 0;
}

static int
snd_ctalsa_spdif_get(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    //TODO:
    return 0;
}

static int
snd_ctalsa_spdif_get_mask(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    //TODO:
    ucontrol->value.iec958.status[0] = 0xff;
    ucontrol->value.iec958.status[1] = 0xff;
    ucontrol->value.iec958.status[2] = 0xff;
    ucontrol->value.iec958.status[3] = 0xff;

    return 0;
}

static int
snd_ctalsa_spdif_put(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
    //TODO:
    return 1;
}

static snd_kcontrol_new_t snd_ctalsa_spdif_mask_control = {
    .access = SNDRV_CTL_ELEM_ACCESS_READ,
    .iface  = SNDRV_CTL_ELEM_IFACE_MIXER,
    .name   = SNDRV_CTL_NAME_IEC958("", PLAYBACK, MASK),
    .count  = 4,
    .info   = snd_ctalsa_spdif_info,
    .get    = snd_ctalsa_spdif_get_mask
};

static snd_kcontrol_new_t snd_ctalsa_spdif_control = {
    .iface  = SNDRV_CTL_ELEM_IFACE_MIXER,
    .name   = SNDRV_CTL_NAME_IEC958("", PLAYBACK, DEFAULT),
    .count  = 4,
    .info   = snd_ctalsa_spdif_info,
    .get    = snd_ctalsa_spdif_get,
    .put    = snd_ctalsa_spdif_put
};
#endif //UDA_ENABLE_IEC958_CONTORL


inline int
ctalsa_ctl_new(const snd_kcontrol_new_t *pCtlNew, stHwObj *pHwObj)
{
    int err;
    stKontrolList *kontrol;
    snd_kcontrol_t *pctl;
    snd_card_t *pCard = pHwObj->pCard;
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    stKontrolList *ctalsa_ctl_list = pMixerObj->pKontrolList;

    pctl = snd_ctl_new1(pCtlNew, pHwObj);
    if (!pctl) return -ENOMEM;
    if ((err = snd_ctl_add(pCard, pctl)) < 0) {
        snd_ctl_free_one(pctl);
        return err;
    }
    kontrol = kmalloc(sizeof(stKontrolList), GFP_KERNEL);
    if (!kontrol) return -ENOMEM;
    kontrol->pKctl = pctl;
    list_add(&(kontrol->list), &(ctalsa_ctl_list->list));

    return 0;
}

static int
ctalsa_build_mixer(stHwObj *pHwObj)
{
    int err;
    enum CTALSA_MIXER_DEV tMixer;
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    stKontrolList *ctalsa_ctl_list = pMixerObj->pKontrolList;

//    CTDPF("ctalsa_build_mixer is called");

    pMixerObj->pKontrolList = kmalloc(sizeof(stKontrolList), GFP_KERNEL);
    if (!pMixerObj->pKontrolList) return -ENOMEM;

    ctalsa_ctl_list = pMixerObj->pKontrolList;

    INIT_LIST_HEAD(&(ctalsa_ctl_list->list));

    for (tMixer = 0; tMixer < NUM_CTALSA_MIXERS; tMixer++) {
        if (NULL == pMixerObj->pGroupArray[tMixer])
            continue;
        if (TRUE == ctalsa_kontrol_table[tMixer].bVolCtl) { // volume
            if ((err = ctalsa_ctl_new(&ctalsa_kontrol_table[tMixer].VolCtl, pHwObj)) < 0)
                return err;
        }
        if (TRUE == ctalsa_kontrol_table[tMixer].bSwhCtl) { // switch
            if ((err = ctalsa_ctl_new(&ctalsa_kontrol_table[tMixer].SwhCtl, pHwObj)) < 0)
                return err;
        }
        if (TRUE == ctalsa_kontrol_table[tMixer].bRecCtl) { // record
            if ((err = ctalsa_ctl_new(&ctalsa_kontrol_table[tMixer].RecCtl, pHwObj)) < 0)
                return err;
        }
    }
    //added by shikang
    //for record mux device
    if (NULL != pMixerObj->pRecGroup) {
	if ((err = ctalsa_ctl_new(&snd_ctalsa_recordstrip_control, pHwObj)) < 0)
		return err;
    }
    //ended by shikang

    /*
    // for record mux device
    if (NULL != pMixerObj->pRecGroup) {
        if ((err = ctalsa_ctl_new(&snd_ctalsa_capturestrip_control, pHwObj)) < 0)
            return err;
    }
    */

#if defined(UDA_ENABLE_IEC958_CONTROL)
    if ((err = ctalsa_ctl_new(&snd_ctalsa_spdif_mask_control, pHwObj)) < 0)
        return err;

    if ((err = ctalsa_ctl_new(&snd_ctalsa_spdif_control, pHwObj)) < 0)
        return err;
#endif

    return 0;
}

static void
ctalsa_remove_mixer(stHwObj *pHwObj)
{
    snd_card_t *pCard = pHwObj->pCard;
    stMixerObj *pMixerObj = pHwObj->pMixerObj;
    stKontrolList *temp, *ctalsa_ctl_list = pMixerObj->pKontrolList;
    struct list_head *pos, *q;

    list_for_each_safe(pos, q, &(ctalsa_ctl_list->list)) {
        temp = list_entry(pos, stKontrolList, list);
        list_del(pos);
        snd_ctl_remove(pCard, temp->pKctl);
        kfree(temp);
    }
    kfree(pMixerObj->pKontrolList);
}

int __devinit
mixer_init(stHwObj *pHwObj, char *szName)
{
    CTDWORD res;
    HRESULT hr = E_FAIL;
    stMixerObj *pMixerObj;
    PCTVOID pIhaBMixer = pHwObj->apIha[BMIXER_DEV];

//    CTDPF("mixer_init is called");
    if (!pIhaBMixer) {
        CTDPF("ERROR: cannot find bmixer device");
        return -EINVAL;
    }

    if (pHwObj->pMixerObj != NULL) {
        CTDPF("mixer has been initialized already");
        return 0;
    }

    pMixerObj = (stMixerObj *)kmalloc(sizeof(stMixerObj), GFP_KERNEL);
    if (!pMixerObj)
        return -ENOMEM;
    memset(pMixerObj, 0, sizeof(stMixerObj));
    pMixerObj->MixerLock = RW_LOCK_UNLOCKED;

    pMixerObj->pIhaBMixer = pIhaBMixer;

    pHwObj->pMixerObj = pMixerObj;

    hr = ctdrv_getszRecName(pMixerObj,ctalsa_rec_num,&nNumOfRecGroupFound);
    CTDPF("nNumOfRecGroupFound = %d\n",nNumOfRecGroupFound);
    get_rec_sel(&n_rec); 
    CTDPF(" n_rec = %d\n",n_rec);
    hr = ctdrv_LineGroup_Build(pMixerObj);
    CTASSERT(SUCCEEDED(hr));

    res = ctalsa_build_mixer(pHwObj);
    CTASSERT(res >= 0);

    strcpy(((snd_card_t *)(pHwObj->pCard))->mixername, "CTALSA Mixer");

    return 0;
}

void __devexit
mixer_exit(stHwObj *pHwObj, int nDev)
{
    stMixerObj *pMixerObj = pHwObj->pMixerObj;

//    CTDPF("mixer_exit is called");

    if (pMixerObj == NULL) {
        CTDPF("there is no mixer to uninitialize!");
        return;
    }

    ctalsa_remove_mixer(pHwObj);

    ctdrv_LineGroup_Free(pMixerObj);

    kfree(pMixerObj);
    pHwObj->pMixerObj = NULL;

    if(ctalsa_rec_name) {
        kfree(ctalsa_rec_name);
        ctalsa_rec_name = NULL;
    }
	
    if (pHwObj->anDev[BMIXER_DEV] < 0)
        return;
}

