/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/
#define EXPORT_SYMTAB
#include <sound/driver.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/moduleparam.h>
#include <linux/sound.h>
#include <linux/major.h>
#include <sound/core.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/info.h>
#include <sound/rawmidi.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#	include <linux/devfs_fs_kernel.h>
#endif
#include <linux/smp.h>

#include "ctalsa.h"

MODULE_AUTHOR("Copyright (C) 1999-2008 Creative Technology Ltd");
MODULE_DESCRIPTION("CTALSA driver version 1.18.0000");
MODULE_LICENSE("Creative Proprietary License");
MODULE_SUPPORTED_DEVICE("{{Creative Labs, Sound Blaster X-Fi}");

// module parameters (see "Module Parameters")
static int index[SNDRV_CARDS] = SNDRV_DEFAULT_IDX;
static char *id[SNDRV_CARDS] = SNDRV_DEFAULT_STR;
static int enable[SNDRV_CARDS] = SNDRV_DEFAULT_ENABLE_PNP;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)
#	define arg(num) num 
#else
#	define arg(num) &num
#endif

static int boot_devs;
module_param_array(index, int, arg(boot_devs), 0444);
MODULE_PARM_DESC(index, "Index value for the ctalsa enabled soundcard.");
module_param_array(id, charp, arg(boot_devs), 0444);
MODULE_PARM_DESC(id, "ID string for the ctalsa enabled soundcard.");
module_param_array(enable, bool, arg(boot_devs), 0444);
MODULE_PARM_DESC(enable, "Enable the ctalsa enabled soundcard.");

// PCI IDs
static struct pci_device_id ctalsa_pci_dev_ids[] = {
#if 0
    { PCI_VENDOR_CREATIVE, PCI_DEVICE_CREATIVE_EMU10K1, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
    { PCI_VENDOR_CREATIVE, PCI_DEVICE_CREATIVE_EMU10K2, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
    { PCI_VENDOR_CREATIVE, PCI_DEVICE_CREATIVE_TINA,    PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
#endif
    // only X-Fi is supported, so...
    { PCI_VENDOR_CREATIVE, PCI_DEVICE_CREATIVE_CTXFI,   PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
    { 0, }
};
MODULE_DEVICE_TABLE(pci, ctalsa_pci_dev_ids);

static struct {
    FPDEVINIT pfnInit;
    FPDEVEXIT pfnExit;
    char szPublicName[16];
} gFuncTable[NUM_UDA_DEVS] =
{
    [WAVEIN_DEV]    = { .pfnInit = wavein_init,        .pfnExit = wavein_exit,        .szPublicName = "WaveIn"   },
    [WAVEOUT1_DEV]  = { .pfnInit = envaudio_init,      .pfnExit = envaudio_exit,      .szPublicName = "EnvAudio" },
    [WAVEOUT2_DEV]  = { .pfnInit = waveout_init,       .pfnExit = waveout_exit,       .szPublicName = "WaveOut/WaveIn"  },
    [BMIXER_DEV]    = { .pfnInit = mixer_init,         .pfnExit = mixer_exit,         .szPublicName = "Mixer"    },
    [UARTIN_DEV]    = { .pfnInit = NULL,               .pfnExit = NULL,               .szPublicName = ""         },
    [UARTOUT_DEV]   = { .pfnInit = NULL,               .pfnExit = NULL,               .szPublicName = ""         },
    [HWSYNTH_DEV]   = { .pfnInit = NULL,               .pfnExit = NULL,               .szPublicName = ""         },
    [AUTOIN_DEV]    = { .pfnInit = NULL,               .pfnExit = NULL,               .szPublicName = ""         },
    [AUTOOUT_DEV]   = { .pfnInit = NULL,               .pfnExit = NULL,               .szPublicName = ""         },
    [TIMER_DEV]     = { .pfnInit = NULL,               .pfnExit = NULL,               .szPublicName = "Timer"    },
    [SBLIVE_DEV]    = { .pfnInit = NULL,               .pfnExit = NULL,               .szPublicName = "Product"  },
    [AC3MGR_DEV]    = { .pfnInit = ac3mgr_init,        .pfnExit = ac3mgr_exit,        .szPublicName = "AC3Mgr"   }
};

static int s_nDevDispatch = -1; // This must be initialized to -1
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#   ifdef DEVFS_ENABLED
static devfs_handle_t g_dev_emu10kx;
#   endif
CT_EXPORT_NO_SYMBOLS;
#endif


// Core functions here...
int ctUAACoreChange(struct pci_dev *pDev);
int ctIsHendrix(CTWORD wDevID, CTWORD wSubSysID);


/////////////////////////
// Driver dispatcher fops
/////////////////////////
static int
ctalsa_dispatch_ioctl(struct inode *pInode, struct file *pFile, unsigned int uCmd, unsigned long ulArg)
{
    CTDPF("ctalsa_dispatch_ioctl(): cmd = 0x%x ", uCmd);

    switch (uCmd) {
        case INIT_DRIVER:
            //return ctalsa_init();   // FIXME:

        case EXIT_DRIVER:
            //ctalsa_exit();          // FIXME:
            return 0;

        case RESET_COUNT:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
            if (MOD_IN_USE)
                MOD_DEC_USE_COUNT;
#endif
            return 0;

        default:
            return SUCCEEDED(ioctl_dispatch(ulArg, ctdrv_dispatch)) ? 0 : -EINVAL;
    }
}

CTHANDLE g_hLastBuffer = NULL;

// FIXME: according to beijing team's code
static int
ctalsa_dispatch_mmap(struct file *file, struct vm_area_struct *vma)
{
    CTDWORD dwSize = vma->vm_end - vma->vm_start;
    OSPHYSMEMHDL hBuffer;

    PDEBUG("ctalsa_dispatch_mmap() called, dwSize=0x%08x\n", dwSize);

    hBuffer = g_hLastBuffer; // Ugly hack
    if (!hBuffer)
        CTRETURN(-EINVAL);

    if (vma->vm_flags & VM_WRITE || vma->vm_flags & VM_READ) {
        if (dwSize > PAGE_ALIGN(((stMemHandle *) hBuffer)->dwPageRange))
            CTRETURN(-EINVAL);
        //if (FAILED(reserve_remap_pages(vma->vm_start, (stMemHandle *) hBuffer, dwSize, vma->vm_page_prot)))
        //    CTRETURN(-EAGAIN);
    } else {
        CTRETURN(-EINVAL);
    }

    return 0;
}

static int
ctalsa_dispatch_open(struct inode *pInode, struct file *pFile)
{
    CTDPF("ctalsa_dispatch_open is called");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    CT_MOD_INC_USE_COUNT;
#endif

    return 0;
}

static int
ctalsa_dispatch_release(struct inode *pInode, struct file *pFile)
{
    CTDPF("ctalsa_dispatch_release is called");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    CT_MOD_DEC_USE_COUNT;
#endif

    return 0;
}

struct file_operations ctalsa_dispatch_fops = {
    ioctl:   ctalsa_dispatch_ioctl,
    mmap:    ctalsa_dispatch_mmap,
    open:    ctalsa_dispatch_open,
    release: ctalsa_dispatch_release,
};

void
register_ctalsa_dispatcher(void)
{
    if (s_nDevDispatch < 0) {
        s_nDevDispatch = register_sound_special(&ctalsa_dispatch_fops, EMU10KX_MINOR);
#ifdef DEVFS_ENABLED
#   if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
        g_dev_emu10kx = devfs_register(NULL, "x-fi", DEVFS_FL_DEFAULT, SOUND_MAJOR, EMU10KX_MINOR,
                S_IRUSR | S_IWUSR | S_IFCHR, &ctalsa_dispatch_fops, NULL);
#   else
        devfs_mk_cdev(MKDEV(SOUND_MAJOR, EMU10KX_MINOR), S_IRUSR|S_IWUSR|S_IFCHR, "x-fi");
#   endif
#endif
    }
}

void
unregister_ctalsa_dispatcher(void)
{
    if (s_nDevDispatch >= 0) {
        unregister_sound_special(s_nDevDispatch);
#ifdef DEVFS_ENABLED
#   if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
        devfs_unregister(g_dev_emu10kx);
#   else
        devfs_remove("x-fi");
#   endif
#endif
    }

}


/////////////////////////////////////////
static int
ctalsa_HwObj_free(stHwObj *pHwObj)
{
    if (pHwObj->pJoyDev) {
        // Restore original joystick base address
        pci_write_config_dword(pHwObj->pJoyDev, PCI_BASE_ADDRESS_0, pHwObj->dwJoyBase);
    }

    // call the destructor for this HwObj
    ctdrv_HwObj_dtor(pHwObj);

    // disable the device
    pci_disable_device(pHwObj->pPCIDev);

    // free the allocated memory for this HwObj
    kfree(pHwObj);

    return 0;
}

static int
ctalsa_HwObj_dev_free(snd_device_t *device)
{
    stHwObj *pHwObj = device->device_data;

    return ctalsa_HwObj_free(pHwObj);
}

void *
map_io_space(unsigned long mem_addr, unsigned long mem_size)
{
    unsigned long offset, last_addr, size;

    if (check_mem_region(mem_addr, mem_size)) {
        printk("<1>ctxficore: memory already in use\n");
        return NULL;
    }
    request_mem_region(mem_addr, mem_size, "ctxficore");

    last_addr = mem_addr + mem_size;
    offset = mem_addr & ~PAGE_MASK;
    mem_addr &= PAGE_MASK;
    size = PAGE_ALIGN(last_addr) - mem_addr;

    return ioremap(mem_addr, size) + offset;
}

static stHwObj *
ctalsa_HwObj_new(snd_card_t *pCard, struct pci_dev *pDev)
{
    int i;
    HRESULT hr;
    stHwObj *pHwObj;
    CTDWORD dwIOBase=0, dwDMAMask;
    static snd_device_ops_t ops = {
        .dev_free = ctalsa_HwObj_dev_free,
    };

    pHwObj = kmalloc(sizeof(stHwObj), GFP_KERNEL);
    if (NULL == pHwObj) return NULL;
    memset(pHwObj, 0, sizeof(stHwObj));

    pHwObj->pCard = pCard;

    pHwObj->pOSSrv = get_ossrv();
    if (!pHwObj->pOSSrv) {
        CTDPF("Error: init_module() encountered invalid OSSrv pointer\n");
        return NULL;
    }
    CTDPF("Got an OSSrv pointer: 0x%p\n", pHwObj->pOSSrv);

    if ((pci_enable_device(pDev)) < 0) {
        kfree(pHwObj);
        return NULL;
    }

    pci_set_master(pDev);

    //
    // Get card info
    //
    pci_read_config_word(pDev, PCI_VENDOR_ID   , &pHwObj->wVendorID);
    pci_read_config_word(pDev, PCI_DEVICE_ID   , &pHwObj->wDeviceID);
    pci_read_config_word(pDev, PCI_SUBSYSTEM_ID, &pHwObj->wSubsysID);
    pci_read_config_word(pDev, PCI_REVISION_ID , &pHwObj->wRevision);
    CTDPF("VendorID : %x, DeviceID : %x, SubSystemID : %x", pHwObj->wVendorID, pHwObj->wDeviceID, pHwObj->wSubsysID);

    //
    // Hendrix card detection
    //
    if( ctIsHendrix(pHwObj->wDeviceID, pHwObj->wSubsysID) )
    {
        ctUAACoreChange(pDev);

        // Search for BAR0 describing I/O space
        for( i = 0 ; i < DEVICE_COUNT_RESOURCE ; i++ )
        {
            if( RSRCISIOREGION(pDev, i) )
            {
                // Reserve I/O region, etc
                dwIOBase = RSRCADDRESS(pDev, i) & PCI_BASE_ADDRESS_IO_MASK;
                CTDPF("[ResourceIndex:%d] I/O region reserved : 0x%x", i, dwIOBase);
                break;
            }
        }
    }
    else
    {
        // do a little sanity test - does configuration register: BAR0 describe an I/O space?
        if (!RSRCISIOREGION(pDev, 0)) {
            CTDPF("Invalid base I/O address");
            CTASSERT(0);
            pci_disable_device(pDev);
            kfree(pHwObj);
            return NULL;
        }

        // Reserve I/O region, etc.
        dwIOBase = RSRCADDRESS(pDev, 0) & PCI_BASE_ADDRESS_IO_MASK;
    }

    dwDMAMask = (pHwObj->tCard == SBLIVE_CARD) ? EMU10K1_DMA_MASK : AUDIGY_DMA_MASK;
    if (pci_set_dma_mask(pDev, dwDMAMask) < 0 || pci_set_consistent_dma_mask(pDev, dwDMAMask) < 0) {
        CTDPF("error to set PCI mask DMA");
        CTASSERT(0);
        pci_disable_device(pDev);
        kfree(pHwObj);
        return NULL;
    }

    for (i = 0; i < DEVICE_COUNT_RESOURCE; i++) {
        // is this a copy of a boot ROM ?
        // IOW, is this resource memory (0x200) & are the bus-specific bits = IORESOURCE_ROM_COPY ?
        if (pci_resource_flags(pDev, i) != 0x204) {
            pHwObj->aMemWin[i].bValid = FALSE;      // nope, flag our H/W obj's assoc'd memory window
            continue;
        }

        pHwObj->aMemWin[i].dwSize = pci_resource_end(pDev, i) - pci_resource_start(pDev, i);
        pHwObj->aMemWin[i].dwStart = pci_resource_start(pDev, i);
        pHwObj->aMemWin[i].pStart = map_io_space(pci_resource_start(pDev, i), pHwObj->aMemWin[i].dwSize);
        pHwObj->aMemWin[i].bValid = TRUE;
    }

#ifdef DEBUG
    int te;
    for (te = 0; te < DEVICE_COUNT_RESOURCE; te++)
        CTDPF("name=%s, start=%lx, end=%lx, flags=%lx, mapped=%lx.\n", pDev->resource[te].name,
                    pDev->resource[te].start, pDev->resource[te].end, pDev->resource[te].flags,
                    (unsigned long)pHwObj->aMemWin[te].pStart);
#endif

    // there is no joystick port on X-Fi
#if 0
    struct pci_dev *pJoyDev;

    ///////////////////////////////////
    // Do joystick stuff
    pJoyDev = pci_find_slot(pDev->bus->number, pDev->devfn+JOY_FUNCTION);

    if (pJoyDev) {
        CTDWORD dwJoyBase = JOY_BASE;

        // Save original base address
        pHwObj->dwJoyBase = RSRCADDRESS(pJoyDev, 0) & PCI_BASE_ADDRESS_IO_MASK;

        // Check requested joystick region
        if (!check_region(dwJoyBase, JOY_EXTENT)) {
            pHwObj->pJoyDev = pJoyDev;
            // Write new joystick base address to PCI configuration space
            pci_write_config_dword(pJoyDev, PCI_BASE_ADDRESS_0, dwJoyBase);
        }
        CTDPF("Joystick.: IO Base = %#x, IO Base new=%#x IRQ = %d", (unsigned int)pHwObj->dwJoyBase,
                    dwJoyBase, pJoyDev->irq);
    }
#endif

    pHwObj->pPCIDev = pDev;
    pHwObj->dwIOBase = dwIOBase;
    pHwObj->dwIRQ = (CTDWORD)pDev->irq;
    CTDPF("Card enum.: Device found: IO Base = %#x, IRQ = %d, dwIRQ = %d", (unsigned int)dwIOBase, pDev->irq,
                    pHwObj->dwIRQ);

    // initialize the current hardware object
    hr = ctdrv_HwObj_ctor(pHwObj);
    if (FAILED(hr)) {
        CTDPF("error when the contructor of HwObj is called");
        CTASSERT(0);
        pci_disable_device(pDev);
        kfree(pHwObj);
        return NULL;
    }

    strcpy(pCard->id, pHwObj->szModel);
    strcpy(pCard->driver, pHwObj->szModel);
    sprintf(pCard->shortname, "%s [%lx]", pHwObj->szCard, (unsigned long)pHwObj->dwIOBase);
    sprintf(pCard->longname, "%s at 0x%lx, irq %i", pHwObj->szCard, (unsigned long)pHwObj->dwIOBase, pHwObj->dwIRQ);
    
    if ((snd_device_new(pCard, SNDRV_DEV_LOWLEVEL, pHwObj, &ops)) < 0) {
        ctalsa_HwObj_free(pHwObj);
        return NULL;
    }

    return pHwObj;
}

static int __devinit
snd_ctalsa_probe(struct pci_dev *pci, const struct pci_device_id *pci_id)
{
    snd_card_t *pCard;
    enum UDADEVS tUdaDev;
    char szDevName[64];
    stHwObj *pHwObj;
    static int dev = 0;
    int err;

    if (dev >= SNDRV_CARDS)
        return -ENODEV;
    if (!enable[dev]) {
        dev++;
        return -ENOENT;
    }

    pCard = snd_card_new(index[dev], id[dev], THIS_MODULE, 0);
    if (pCard == NULL)
        return -ENOMEM;

    pHwObj = ctalsa_HwObj_new(pCard, pci);
    if (NULL == pHwObj) {
        snd_card_free(pCard);
        return -ENOMEM;
    }

    pCard->private_data = pHwObj;

    // to initialize the different uda device objects
    CTDPF("Initializing devices\n");
    for (tUdaDev = 0; tUdaDev < NUM_UDA_DEVS; tUdaDev++) {
        if ((NULL != pHwObj->apIha[tUdaDev]) && (NULL != gFuncTable[tUdaDev].pfnInit)) {
            sprintf(szDevName, SZ_DEVNAME_FORMAT, gFuncTable[tUdaDev].szPublicName);
            if ((pHwObj->anDev[tUdaDev] = gFuncTable[tUdaDev].pfnInit(pHwObj, szDevName)) < 0) {
                err = pHwObj->anDev[tUdaDev];
                printk(KERN_ERR "fail to start device %s.\n", szDevName);
                snd_card_free(pCard);
                return err;
            }
        } else {
            pHwObj->anDev[tUdaDev] = -1;
        }
    }
    CTDPF("Initializing devices done\n");
    
    if ((err = snd_card_register(pCard)) < 0) {
        snd_card_free(pCard);
        return err;
    }

    pci_set_drvdata(pci, pCard);
    dev++;

    return 0;
}

static void __devexit
snd_ctalsa_remove(struct pci_dev *pci)
{
    //CTDPF("ctalsa_remove ");

    snd_card_t *pCard = pci_get_drvdata(pci);
    stHwObj *pHwObj = pCard->private_data;
    enum UDADEVS tUdaDev;
    
    // to uninitialize the different uda device objects
    for (tUdaDev = 0; tUdaDev < NUM_UDA_DEVS; tUdaDev++) {
        if ((NULL != pHwObj->apIha[tUdaDev]) && (NULL != gFuncTable[tUdaDev].pfnExit)) {
            gFuncTable[tUdaDev].pfnExit(pHwObj, pHwObj->anDev[tUdaDev]);
        }
    }

    snd_card_free(pCard);
    pci_set_drvdata(pci, NULL);

    // Unload I/O memory from kernel address space and release I/O region
    int i;
    for (i = 0; i < DEVICE_COUNT_RESOURCE; i++) {
        if (pHwObj->aMemWin[i].bValid == FALSE)
            continue;
        
        if (check_mem_region(pci_resource_start(pci, i), pHwObj->aMemWin[i].dwSize) != 0) {
            if (pHwObj->aMemWin[i].pStart) {
                //CTDPF("ctalsa_remove - iounmap : 0x%p",pHwObj->aMemWin[i].pStart);
                iounmap(pHwObj->aMemWin[i].pStart); 
            }
            release_mem_region( pci_resource_start(pci, i), pHwObj->aMemWin[i].dwSize);
        }
    }
}

// pci_driver definition
static struct pci_driver ctalsa_pci_driver = {
    .name       = "CTALSA",
    .id_table   = ctalsa_pci_dev_ids,
    .probe      = snd_ctalsa_probe,
    .remove     = __devexit_p(snd_ctalsa_remove),
};

// initialization of the module
static int __init
ctalsa_init(void)
{
    int err;

    CTDPF("ctalsa_init begin...");

    __do_global_ctors_aux();

    // NOTE : Redhat returns num of pci devices which were claimed by
    //        the driver during registration. Negative value on error,
    //        otherwise 0.
    //if ((err = pci_register_driver(&ctalsa_pci_driver)) > 0)
    if ((err = pci_register_driver(&ctalsa_pci_driver)) < 0)
    {
        CTDPF("pci_register_driver failed.  err=0x%x", err);    
        return err;
    }

    register_ctalsa_dispatcher();

    CTDPF("ctalsa_init end...");

    return 0;
}

// clean up the module
static void __exit
ctalsa_exit(void)
{
    CTDPF("ctalsa_exit begin...");

    unregister_ctalsa_dispatcher();

    pci_unregister_driver(&ctalsa_pci_driver);

    CTDPF("ctalsa_exit end...");
}

module_init(ctalsa_init);
module_exit(ctalsa_exit);
CT_EXPORT_SYMBOL_NOVERS(map_io_space);
