/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
ALSA PCM implementation
Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/
#include "ctalsa.h"

#include <sound/pcm.h>
#include <linux/pci.h>

///////////////////////////////////////
// hardware descriptions for playback
static snd_pcm_hardware_t ctalsa_playback_hw[NUM_UDA_CARDS] =
{
    [SBLIVE_CARD] = {
        .info               = (SNDRV_PCM_INFO_MMAP |
                               SNDRV_PCM_INFO_INTERLEAVED |
                               SNDRV_PCM_INFO_BLOCK_TRANSFER |
                               SNDRV_PCM_INFO_MMAP_VALID |
                               SNDRV_PCM_INFO_PAUSE),
        .formats            = (SNDRV_PCM_FMTBIT_S8 |
                               SNDRV_PCM_FMTBIT_U8 |
                               SNDRV_PCM_FMTBIT_S16_LE |
                               SNDRV_PCM_FMTBIT_U16_LE),
        .rates              = (SNDRV_PCM_RATE_CONTINUOUS |
                               SNDRV_PCM_RATE_8000_48000),
        .rate_min           = 8000,
        .rate_max           = 48000,
        .channels_min       = 1,
        .channels_max       = 8,
        .buffer_bytes_max   = (128*1024),
        .period_bytes_min   = 64,
        .period_bytes_max   = (128*1024),
        .periods_min        = 1,
        .periods_max        = 1024,
        .fifo_size          = 0,
    },

    [AUDIGY_CARD] = {
        .info               = (SNDRV_PCM_INFO_MMAP |
                               SNDRV_PCM_INFO_INTERLEAVED |
                               SNDRV_PCM_INFO_BLOCK_TRANSFER |
                               SNDRV_PCM_INFO_MMAP_VALID |
                               SNDRV_PCM_INFO_PAUSE),
        .formats            = (SNDRV_PCM_FMTBIT_S8 |
                               SNDRV_PCM_FMTBIT_U8 |
                               SNDRV_PCM_FMTBIT_S16_LE |
                               SNDRV_PCM_FMTBIT_U16_LE |
                               SNDRV_PCM_FMTBIT_S24_3LE |
                               SNDRV_PCM_FORMAT_S32_LE),
        .rates              = (SNDRV_PCM_RATE_CONTINUOUS |
                               SNDRV_PCM_RATE_8000_192000),
        .rate_min           = 8000,
        .rate_max           = 192000,
        .channels_min       = 1,
        .channels_max       = 8,
        .buffer_bytes_max   = (128*1024),
        .period_bytes_min   = 64,
        .period_bytes_max   = (128*1024),
        .periods_min        = 1,
        .periods_max        = 1024,
        .fifo_size          = 0,
    },

    [CTXFI_CARD] = {
        .info               = (SNDRV_PCM_INFO_MMAP |
                               SNDRV_PCM_INFO_INTERLEAVED |
                               SNDRV_PCM_INFO_BLOCK_TRANSFER |
                               SNDRV_PCM_INFO_MMAP_VALID |
                               SNDRV_PCM_INFO_PAUSE),
        .formats            = (SNDRV_PCM_FMTBIT_S8 |
                               SNDRV_PCM_FMTBIT_U8 |
                               SNDRV_PCM_FMTBIT_S16_LE |
                               SNDRV_PCM_FMTBIT_U16_LE |
                               SNDRV_PCM_FMTBIT_S24_3LE |
                               SNDRV_PCM_FORMAT_S32_LE),
        .rates              = (SNDRV_PCM_RATE_CONTINUOUS |
                               SNDRV_PCM_RATE_8000_192000),
        .rate_min           = 8000,
        .rate_max           = 192000,
        .channels_min       = 1,
        .channels_max       = 8,
        .buffer_bytes_max   = (128*1024),
        .period_bytes_min   = 64,
        .period_bytes_max   = (128*1024),
        .periods_min        = 1,
        .periods_max        = 1024,
        .fifo_size          = 0,
    }
};

// hardware descriptions for capture
static snd_pcm_hardware_t ctalsa_capture_hw[NUM_UDA_CARDS] =
{
    [SBLIVE_CARD] = {
        .info               = (SNDRV_PCM_INFO_MMAP |
                               SNDRV_PCM_INFO_INTERLEAVED |
                               SNDRV_PCM_INFO_BLOCK_TRANSFER |
                               SNDRV_PCM_INFO_PAUSE |
                               SNDRV_PCM_INFO_MMAP_VALID),
        .formats            = (SNDRV_PCM_FMTBIT_S8 |
                               SNDRV_PCM_FMTBIT_U8 |
                               SNDRV_PCM_FMTBIT_S16_LE |
                               SNDRV_PCM_FMTBIT_U16_LE),
        .rates              = (SNDRV_PCM_RATE_CONTINUOUS |
                               SNDRV_PCM_RATE_8000_48000),
        .rate_min           = 8000,
        .rate_max           = 48000,
        .channels_min       = 1,
        .channels_max       = 2,
        .buffer_bytes_max   = (64*1024),
        .period_bytes_min   = 384,
        .period_bytes_max   = (64*1024),
        .periods_min        = 2,
        .periods_max        = 2,
        .fifo_size          = 0,
    },

    [AUDIGY_CARD] = {
        .info               = (SNDRV_PCM_INFO_MMAP |
                               SNDRV_PCM_INFO_INTERLEAVED |
                               SNDRV_PCM_INFO_BLOCK_TRANSFER |
                               SNDRV_PCM_INFO_PAUSE|
                               SNDRV_PCM_INFO_MMAP_VALID),
        .formats            = (SNDRV_PCM_FMTBIT_S8 |
                               SNDRV_PCM_FMTBIT_U8 |
                               SNDRV_PCM_FMTBIT_S16_LE |
                               SNDRV_PCM_FMTBIT_U16_LE |
                               SNDRV_PCM_FMTBIT_S24_3LE |
                               SNDRV_PCM_FORMAT_S32_LE),
        .rates              = (SNDRV_PCM_RATE_CONTINUOUS |
                               SNDRV_PCM_RATE_8000_192000),
        .rate_min           = 8000,
        .rate_max           = 192000,
        .channels_min       = 1,
        .channels_max       = 2,
        .buffer_bytes_max   = (64*1024),
        .period_bytes_min   = 384,
        .period_bytes_max   = (64*1024),
        .periods_min        = 2,
        .periods_max        = 2,
        .fifo_size          = 0,
    },

    [CTXFI_CARD] = {
        .info               = (SNDRV_PCM_INFO_MMAP |
                               SNDRV_PCM_INFO_INTERLEAVED |
                               SNDRV_PCM_INFO_BLOCK_TRANSFER |
                               SNDRV_PCM_INFO_PAUSE|
                               SNDRV_PCM_INFO_MMAP_VALID),
        .formats            = (SNDRV_PCM_FMTBIT_S8 |
                               SNDRV_PCM_FMTBIT_U8 |
                               SNDRV_PCM_FMTBIT_S16_LE |
                               SNDRV_PCM_FMTBIT_U16_LE |
                               SNDRV_PCM_FMTBIT_S24_3LE |
                               SNDRV_PCM_FORMAT_S32_LE),
        .rates              = (SNDRV_PCM_RATE_CONTINUOUS |
                               SNDRV_PCM_RATE_8000_192000),
        .rate_min           = 8000,
        .rate_max           = 192000,
        .channels_min       = 1,
        .channels_max       = 2,
        .buffer_bytes_max   = (64*1024),
        .period_bytes_min   = 384,
        .period_bytes_max   = (64*1024),
        .periods_min        = 2,
        .periods_max        = 4,
        .fifo_size          = 0,
    },
};

// hardware description for IEC958 subframe
static snd_pcm_hardware_t ctalsa_iec958_hw =
{
    .info               = (SNDRV_PCM_INFO_MMAP |
                           SNDRV_PCM_INFO_INTERLEAVED |
                           SNDRV_PCM_INFO_PAUSE),
    .formats            = SNDRV_PCM_FMTBIT_S16_LE,
    .rates              = SNDRV_PCM_RATE_48000,
    .rate_min           = 48000,
    .rate_max           = 48000,
    .channels_min       = 2,
    .channels_max       = 2,
    .buffer_bytes_max   = (128*1024),
    .period_bytes_min   = (4*1024),
    .period_bytes_max   = (128*1024),
    .periods_min        = 1,
    .periods_max        = 1024,
    .fifo_size          = 0,
};


//////////////////////////
// some helper functions
//////////////////////////
static CTDWORD
ConvertFormat(snd_pcm_format_t sndFormat)
{
    CTDWORD dwUDAFormat = HASF_SIGNED_16PCM;

    switch(sndFormat) {
        case SNDRV_PCM_FORMAT_U8:
            dwUDAFormat = HASF_UNSIGNED_8PCM;
            break;
        case SNDRV_PCM_FORMAT_S8:
            dwUDAFormat = HASF_SIGNED_8PCM;
            break;
        case SNDRV_PCM_FMTBIT_S16_LE:
            dwUDAFormat = HASF_SIGNED_16PCM;
            break;
        case SNDRV_PCM_FMTBIT_U16_LE:
            dwUDAFormat = HASF_UNSIGNED_16PCM;
            break;
        case SNDRV_PCM_FORMAT_S24_3LE:
            dwUDAFormat = HASF_SIGNED_24PCM;
            break;
        case SNDRV_PCM_FORMAT_S32_LE:
            dwUDAFormat = HASF_SIGNED_32PCM;
            break;
        default:
            dwUDAFormat = HASF_SIGNED_16PCM;
    }

    return dwUDAFormat;
}

/* Legacy
static CTDWORD
caculate_offset(unsigned int start_addr, snd_pcm_runtime_t *runtime)
{
    int stereo, w_16;
    CTDWORD dwTemp=0;

    stereo = runtime->channels == 2;
    w_16 = snd_pcm_format_width(runtime->format) == 16;
    if (stereo) {
        start_addr >>= 1;
    }
    if (w_16) {
        start_addr >>= 1;
    }

    unsigned int ccis = stereo ? 28 : 30;
    if (w_16)
        ccis *= 2;

    dwTemp = start_addr + ccis;

    return dwTemp;
}
*/


// callback functions
static CTVOID
envaudio_buffer_callback(enum HAEASS_TRANSPORTMESSAGE Message, CTHANDLE Argument, CTHANDLE Context)
{
    stBuffObj *pBuffObj = (stBuffObj *)Context;

    //CTDPF("env_audio_callback is called");

    if (Message == HAEASS_MSG_NOTIFY) {
        pBuffObj->dwOffset = pBuffObj->dwBufferSize;
        if(NULL == pBuffObj->substream)
            return;
        snd_pcm_period_elapsed((snd_pcm_substream_t *)pBuffObj->substream);
    } else if (Message == HAEASS_MSG_RIPOFF)
        CTASSERT(0);
}

static CTVOID
wavein_buffer_callback(PCTVOID pWaveInBuffer)
{
    stBuffObj *pBuffObj;

    //CTDPF("wavein_buffer_callback is called");

    pBuffObj = (stBuffObj *)ctdrv_WIB_GetUserData(pWaveInBuffer);
    if (NULL == pBuffObj->substream) return;
    snd_pcm_period_elapsed((snd_pcm_substream_t *)pBuffObj->substream);
}

static CTVOID
waveout_buffer_callback(PCTVOID pWaveOutBuffer, HAWODMESSAGETYPE message)
{
    stBuffObj *pBuffObj;

    //CTDPF("waveout_buffer_callback is called");

    if (message == HAWOD_NOTIFY) {
        pBuffObj = (stBuffObj *)ctdrv_WOB_GetUserData(pWaveOutBuffer);
        pBuffObj->dwOffset = pBuffObj->dwBufferSize;
        if (NULL == pBuffObj->substream) return;
        snd_pcm_period_elapsed((snd_pcm_substream_t *)pBuffObj->substream);
    }
}

static CTVOID
ac3_buffer_callback(PCTVOID pWaveOutBuffer, HAWODMESSAGETYPE message)
{
    stBuffObj *pBuffObj;

    //CTDPF("ac3_buffer_callback is called");

    if (message == HAWOD_NOTIFY) {
        pBuffObj = (stBuffObj *)ctdrv_WOB_GetUserData(pWaveOutBuffer);
        if (NULL == pBuffObj->substream) return;
        snd_pcm_period_elapsed((snd_pcm_substream_t *)pBuffObj->substream);
    }
}


/*********************************
 * playback operations
 *********************************/
static int
snd_ctalsa_playback_open(snd_pcm_substream_t *substream)
{
    stHwObj *pHwObj = snd_pcm_substream_chip(substream);
    snd_pcm_runtime_t *runtime = substream->runtime;
    stBuffObj *pBuffObj;
    int err;

    CTDPF("snd_ctalsa_playback_open is called, tPCMDev = %d", substream->pcm->device);

    pBuffObj = kmalloc(sizeof(stBuffObj), GFP_KERNEL);
    if (!pBuffObj) return -ENOMEM;
    memset(pBuffObj, 0, sizeof(stBuffObj));

    pBuffObj->substream = substream;
    pBuffObj->tPCMDev = substream->pcm->device; // who am i?

    if ((err = snd_pcm_hw_constraint_integer(runtime, SNDRV_PCM_HW_PARAM_PERIODS)) < 0) {
        kfree(pBuffObj);
        return err;
    }
    if ((err = snd_pcm_hw_constraint_minmax(runtime, SNDRV_PCM_HW_PARAM_BUFFER_BYTES, 256, UINT_MAX)) < 0) {
        kfree(pBuffObj);
        return err;
    }

    runtime->private_data = pBuffObj;
    if (PCM_IEC958 == pBuffObj->tPCMDev) {
        runtime->hw = ctalsa_iec958_hw;
    } else {
        runtime->hw = ctalsa_playback_hw[pHwObj->tCard];
    }

    return 0;
}

static int
snd_ctalsa_playback_close(snd_pcm_substream_t *substream)
{
    stBuffObj *pBuffObj = substream->runtime->private_data;
    HRESULT hr = E_FAIL;

    CTDPF("snd_ctalsa_playback_close is called");

    if (!pBuffObj) return -ENODEV;
    
    CTDPF("%s: tPCMDev = %d", __func__, pBuffObj->tPCMDev);

    if (pBuffObj->pWOB) {
    	hr = ctdrv_WOB_Stop(pBuffObj); 
    	hr = ctdrv_WOB_Release(pBuffObj);
    }

    kfree(pBuffObj);

    return 0;
}

static int
snd_ctalsa_playback_prepare(snd_pcm_substream_t *substream)
{
    stHwObj *pHwObj = snd_pcm_substream_chip(substream);
    snd_pcm_runtime_t *runtime = substream->runtime;
    stBuffObj *pBuffObj = runtime->private_data;
    CTHANDLE pWOD;
    CTDWORD dwFragSize, dwFrameSize;
    WOB_CALLBACK pWOBCallback;
    HRESULT hr = E_FAIL;

    CTDPF("snd_ctalsa_playback_prepare is called, tPCMDev = %d, dwNumChans = %d", pBuffObj->tPCMDev, runtime->channels);

    switch (pBuffObj->tPCMDev) {
        case PCM_ENVAUDIO:
            pWOD = pHwObj->apIha[WAVEOUT1_DEV];
            pWOBCallback = (WOB_CALLBACK)envaudio_buffer_callback;
            pBuffObj->bAC3Format = FALSE;
            pBuffObj->bExtFormat = FALSE;
            break;
        case PCM_WAVEOUT:
            pWOD = pHwObj->apIha[WAVEOUT2_DEV];
            pWOBCallback = (WOB_CALLBACK)waveout_buffer_callback;
            break;
        case PCM_AC3MGR:
            pWOD = pHwObj->apIha[AC3MGR_DEV];
            pWOBCallback = (WOB_CALLBACK)ac3_buffer_callback;
            break;
        default:
            return -ENODEV;
    }
    if (NULL == pWOD) return -ENODEV;
    
    if (pBuffObj->pWOB) {
        CTDPF("this playback buffer has been prepared already");
        return 0;
    }

    dwFragSize = snd_pcm_lib_period_bytes(substream);
    dwFrameSize = snd_pcm_format_width(runtime->format)*runtime->channels/8;
 
    pBuffObj->dwFreq = dwFragSize/dwFrameSize;
    pBuffObj->dwPosition = 0;
    pBuffObj->dwOffset = 0;
    pBuffObj->dwBufferSize = runtime->dma_bytes;
    pBuffObj->dwSampleRate = runtime->rate;
    pBuffObj->dwSampleFormat = ConvertFormat(runtime->format);
    pBuffObj->dwNumChans = runtime->channels;
    pBuffObj->pVirtAddr = runtime->dma_area;
    pBuffObj->bLoopState = TRUE;

#if defined(CONFIG_SND_PCM_OSS) || defined(CONFIG_SND_PCM_OSS_MODULE)
    CTDPF("runtime->oss.prepare : %d", runtime->oss.prepare);
    pBuffObj->bOSSEmul = runtime->oss.prepare;
#else
    pBuffObj->bOSSEmul = FALSE;
#endif
 
    hr = ctdrv_WOD_CreateBuffer(pWOD, pWOBCallback, pBuffObj);
    if (FAILED(hr)) return -ENOMEM;

    return 0;
}

static int
snd_ctalsa_playback_trigger(snd_pcm_substream_t *substream, int cmd)
{
    snd_pcm_runtime_t *runtime = substream->runtime;
    stBuffObj *pBuffObj = runtime->private_data;
    HRESULT hr = E_FAIL;

    CTDPF("snd_ctalsa_playback_trigger is called, tPCMDev = %d, dwNumChans = %d", pBuffObj->tPCMDev,
                    pBuffObj->dwNumChans);

    if (!pBuffObj->pWOB) return -ENODEV;

    switch (cmd) {
        case SNDRV_PCM_TRIGGER_START:
            CTDPF("snd_ctalsa_playback_trigger: START");
            pBuffObj->dwOffset = pBuffObj->dwBufferSize;
            hr = ctdrv_WOB_UpdateBuffer(pBuffObj, pBuffObj->dwPosition, pBuffObj->dwOffset);
            pBuffObj->dwPosition = (pBuffObj->dwPosition+pBuffObj->dwOffset)%(runtime->dma_bytes);
            pBuffObj->dwOffset = 0;
            hr = ctdrv_WOB_Play(pBuffObj);
            CTASSERT(SUCCEEDED(hr));
            break;

        case SNDRV_PCM_TRIGGER_STOP:
            CTDPF("snd_ctalsa_playback_trigger: STOP");
            // FIXME: ~~~
            hr = ctdrv_WOB_Stop(pBuffObj);
            CTASSERT(SUCCEEDED(hr));
            break;

        case SNDRV_PCM_TRIGGER_SUSPEND:
        case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
            CTDPF("snd_ctalsa_playback_trigger: PAUSE_PUSH");
            hr = ctdrv_WOB_Pause(pBuffObj);
            CTASSERT(SUCCEEDED(hr));
            break;

        case SNDRV_PCM_TRIGGER_RESUME:
        case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
            CTDPF("snd_ctalsa_playback_trigger: PAUSE_RELEASE");
            hr = ctdrv_WOB_Play(pBuffObj);
            CTASSERT(SUCCEEDED(hr));
            break;

        default:
            CTDPF("snd_ctalsa_playback_trigger: Unknown command");
            return -EINVAL;
    }

    return 0;
}

static snd_pcm_uframes_t
snd_ctalsa_playback_pointer(snd_pcm_substream_t *substream)
{
    snd_pcm_runtime_t *runtime = substream->runtime;
    stBuffObj *pBuffObj = runtime->private_data;
    CTDWORD dwInFrames = 0, dwInBytes = 0;

    //CTDPF("snd_ctalsa_playback_pointer is called, tPCMDev = %d, dwNumChans = %d", pBuffObj->tPCMDev, pBuffObj->dwNumChans);

    if (!pBuffObj->pWOB) return -ENODEV;

    dwInBytes = ctdrv_WOB_GetPlaybackPosition(pBuffObj);
    ctdrv_WOB_UpdateBuffer(pBuffObj, pBuffObj->dwPosition, pBuffObj->dwOffset);
    pBuffObj->dwPosition = (pBuffObj->dwPosition+pBuffObj->dwOffset)%runtime->dma_bytes;
    pBuffObj->dwOffset = 0;

    // FIXME : Returning current hardware position on the buffer
    // Symptoms : static sound when using "play" program. No problem on mplayer or xmms using OSS.
    if (pBuffObj->bOSSEmul) {
        //CTDPF("snd_ctalsa_playback_pointer : OSSEmulation");
        dwInBytes -= snd_pcm_format_width(runtime->format)*runtime->channels;
    }
    dwInFrames = bytes_to_frames(substream->runtime, dwInBytes);
    
    return dwInFrames;
}


/*********************************
 * capture operations
 *********************************/
static int
snd_ctalsa_capture_open(snd_pcm_substream_t *substream)
{
    stHwObj *pHwObj = snd_pcm_substream_chip(substream);
    snd_pcm_runtime_t *runtime = substream->runtime;
    stBuffObj *pBuffObj;

    CTDPF("snd_ctalsa_capture_open is called");

    pBuffObj = kmalloc(sizeof(stBuffObj), GFP_KERNEL);
    if (!pBuffObj) return -ENOMEM;
    memset(pBuffObj, 0, sizeof(stBuffObj));

    pBuffObj->substream = substream;

    runtime->private_data = pBuffObj;
    runtime->hw = ctalsa_capture_hw[pHwObj->tCard];

    return 0;
}

static int
snd_ctalsa_capture_close(snd_pcm_substream_t *substream)
{
    stBuffObj *pBuffObj = substream->runtime->private_data;
    HRESULT hr = E_FAIL;

    CTDPF("snd_ctalsa_capture_close is called");

    if (!pBuffObj) return -ENODEV;

    if (pBuffObj->pWIB) {
        hr = ctdrv_WIB_Stop(pBuffObj);
        hr = ctdrv_WIB_Release(pBuffObj);
    }

    kfree(pBuffObj);

    return 0;
}

static int
snd_ctalsa_capture_prepare(snd_pcm_substream_t *substream)
{
    stHwObj *pHwObj = snd_pcm_substream_chip(substream);
    stBuffObj *pBuffObj = substream->runtime->private_data;
    snd_pcm_runtime_t *runtime = substream->runtime;
    CTDWORD dwFragSize, dwFrameSize;
    CTHANDLE pWID = pHwObj->apIha[WAVEIN_DEV];
    HRESULT hr;

    CTDPF("snd_ctalsa_capture_prepare is called");

    if (NULL == pWID) return -ENODEV;

    if (pBuffObj->pWIB) {
        CTDPF("the wavein buffer has been prepared already");
        return 0;
    }

    dwFragSize = snd_pcm_lib_period_bytes(substream);
    dwFrameSize = (snd_pcm_format_width(runtime->format)*(runtime->channels))/8;

    pBuffObj->dwFreq = (dwFragSize/dwFrameSize);
    pBuffObj->dwSampleRate = runtime->rate;
    pBuffObj->dwSampleFormat = ConvertFormat(runtime->format);
    pBuffObj->dwNumChans = runtime->channels;
    pBuffObj->dwBufferSize = snd_pcm_lib_buffer_bytes(substream);
    pBuffObj->pVirtAddr = runtime->dma_area;

    hr = ctdrv_WID_CreateBuffer(pWID, wavein_buffer_callback, pBuffObj);
    if (FAILED(hr)) return -ENOMEM;

    return 0;
}

static int
snd_ctalsa_capture_trigger(snd_pcm_substream_t *substream, int cmd)
{
    stBuffObj *pBuffObj = substream->runtime->private_data;
    HRESULT hr = E_FAIL;

    if (!pBuffObj->pWIB) return -ENODEV;

    switch (cmd) {
        case SNDRV_PCM_TRIGGER_STOP:
            CTDPF("snd_ctalsa_capture_trigger(): STOP");
            hr = ctdrv_WIB_Stop(pBuffObj);
            CTASSERT(SUCCEEDED(hr));
            break;

        case SNDRV_PCM_TRIGGER_START:
            CTDPF("snd_ctalsa_capture_trigger(): START");
            hr = ctdrv_WIB_Record(pBuffObj);
            CTASSERT(SUCCEEDED(hr));
            break;

        case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
            CTDPF("snd_ctalsa_capture_trigger(): PAUSE_PUSH");
            hr = ctdrv_WIB_Pause(pBuffObj);
            CTASSERT(SUCCEEDED(hr));
            break;

        case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
            CTDPF("snd_ctalsa_capture_trigger(): PAUSE_RELEASE");
            hr = ctdrv_WIB_Record(pBuffObj);
            CTASSERT(SUCCEEDED(hr));
            break;

        default:
            CTDPF("snd_ctalsa_capture_trigger(): Unknown command");
            return -EINVAL;
    }

    return 0;
}

static snd_pcm_uframes_t
snd_ctalsa_capture_pointer(snd_pcm_substream_t *substream)
{
    stBuffObj *pBuffObj = substream->runtime->private_data;
    CTDWORD dwInFrames = 0, dwInBytes = 0;

    CTDPF("snd_ctalsa_capture_pointer is called");

    if (!pBuffObj->pWIB) return -ENODEV;

    dwInBytes = ctdrv_WIB_GetRecordPosition(pBuffObj);
    dwInFrames = bytes_to_frames(substream->runtime, dwInBytes);

    return dwInFrames;
}


// common pcm ops
static int
snd_ctalsa_pcm_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params)
{
    int err;

    CTDPF("snd_ctalsa_pcm_hw_params is called");

    if ((err = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params))) < 0)
        return err;

    return 0;
}

static int
snd_ctalsa_pcm_hw_free(snd_pcm_substream_t *substream)
{
    CTDPF("snd_ctalsa_pcm_hw_free is called");

    return snd_pcm_lib_free_pages(substream);
}


///////////////////////////////
// pcm operators for playback
static snd_pcm_ops_t snd_ctalsa_playback_ops = {
    .open       = snd_ctalsa_playback_open,
    .close      = snd_ctalsa_playback_close,
    .ioctl      = snd_pcm_lib_ioctl,
    .hw_params  = snd_ctalsa_pcm_hw_params,
    .hw_free    = snd_ctalsa_pcm_hw_free,
    .prepare    = snd_ctalsa_playback_prepare,
    .trigger    = snd_ctalsa_playback_trigger,
    .pointer    = snd_ctalsa_playback_pointer
};

// pcm operators for capture
static snd_pcm_ops_t snd_ctalsa_capture_ops = {
    .open       = snd_ctalsa_capture_open,
    .close      = snd_ctalsa_capture_close,
    .ioctl      = snd_pcm_lib_ioctl,
    .hw_params  = snd_ctalsa_pcm_hw_params,
    .hw_free    = snd_ctalsa_pcm_hw_free,
    .prepare    = snd_ctalsa_capture_prepare,
    .trigger    = snd_ctalsa_capture_trigger,
    .pointer    = snd_ctalsa_capture_pointer,
};


/////////////////////////////////
// init and exit points of PCM
/////////////////////////////////
static void
ctalsa_pcm_free(snd_pcm_t *pcm)
{
    CTDPF("ctalsa_pcm_free()");
    snd_pcm_lib_preallocate_free_for_all(pcm);
}

static int __devinit
ctalsa_pcm_new(stHwObj *pHwObj, char *szName, enum CTALSA_PCM_DEV device)
{
    CTDPF("ctalsa_pcm_new is called.  device=%d", (int)device);
    snd_pcm_t *pcm;
    int err, playback_count, capture_count;
    
    if (PCM_WAVEIN == device) {
	    return 0;
    }
    
    playback_count = (device == PCM_IEC958) ? 1 : 32;
    capture_count = (PCM_WAVEOUT == device) ? 1 : 0;

    if ((err = snd_pcm_new(pHwObj->pCard, szName, device, playback_count, capture_count, &pcm)) < 0)
    {
        CTDPF("ctalsa_pcm_new calling snd_pcm_new failed!! Err=%d", err);
        return err;
    }

    pcm->private_data = pHwObj;
    pcm->private_free = ctalsa_pcm_free;
    pcm->info_flags = 0;
    pcm->dev_subclass = SNDRV_PCM_SUBCLASS_GENERIC_MIX;

    switch (device) {
        case PCM_ENVAUDIO:
            strcpy(pcm->name, "EnvAudio");
            snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_ctalsa_playback_ops);
            break;

        case PCM_WAVEOUT:
	    /* Integrate WaveIn device into WaveOut device */
            strcpy(pcm->name, "WaveOut/WaveIn");
            snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_ctalsa_playback_ops);
            snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &snd_ctalsa_capture_ops);
            break;
           
        case PCM_AC3MGR:
            strcpy(pcm->name, "AC3Mgr");
            snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_ctalsa_playback_ops);
            break;
 
        case PCM_IEC958:
            strcpy(pcm->name, "IEC958");
            snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_ctalsa_playback_ops);
            break;

        default:
            break;
    }

    snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV, snd_dma_pci_data(pHwObj->pPCIDev),
            128*1024, 128*1024);

    return 0;
}

int __devinit
envaudio_init(stHwObj *pHwObj, char *szName)
{
    int err;

    CTDPF("envaudio_init is called");

    if ((err = ctalsa_pcm_new(pHwObj, szName, PCM_ENVAUDIO)) < 0) {
        snd_card_free(pHwObj->pCard);
        return err;
    }

    return err;
}

void __devexit
envaudio_exit(stHwObj *pHwObj, int dev)
{
    CTDPF("envaudio_exit is called");
}


int __devinit
waveout_init(stHwObj *pHwObj, char *szName)
{
    int err;

    CTDPF("waveout_init is called");

    if ((err = ctalsa_pcm_new(pHwObj, szName, PCM_WAVEOUT)) < 0) {
        snd_card_free(pHwObj->pCard);
        return err;
    }

    return err;
}

void __devexit
waveout_exit(stHwObj *pHwObj, int dev)
{
    CTDPF("waveout_exit is called");
}


int __devinit
wavein_init(stHwObj *pHwObj, char *szName)
{
    int err;

    CTDPF("wavein_init is called");

    if ((err = ctalsa_pcm_new(pHwObj, szName, PCM_WAVEIN)) < 0) {
        CTDPF("wavein_init error.  Err=%d", err);
        snd_card_free(pHwObj->pCard);
        return err;
    }

    return err;
}

void __devexit
wavein_exit(stHwObj *pHwObj, int dev)
{
    CTDPF("wavein_exit is called");
}

int __devinit
ac3mgr_init(stHwObj *pHwObj, char *szName)
{
    int err;

    CTDPF("ac3mgr_init is called");

    if ((err = ctalsa_pcm_new(pHwObj, szName, PCM_AC3MGR)) < 0) {
        snd_card_free(pHwObj->pCard);
        return err;
    }

    return err;
}

void __devexit
ac3mgr_exit(stHwObj *pHwObj, int dev)
{
    CTDPF("ac3mgr_exit is called");
}


