/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
Main code module for EmuPIA
Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#define EXPORT_SYMTAB
#include <linux/version.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/sound.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/devfs_fs_kernel.h>
#endif



#include "PIAMain.h"
#include "PIAServ.h"
#include "osutils.h"

// Global vars
INSERT_VERSION_INFO;

static int disable = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#   ifdef DEVFS_ENABLED
static devfs_handle_t g_dev_emupia;
#   endif
#endif


int emupia_InitDriver(void);
int emupia_ExitDriver(void);
int emupia_Cleanup(void);

/////////////////////////
// EmuPIA file operations
/////////////////////////

int 
emu10kx_emupia_ioctl(struct inode *pInode, struct file *pFile, unsigned int uCmd, unsigned long ulArg)
{
    return ioctl_dispatch(ulArg, PIADispatch);
}


int 
emu10kx_emupia_open(struct inode *pInode, struct file *pFile)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    CT_MOD_INC_USE_COUNT;
#endif
    
    return 0;
}

int 
emu10kx_emupia_release(struct inode *pInode, struct file *pFile)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    CT_MOD_DEC_USE_COUNT;
#endif

    return 0;
}

struct file_operations emu10kx_emupia_fops = 
{
    ioctl:   emu10kx_emupia_ioctl,
    open:    emu10kx_emupia_open,
    release: emu10kx_emupia_release,
};


////////////////////////
// Entry and exit points
////////////////////////

static int g_nDevEmuPIA = -1;

static int __init
emupia_init(void)
{

    if (disable)
        return 0;

    __do_global_ctors_aux();

    emupia_InitDriver();

    PDEBUG("Registering PIA device\n");
    g_nDevEmuPIA = register_sound_special(&emu10kx_emupia_fops, EMUPIA_MINOR);
    if (g_nDevEmuPIA < 0) {
        printk(KERN_ERR "emupia: cannot register device\n");
        emupia_Cleanup();
        return -ENODEV;
#ifdef DEVFS_ENABLED
#   if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    } else {
        PDEBUG("Registering PIA device via devfs_register\n");
        g_dev_emupia = devfs_register(NULL, "emupia", DEVFS_FL_DEFAULT, SOUND_MAJOR, EMUPIA_MINOR, 
                S_IRUSR | S_IWUSR | S_IFCHR, &emu10kx_emupia_fops, NULL);
        PDEBUG("devfs_register returns: %d\n", (int)g_dev_emupia);

#   else
    } else {
        PDEBUG("Registering PIA device via devfs_mk_cdev\n");
        devfs_mk_cdev(MKDEV(SOUND_MAJOR, EMUPIA_MINOR), S_IFCHR|S_IRUSR|S_IWUSR, "emupia");
#   endif
#endif
    }

    return 0;
}

static void __exit 
emupia_exit(void)
{
    if (disable)    
        return;

    PDEBUG("Unregistering PIA device\n");
    if (g_nDevEmuPIA >= 0) {
        unregister_sound_special(g_nDevEmuPIA);
#ifdef DEVFS_ENABLED
#   if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
        PDEBUG("Unregistering PIA device in DEVFS\n");
        devfs_unregister(g_dev_emupia);
#   else
        PDEBUG("Unregistering PIA device in DEVFS\n");
        devfs_remove("emupia");
#   endif
#endif
    }

    emupia_ExitDriver();
}

module_init(emupia_init);
module_exit(emupia_exit);


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
MODULE_PARM(disable, "i");
#else
module_param(disable, int, 0);
#endif
MODULE_PARM_DESC(disable, "Sets module to disabled state");

MODULE_AUTHOR("Copyright (C) 1999-2008 Creative Technology Ltd");
MODULE_DESCRIPTION("EMUPIA driver version 1.18.0000");
MODULE_LICENSE("Creative Proprietary License");

