/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
IhwReg implementation for Linux
Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#include <linux/kernel.h>
#include <linux/slab.h>
#include <asm/io.h>

#include "IasOSSrv.h"
#include "IhwReg.h"


typedef struct {
    IhwRegVtbl vtbl;
} tRegVtbl;

typedef struct {
    tRegVtbl *lpVtbl;
    CTDWORD m_cRef;
    ICtUnknown *m_pIUnknownOuter;
    IasOSSrv *m_pIasOSSrv;
} HWREGOBJ, *PHWREGOBJ;

// Function prototypes
CTVOID regConstruct(CTVOID *This, IasOSSrv *pIasOSSrv);
CTVOID regDestruct(CTVOID *This);

CTVOID *
CreateHwRegObject(IasOSSrv *pIasOSSrv)
{
    PHWREGOBJ pObj;

    pObj = kmalloc(sizeof(HWREGOBJ), GFP_KERNEL);
    regConstruct(pObj, pIasOSSrv);
    
    return pObj;
}

CTDWORD 
regNondelegatingAddRef(ICTUNK_IARGLIST)
{
    PHWREGOBJ pObj = This;
    
    return ++pObj->m_cRef;
}

CTDWORD 
regNondelegatingRelease(ICTUNK_IARGLIST)
{
    PHWREGOBJ pObj = This;
    CTDWORD dwRef;

    dwRef = --pObj->m_cRef;

    if (dwRef == 0) {
        CTDPF("Warning: regNondelegatingRelease() is deleting object");
        regDestruct(pObj);
        kfree(pObj);
    }

    return dwRef;
}

HRESULT 
regNondelegatingQueryInterface(ICTUNK_QIARGLIST)
{
    if (!ppv) CTRETURN(E_INVALIDARG);

    if( IsEqualGUID(iid, &IID_ICtUnknown))
    {
        *ppv = This;
    } else if ( IsEqualGUID(iid, &IID_IhwReg)) {
        *ppv = This;
    } else {
        *ppv = NULL;
        return E_NOINTERFACE;
    }

    regNondelegatingAddRef(CICTUNK_PASSTHRU(This));

    return S_OK;
}

HRESULT 
regInit(THIS_ ICtUnknown *pIUnknownOuter)
{
    PHWREGOBJ pObj = This;

    if (pIUnknownOuter)
        pObj->m_pIUnknownOuter = pIUnknownOuter;

    return S_OK;
}

HRESULT 
regQueryInterface(ICTUNK_QIARGLIST)
{
    PHWREGOBJ pObj = This;

    return COMCALL(ICtUnknown, pObj->m_pIUnknownOuter, QueryInterface)(pObj->m_pIUnknownOuter, iid, ppv QICTUNK);
}

CTDWORD 
regAddRef(ICTUNK_IARGLIST)
{
    PHWREGOBJ pObj = This;

    return COMCALL(ICtUnknown, pObj->m_pIUnknownOuter, AddRef)(CICTUNK(pObj->m_pIUnknownOuter));
}

CTDWORD 
regRelease(ICTUNK_IARGLIST)
{
    PHWREGOBJ pObj = This;

    return COMCALL(ICtUnknown, pObj->m_pIUnknownOuter, Release)(CICTUNK(pObj->m_pIUnknownOuter));
}

CTBYTE 
regReadPortByte(THIS_ CTHANDLE hAddr)
{
    return inb((unsigned long) hAddr);
}

CTWORD 
regReadPortWord(THIS_ CTHANDLE hAddr)
{
    return inw((unsigned long) hAddr);
}

CTDWORD 
regReadPortDword(THIS_ CTHANDLE hAddr)
{
    return inl((unsigned long) hAddr);
}

CTVOID 
regWritePortByte(THIS_ CTHANDLE hAddr, CTBYTE byValue)
{
    outb(byValue, (unsigned long) hAddr);
}

CTVOID 
regWritePortWord(THIS_ CTHANDLE hAddr, CTWORD wValue)
{
    outw(wValue, (unsigned long) hAddr);
}

CTVOID 
regWritePortDword(THIS_ CTHANDLE hAddr, CTDWORD dwValue)
{
    outl(dwValue, (unsigned long) hAddr);
}


// Function table
static const tRegVtbl g_RegVtbl = 
{
    .vtbl =
    {
        regNondelegatingQueryInterface,
        regNondelegatingAddRef,
        regNondelegatingRelease,
        regInit,
        regQueryInterface,
        regAddRef,
        regRelease,
        regReadPortByte,
        regReadPortWord,
        regReadPortDword,
        regWritePortByte,
        regWritePortWord,
        regWritePortDword
    }
};

// Constructor/Destructor
CTVOID 
regConstruct(CTVOID *This, IasOSSrv *pIasOSSrv)
{
    PHWREGOBJ pObj = This;

    pObj->lpVtbl = kmalloc(sizeof(tRegVtbl), GFP_KERNEL);
    memcpy(pObj->lpVtbl, &g_RegVtbl, sizeof(tRegVtbl));
    // pObj->lpVtbl->pObj = This;
    pObj->m_cRef = 1;
    pObj->m_pIasOSSrv = pIasOSSrv;
    pObj->m_pIUnknownOuter = This;    
}

CTVOID 
regDestruct(CTVOID *This)
{
    PHWREGOBJ pObj = This;

    if (pObj->lpVtbl) {
        kfree(pObj->lpVtbl);
        pObj->lpVtbl = NULL;
    }
}

