/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
IasSystem implementation for Linux
Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
See bottom for preserved copyright notices
*******************************************************************************
*/

#define __NO_VERSION__
#define __KERNEL_SYSCALLS__
#include <linux/version.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/unistd.h>
#include <linux/kmod.h> // For loading modules
#include <linux/mman.h> // For mmap
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/smp.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>

//#define __USE_LIBM_A__
#ifdef __USE_LIBM_A__
    #define _USE_LIBM_MATH_H
    #include <stdio.h>
    #include <math.h>
    #include <bits/nan.h>
    //#include <endian.h>
    #include <float.h>

    _LIB_VERSION_TYPE _LIB_VERSION = _IEEE_;

#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#   include <linux/tqueue.h>
#else 
#   include <linux/workqueue.h>
#   include <linux/interrupt.h>
#   include <linux/syscalls.h>
#   include <asm/cacheflush.h>
#   include <asm/unistd.h>
#   define __NR_ct_delete_module    __NR_delete_module
//_syscall2(int, ct_delete_module, const char *, name, unsigned int, flags
#   define delete_module(a)     ct_delete_module(a, O_TRUNC)
#endif

#include "IasSystem.h"
#include "ctossrv.h"

#    ifndef HUGE_VAL
#       if defined(__GNUC__) && (__GNUC__ >= 3)
#           define HUGE_VAL    (__extension__ 0x1.0p2047)
#       else
            static union { 
                unsigned char __c[8];
                double __d;
            } __huge_val = { 0, 0, 0, 0, 0, 0, 0xf0, 0x7f };
#           define HUGE_VAL    (__huge_val.__d)
#       endif
#   endif



/*
#if __FLOAT_WORD_ORDER == BIG_ENDIAN
typedef union {
    double value;
    struct {
        u_int32_t msw;
        u_int32_t lsw;
    } parts;
} ieee_double_shape_type;
#endif
*/

//#if __FLOAT_WORD_ORDER == LITTLE_ENDIAN
typedef union {
    double value;
    struct {
        u_int32_t lsw;
        u_int32_t msw;
    } parts;
} ieee_double_shape_type;
//#endif

/* Get two 32 bit ints from a double.  */
#define EXTRACT_WORDS(ix0,ix1,d)                                \
    do {                                                        \
        ieee_double_shape_type ew_u;                            \
        ew_u.value = (d);                                       \
        (ix0) = ew_u.parts.msw;                                 \
        (ix1) = ew_u.parts.lsw;                                 \
    } while (0)

/* Get the more significant 32 bit int from a double.  */
#define GET_HIGH_WORD(i,d)                                      \
    do {                                                        \
        ieee_double_shape_type gh_u;                            \
        gh_u.value = (d);                                       \
        (i) = gh_u.parts.msw;                                   \
    } while (0)

// Disable Math Functions for x86_64 for now,
// until we have ported them to x86_64 arch. 
#if defined(__x86_64__)
//#define NO_MATH
#endif

//#define USE_NATIVE_MATH
#ifdef USE_NATIVE_MATH
#   define FPCALL(fn) fn
#   define FPUBUFSIZE 112
#elif defined(i386) // Use x86 asm library
#   include <x86math.h> 
#   define FPCALL(fn) fp_##fn
#elif defined(__x86_64__) // Use x86_64 asm library
#   include <x86math.h> 
#   define FPCALL(fn) fp_##fn
#else // Non x86 platform
#   error Floating point library not available for this platform
#endif // USE_NATIVE_MATH

/* Maximum length of the registry value name. */
#define MAX_REGVALUENAME_LEN        64
/* Maximum length of the registry path. */
#define MAX_REGISTRYPATH_LEN        512
#define CTCHARTOUPPER(c) (((c)>='a'&&(c)<='z')?((c)-'a'+'A'):(c))
#define CTCHARTOHEX(c) (((c)>='0'&&(c)<='9')?(c)-'0':(CTCHARTOUPPER(c)>='A'&&CTCHARTOUPPER(c)<='F')?CTCHARTOUPPER(c)-'A'+10:0xff)

typedef struct{
    IasSystemVtbl  vtbl1;
    IasSystem2Vtbl vtbl2;
    IasSystem3Vtbl vtbl3;
    IasSystem4Vtbl vtbl4;
    IasSystem5Vtbl vtbl5;
} tSysVtbl;

typedef struct {
    tSysVtbl *lpVtbl;
    CTDWORD m_cRef;
    ICtUnknown *m_pIUnknownOuter;
    IasOSSrv *m_pIasOSSrv;
} SYSOBJ, *PSYSOBJ;


//////////////////////////////////////
// Function prototypes
CTVOID sysConstruct(CTVOID *This, IasOSSrv *pIasOSSrv);
CTVOID sysDestruct(CTVOID *This);
unsigned long kvirt_to_phys(unsigned long adr);
unsigned long kvirt_to_bus(unsigned long adr);
unsigned long bytes_to_order(CTDWORD dwBytes);
HRESULT reserve_remap_pages(unsigned long ulUserAddr, stMemHandle *pMem, unsigned long ulSize, 
        pgprot_t prot);
HRESULT unreserve_pages(stMemHandle *pMem, unsigned long ulSize);
CTDWORD sysAddRef(ICTUNK_IARGLIST);


//////////////////////////////////////
// Object creation function
CTVOID *
CreateSystemObject(IasOSSrv *pIasOSSrv)
{
    PSYSOBJ pObj;

    //CTDPF("CreateSystemObject called");
    pObj = kmalloc(sizeof(SYSOBJ), GFP_KERNEL);
    sysConstruct(pObj, pIasOSSrv);
    CTDPF("CreateSystemObject() %p", pObj); 
    
    return pObj;
}


//////////////////////////////////
// ICtNondelegatingUnknown methods
//////////////////////////////////

CTDWORD 
sysNondelegatingAddRef(ICTUNK_IARGLIST)
{
    // CTDPF("sysNondelegatingAddRef called");
    PSYSOBJ pObj = This;
    CTASSERT(pObj);
    
    return ++pObj->m_cRef;
}

CTDWORD 
sysNondelegatingRelease(ICTUNK_IARGLIST)
{
    PSYSOBJ pObj = This;
    CTDWORD dwRef;
    
    //CTDPF("sysNondelegatingRelease() %lx", This);
    CTASSERT(pObj);
    dwRef = --pObj->m_cRef;

    if (dwRef == 0) {
        CTDPF("Warning: sysNondelegatingRelease() is deleting object");
        sysDestruct(pObj);
        kfree(pObj);
    }

    return dwRef;
}

HRESULT 
sysNondelegatingQueryInterface(ICTUNK_QIARGLIST)
{
    PSYSOBJ pObj = This;

    //CTDPF("sysNondelegatingQueryInterface() %lx", This);
    if (!ppv) 
        CTRETURN(E_INVALIDARG);

    if( IsEqualGUID(iid, &IID_ICtUnknown)) {
        *ppv = This;
        //sysNondelegatingAddRef(CICTUNK(This));
    } else if ( IsEqualGUID(iid, &IID_IasSystem)) {
        *ppv = This;
        //sysNondelegatingAddRef(CICTUNK(This));
    } else if( IsEqualGUID(iid, &IID_IasSystem2)) {
        *ppv = This;
    } else if( IsEqualGUID(iid, &IID_IasSystem3)) {
        *ppv = This;
    } else if( IsEqualGUID(iid, &IID_IasSystem4)) {
        *ppv = This;
    } else if( IsEqualGUID(iid, &IID_IasSystem5)) {
        *ppv = This;
    } else {
        *ppv = NULL;                
        return E_NOINTERFACE;
    }

    if (*ppv) {
        if (pObj->m_pIUnknownOuter == This)
            sysNondelegatingAddRef(CICTUNK_PASSTHRU(This));
        else
            sysAddRef(CICTUNK_PASSTHRU(This));
    }
    //sysNondelegatingAddRef(CICTUNK_PASSTHRU(This));

    return S_OK;
}

HRESULT 
sysInit(CTVOID *This, ICtUnknown *pIUnknownOuter)
{
    PSYSOBJ pObj = This;
    
    //CTDPF("sysInit called\n");  
    if (!pObj) 
        CTRETURN(E_INVALIDARG);
    if (pIUnknownOuter)
        pObj->m_pIUnknownOuter = pIUnknownOuter;

    return S_OK;
}


/////////////////////
// ICtUnknown methods
/////////////////////

HRESULT 
sysQueryInterface(ICTUNK_QIARGLIST)
{
    PSYSOBJ pObj = This;

    //CTDPF("sysQueryInterface called, %d.\n", pObj->m_pIUnknownOuter);
    if (!pObj) 
        CTRETURN(E_INVALIDARG);    
    
    return COMCALL(ICtUnknown, pObj->m_pIUnknownOuter, QueryInterface)(pObj->m_pIUnknownOuter, iid,
            ppv QICTUNK);
}

CTDWORD 
sysAddRef(ICTUNK_IARGLIST)
{
    PSYSOBJ pObj = This;

    // CTDPF("sysAddRef called\n");
    if (!pObj) 
        CTRETURN(E_INVALIDARG);
    
    return COMCALL(ICtUnknown, pObj->m_pIUnknownOuter, AddRef)(CICTUNK(pObj->m_pIUnknownOuter));
}

CTDWORD 
sysRelease(ICTUNK_IARGLIST)
{
    PSYSOBJ pObj = This;

    //CTDPF("sysRelease called\n");
    if (!pObj) { 
        CTRETURN(E_INVALIDARG);
    }
    
    return COMCALL(ICtUnknown, pObj->m_pIUnknownOuter, Release)(CICTUNK(pObj->m_pIUnknownOuter));
}


////////////////////
// IasSystem methods
////////////////////

HRESULT 
sysPostBootInit(CTVOID *This)
{
    //CTDPF("sysPostBootInit called\n");
    
    return S_OK;
}

HRESULT 
sysGetSystemTime(CTVOID *This, CTDWORD *pdwTime)
{
    //CTDPF("sysGetSystemTime called\n");
    // TODO: This may be incorrect for SMP systems
    // We may have to use do_gettimeofday()
    if (!pdwTime) 
        CTRETURN(E_INVALIDARG);
    *pdwTime = (jiffies * 1000) / HZ;

    return S_OK;
}


////////////////////////////
// Timeout and DPC functions
////////////////////////////

typedef struct {
    CTBOOL bNonCriticalTimeoutWork;
    PFCALLBACK fpCallback;
    CTHANDLE hUser;
    struct timer_list TimerList;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    struct tq_struct SchedulerQueue;
#else
    struct work_struct SchedulerQueue;
#endif
} stTimerCbkHandle;

static void 
timer_callback(unsigned long ulData)
{
    stTimerCbkHandle *phTimerCbk = (stTimerCbkHandle *) ulData;

    //PDEBUG("timer_callback called\n");
    if (!phTimerCbk || !phTimerCbk->fpCallback) {
        PDEBUG("Error: Invalid timer callback handle!\n");
        CTASSERT(FALSE);
        return;
    }

    if (phTimerCbk->bNonCriticalTimeoutWork) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
        schedule_task(&phTimerCbk->SchedulerQueue);
#else
        schedule_work(&phTimerCbk->SchedulerQueue);
#endif
    } else {
        phTimerCbk->fpCallback(phTimerCbk->hUser);
    }
}

HRESULT 
sysCreateTimeout(CTVOID *This, OSTIMEOUTHANDLE *phTmo)
{
    struct timer_list *pTimerList;
    stTimerCbkHandle *phTimerCbk;

    CTDPF("sysCreateTimeout called");
    if (!phTmo) 
        CTRETURN(E_INVALIDARG);
    *phTmo = CTHANDLE_NULL;

    phTimerCbk = kmalloc(sizeof(stTimerCbkHandle), GFP_KERNEL);
    if (!phTimerCbk) 
        CTRETURN(E_OUTOFMEMORY);
    memset(phTimerCbk, 0, sizeof(stTimerCbkHandle));
    phTimerCbk->bNonCriticalTimeoutWork = FALSE;

    pTimerList = &phTimerCbk->TimerList;
    init_timer(pTimerList);
    pTimerList->function = timer_callback;
    pTimerList->data = (unsigned long) phTimerCbk;
    
    *phTmo = (OSTIMEOUTHANDLE) phTimerCbk;

    return S_OK;
}

HRESULT 
sysDestroyTimeout(CTVOID *This, OSTIMEOUTHANDLE hTmo)
{
    stTimerCbkHandle *phTimerCbk = (stTimerCbkHandle *) hTmo;
    
    CTDPF("sysDestroyTimeout called");
    if (!phTimerCbk) 
        CTRETURN(E_INVALIDARG);
    del_timer(&phTimerCbk->TimerList);

    // Free allocated resources, never to be called again
    kfree(phTimerCbk);

    return S_OK;
}

HRESULT 
sysScheduleTimeout(CTVOID *This, OSTIMEOUTHANDLE hTmo, PFCALLBACK fpCallback, CTHANDLE hUser, 
        CTDWORD dwMillisecs)
{
    stTimerCbkHandle *phTimerCbk = (stTimerCbkHandle *) hTmo;    

    //CTDPF("sysScheduleTimeout called");
    if (!phTimerCbk) 
        CTRETURN(E_INVALIDARG);
    phTimerCbk->fpCallback = fpCallback;
    phTimerCbk->hUser = hUser;
    phTimerCbk->TimerList.expires = ((dwMillisecs * HZ) / 1000) + jiffies;
    if (timer_pending(&phTimerCbk->TimerList)) 
        return S_OK;
    add_timer(&phTimerCbk->TimerList);

    return S_OK;
}

typedef struct {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    struct tq_struct Task;
#else
    struct work_struct Task;
#endif
    PFCALLBACK fpCallback;
    CTHANDLE hUser;
} stDPCCbkHandle;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
static void task_queue_callback(void *pData)
{
    stDPCCbkHandle *phDPCCbk = pData;
#else
static void task_queue_callback(struct work_struct *work)
{
    stDPCCbkHandle *phDPCCbk = container_of(work, stDPCCbkHandle, Task);
#endif
    //PDEBUG("task_queue_callback called\n");
    if (!phDPCCbk || !phDPCCbk->fpCallback) {
        CTASSERT(0);
        return;
    }

    phDPCCbk->fpCallback(phDPCCbk->hUser);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
HRESULT 
sysCreateDpc(CTVOID *This, OSDPCHANDLE *phDPC)
{
    stDPCCbkHandle *phDPCCbk;
    struct tq_struct *pTask;

    //CTDPF("sysCreateDpc called");
    if (!phDPC) 
        CTRETURN(E_INVALIDARG);
    *phDPC = CTHANDLE_NULL;

    phDPCCbk = kmalloc(sizeof(stDPCCbkHandle), GFP_ATOMIC);
    if (!phDPCCbk) {
        CTRETURN(E_OUTOFMEMORY);
    }
    memset(phDPCCbk, 0, sizeof(stDPCCbkHandle));

    pTask = &(phDPCCbk->Task);
    INIT_LIST_HEAD(&pTask->list);
    pTask->sync = 0;
    pTask->routine = task_queue_callback;
    pTask->data = phDPCCbk;

    *phDPC = (OSDPCHANDLE)phDPCCbk;
    
    return S_OK;
}
#else
HRESULT 
sysCreateDpc(CTVOID *This, OSDPCHANDLE *phDPC)
{
    stDPCCbkHandle *phDPCCbk;

    //CTDPF("sysCreateDpc called");
    if (!phDPC) 
        CTRETURN(E_INVALIDARG);
    *phDPC = CTHANDLE_NULL;

    phDPCCbk = kmalloc(sizeof(stDPCCbkHandle), GFP_ATOMIC);
    if (!phDPCCbk) {
        CTRETURN(E_OUTOFMEMORY);
    }
    memset(phDPCCbk, 0, sizeof(stDPCCbkHandle));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
    INIT_WORK(&(phDPCCbk->Task), (void *)task_queue_callback, phDPCCbk);
#else
    INIT_WORK(&(phDPCCbk->Task), task_queue_callback);
#endif
    *phDPC = (OSDPCHANDLE)phDPCCbk;
    
    return S_OK;
}
#endif

HRESULT 
sysDestroyDpc(CTVOID *This, OSDPCHANDLE hDPC)
{
    stDPCCbkHandle *phDPCCbk = (stDPCCbkHandle *)hDPC;

    //PDEBUG("sysDestroyDpc called: hDPC = 0x%x\n", (unsigned int)hDPC);
    if (!phDPCCbk) 
        CTRETURN(E_INVALIDARG);

    kfree(phDPCCbk);
    
    return S_OK;
}

// This function can be called from an ISR
HRESULT 
sysScheduleDpc(CTVOID *This, OSDPCHANDLE hDPC, PFCALLBACK fpCallback, CTHANDLE hUser)
{
    stDPCCbkHandle *phDPCCbk = (stDPCCbkHandle *)hDPC;

    //CTDPF("sysScheduleDpc called");
    if (!phDPCCbk) 
        CTRETURN(E_INVALIDARG);
    phDPCCbk->fpCallback = fpCallback; 
    phDPCCbk->hUser = hUser;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    queue_task(&phDPCCbk->Task, &tq_immediate);
    mark_bh(IMMEDIATE_BH);
#else 
    schedule_work(&phDPCCbk->Task);
#endif

    return S_OK;
}

HRESULT 
sysCreateInterruptTimeout(CTVOID *This, OSTIMEOUTHANDLE *phTmo)
{
    return sysCreateTimeout(This, phTmo);
}

HRESULT 
sysDestroyInterruptTimeout(CTVOID *This, OSTIMEOUTHANDLE hTmo)
{
    return sysDestroyTimeout(This, hTmo);
}

HRESULT 
sysScheduleInterruptTimeout(CTVOID *This, OSTIMEOUTHANDLE hTmo, PFCALLBACK fpCallback, 
        CTHANDLE hUser, CTDWORD dwMillisecs)
{
    return sysScheduleTimeout(This, hTmo, fpCallback, hUser, dwMillisecs);
}


//////////////////////
// Interrupt functions
//////////////////////

typedef struct tInterruptHandle {
    CTDWORD uIRQ;
    PF_ISR_CALLBACK fpISR;
    CTHANDLE hUser;
    CTBOOL bEnabled;
} stInterruptHandle;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
	typedef void irqreturn_t;
	#define IRQ_NONE
	#define IRQ_HANDLED
	#define IRQ_RETVAL(x)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
static irqreturn_t interrupt_callback(int nIRQ, void *dev_id, struct pt_regs *regs)
#else
static irqreturn_t interrupt_callback(int nIRQ, void *dev_id)
#endif
{
    stInterruptHandle *pCurr = (stInterruptHandle *) dev_id;

    //CTDPF("interrupt_callback called");

    CTASSERT(pCurr);
    if (!pCurr->bEnabled) {
        return IRQ_NONE;
    }

    /* This actually returns a BOOL value to indicate if the 
     * interrupt was handled. */	    
    return pCurr->fpISR(pCurr->hUser) ? IRQ_HANDLED : IRQ_NONE;
}

HRESULT 
sysRegisterInterrupt(CTVOID *This, CTHANDLE *phIrq, CTHANDLE hIrqResource, PF_ISR_CALLBACK fpISR, 
        CTHANDLE hUser)
{
    stInterruptHandle *phInterrupt;

    if (!phIrq) 
        CTRETURN(E_INVALIDARG);
    *phIrq = CTHANDLE_NULL; // Disable first

    phInterrupt = kmalloc(sizeof(stInterruptHandle), GFP_KERNEL);
    if (!phInterrupt) 
        CTRETURN(E_OUTOFMEMORY);
    memset(phInterrupt, 0, sizeof(stInterruptHandle));

    phInterrupt->uIRQ = (CTDWORD)hIrqResource;
    phInterrupt->fpISR = fpISR;
    phInterrupt->hUser = hUser;
    phInterrupt->bEnabled = TRUE;

    if (request_irq(phInterrupt->uIRQ, interrupt_callback, SA_SHIRQ, "x-fi", phInterrupt)) {
        kfree(phInterrupt);
        return E_FAIL;
    }

    *phIrq = (CTHANDLE) phInterrupt;

    return S_OK;
}

HRESULT 
sysUnregisterInterrupt(CTVOID *This, CTHANDLE hIrq)
{
    stInterruptHandle *pCurr = (stInterruptHandle *) hIrq;

    if (!pCurr) 
        CTRETURN(E_INVALIDARG);
    free_irq(pCurr->uIRQ, pCurr);
    kfree(pCurr);

    return S_OK;
}

HRESULT 
sysStartInterrupt(CTVOID *This, CTHANDLE hIrq)
{
    stInterruptHandle *phInterrupt = (stInterruptHandle *) hIrq;

    if (!phInterrupt) 
        CTRETURN(E_INVALIDARG);
    if (!phInterrupt->bEnabled) {
        phInterrupt->bEnabled = TRUE;
        // enable_irq() not allowed for shared interrupts
    }
    
    return S_OK;
}

HRESULT 
sysStopInterrupt(CTVOID *This, CTHANDLE hIrq)
{
    stInterruptHandle *phInterrupt = (stInterruptHandle *) hIrq;

    if (!phInterrupt) 
        CTRETURN(E_INVALIDARG);
    if (phInterrupt->bEnabled) {
        phInterrupt->bEnabled = FALSE;
	    // disable_irq() not allowed for shared interrupts
    }

    return S_OK;
}

HRESULT 
sysAckInterrupt(CTVOID *This, CTHANDLE hIrq)
{
    // TODO: Do nothing?
    return S_OK;
}

// Hardware info functions
HRESULT 
sysGetHardwareInfo(CTVOID *This, CTHANDLE hDevice, stHardwareInfo *pInfo)
{
    struct pci_dev *pDev = (struct pci_dev *) hDevice;
    CTDWORD dwSubsysID;
    CTBYTE byRevID;

    //CTDPF("sysGetHardwareInfo is called");
    if (!pDev || !pInfo) 
        CTRETURN(E_INVALIDARG);
    //
    pInfo->busNum = pDev->bus->number;
    pInfo->funcNum = PCI_FUNC(pDev->devfn);
    pInfo->instance = PCI_SLOT(pDev->devfn); // Instance == Slot
    pInfo->vendorID = pDev->vendor; 
    pInfo->deviceID = pDev->device; 
    pci_read_config_dword(pDev, PCI_SUBSYSTEM_VENDOR_ID, (u32*) &dwSubsysID);
    pInfo->subsys = dwSubsysID;
    pci_read_config_byte(pDev, PCI_REVISION_ID, &byRevID);
    pInfo->rev = byRevID;
    // TODO: get the busType?
    pInfo->busType = HWINFO_BUS_TYPE_PCI;

    return S_OK;
}


////////////////////////////
// Heap management functions
////////////////////////////

HRESULT 
sysHeapFree(CTVOID *This, CTVOID *pMem)
{
    heap_free(pMem);
    
    return S_OK;
}

CTVOID *
sysHeapAllocate(CTVOID *This, CTDWORD dwSize)
{
    return heap_alloc(ALLOCARG(dwSize, TRUE));
}

HRESULT 
sysLockedHeapFree(CTVOID *This, CTVOID *pMem)
{
    heap_free(pMem);

    return S_OK;
}

CTVOID *
sysLockedHeapAllocate(CTVOID *This, CTDWORD dwSize)
{
    return heap_alloc(ALLOCARG(dwSize, TRUE));
}


////////////////////////////
// Page management functions
////////////////////////////

// stMemHandle declared in ctossrv.h as type is exported to main driver

OSPHYSMEMHDL 
sysAllocPages(CTVOID *This, CTDWORD dwBytes, CTDWORD dwFlags)
{
    stMemHandle *phMem;

    //PDEBUG("sysAllocPages called: dwBytes=%u\n", dwBytes);
    phMem = kmalloc(sizeof(stMemHandle), GFP_KERNEL);
    if (!phMem) {
        PDEBUG("Error: sysAllocPages could not allocate kmalloc memory\n");
        return 0;
    }
    memset(phMem, 0, sizeof(stMemHandle));

    phMem->ulOrder = bytes_to_order(dwBytes); 
    phMem->pVirtAddr = vmalloc(dwBytes);
    if (!phMem->pVirtAddr) {
        CTASSERT(0);
        PDEBUG("Error: sysAllocPages could not allocate vmalloc memory\n");
        kfree(phMem);
        return 0;
    }
    phMem->MemType = MEMTYPE_VMALLOC;
    phMem->dwPageRange = dwBytes;

    return ((OSPHYSMEMHDL)phMem);
}

OSPHYSMEMHDL 
sysAllocContiguousPages(CTVOID *This, CTDWORD dwBytes, interface ICtDevice *pDevice)
{
    stMemHandle *phMem;

    phMem = kmalloc(sizeof(stMemHandle), GFP_KERNEL);
    if (!phMem) {
        PDEBUG("Error: sysAllocContiguousPages could not allocate kmalloc memory\n");
        return 0;
    }
    memset(phMem, 0, sizeof(stMemHandle));

    phMem->ulOrder = bytes_to_order(dwBytes); 
    phMem->pVirtAddr = (CTVOID *) __get_free_pages(GFP_KERNEL, phMem->ulOrder);
    if (!phMem->pVirtAddr) {
        kfree(phMem);
        CTASSERT(0);
        PDEBUG("Error: sysAllocContiguousPages could not allocate __get_free_pages memory\n");
        return 0;
    }
    phMem->MemType = MEMTYPE_PAGE;
    phMem->dwPageRange = dwBytes;
    
    return ((OSPHYSMEMHDL)phMem);
}

HRESULT 
sysFreePages(CTVOID *This, OSPHYSMEMHDL hMem)
{
    stMemHandle *phMem = (stMemHandle *) hMem;

    if (!phMem) 
        CTRETURN(E_INVALIDARG);

    // Unmark reserved pages, if any
    unreserve_pages(phMem, PAGE_ALIGN(phMem->dwPageRange));

    switch (phMem->MemType) {
        case MEMTYPE_PAGE:
            free_pages((unsigned long)phMem->pVirtAddr, phMem->ulOrder);
            break;
        case MEMTYPE_VMALLOC:
            vfree(phMem->pVirtAddr);
            break;
        case MEMTYPE_KMALLOC:
        default:
            CTRETURN(E_FAIL);
            break;
    }
    kfree(phMem);

    return S_OK;
}

HRESULT 
sysLockPages(CTVOID *This, OSPHYSMEMHDL hMem)
{
    // Pages are locked by default
    return S_OK;
}

HRESULT 
sysUnlockPages(CTVOID *This, OSPHYSMEMHDL hMem)
{
    // Pages are locked by default
    return S_OK;
}

OSVIRTADDR 
sysGetPageVirtAddr(CTVOID *This, OSPHYSMEMHDL hMem)
{
    stMemHandle *phMem = (stMemHandle *) hMem;

    //PDEBUG("sysGetPageVirtAddr() called\n");
    if (!phMem) {
        PDEBUG("Error: sysGetPageVirtAddr() called with invalid handle\n");
        return NULL;
    }
    
    return phMem->pVirtAddr;
}

OSVIRTADDR 
sysGetPageUserModeVirtAddr(CTVOID *This, OSPHYSMEMHDL hMem)
{
    stMemHandle *phMem = (stMemHandle *) hMem;
    long lAddr; // Signed to catch errors
    pgprot_t prot = __S111; // shared rwx
 
    //PDEBUG("sysGetPageUserModeVirtAddr() called\n");
    if (!phMem) {
        CTASSERT(0);
        return NULL;
    }
 
    if (phMem->bMapped) { // Previously called
        // TODO: Find a way to handle threads nicely
        if (current->pid == phMem->pid)
            return (OSVIRTADDR) phMem->ulUserAddr;
        else {
            CTASSERT(0);
            return NULL;
        }
    }

    // Setup VMA
    down_write(&current->mm->mmap_sem);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 20)
#define MAP_ANON MAP_ANONYMOUS
#endif
    lAddr = do_mmap(NULL, 0, phMem->dwPageRange, (PROT_READ|PROT_WRITE), (MAP_SHARED|MAP_ANON), 0);
    up_write(&current->mm->mmap_sem);
    if (lAddr < 0) {
        CTDPF("do_mmap failed, error=%d ", (int)lAddr);
        CTASSERT(0);
        return NULL;
    }

    // Mark pages as reserved and remap into VMA
    if (FAILED(reserve_remap_pages(lAddr, phMem, phMem->dwPageRange, prot))) {
        down_write(&current->mm->mmap_sem);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 21)
        do_munmap(current->mm, lAddr, phMem->dwPageRange, 1);
#else
        do_munmap(current->mm, lAddr, phMem->dwPageRange);
#endif
        up_write(&current->mm->mmap_sem);
        CTDPF("reserve_remap_pages failed");
        CTASSERT(0);
        return NULL;
    }
    phMem->pid = current->pid;
    phMem->ulUserAddr = lAddr;
    phMem->bMapped = TRUE;
    
    return (OSVIRTADDR) lAddr;
}

OSPHYSADDR 
sysGetPagePhysAddr(CTVOID *This, OSPHYSMEMHDL hMem, CTDWORD dwOffset)
{
    stMemHandle *phMem = (stMemHandle *) hMem;

    PDEBUG("sysGetPagePhysAddr() is called\n");
    
    if (!phMem) {
        PDEBUG("Error: sysGetPagePhysAddr() called with invalid handle\n");
        return 0;
    }
    
    switch (phMem->MemType) {    
        case MEMTYPE_PAGE:
            return virt_to_phys(phMem->pVirtAddr + dwOffset);
        case MEMTYPE_VMALLOC:
            return kvirt_to_phys((phMem->pVirtAddr + dwOffset));
            //return kvirt_to_phys((unsigned long) (phMem->pVirtAddr + dwOffset));
        case MEMTYPE_KMALLOC:
            PDEBUG("Error: sysGetPagePhysAddr() called with KMALLOC_MEMTYPE\n");
            return 0;
        default:
            PDEBUG("Error: sysGetPagePhysAddr() called with unknown MEMTYPE\n");
            return 0;
    }
}

OSBUSADDR 
sysGetPageBusAddr(CTVOID *This, OSPHYSMEMHDL hMem, CTDWORD dwOffset, interface ICtDevice *pDevice)
{
    stMemHandle *phMem = (stMemHandle *) hMem;

//    PDEBUG("sysGetPageBusAddr() is called\n");
    
    if (!phMem) {
        PDEBUG("Error: sysGetPageBusAddr() called with invalid handle\n");
        return 0;
    }
    
    switch (phMem->MemType) {
        case MEMTYPE_PAGE:
            return virt_to_bus(phMem->pVirtAddr + dwOffset);
        case MEMTYPE_VMALLOC:
            return kvirt_to_bus((phMem->pVirtAddr + dwOffset));
            //return kvirt_to_bus((unsigned long) (phMem->pVirtAddr + dwOffset));
        case MEMTYPE_KMALLOC:
            PDEBUG("Error: sysGetPageBusAddr() called with KMALLOC_MEMTYPE\n");
            return 0;
        default:
            PDEBUG("Error: sysGetPageBusAddr() called with unknown MEMTYPE\n");
            return 0;
    }

}

CTDWORD 
sysGetPageRangeSize(CTVOID *This, OSPHYSMEMHDL hMem)
{
    stMemHandle *phMem = (stMemHandle *) hMem;

    if (!phMem) {
        PDEBUG("Error: sysGetPageVirtAddr() called with invalid handle\n");
        return 0;
    }

    return phMem->dwPageRange;
}

CTVOID 
sysFlushPageCache(CTVOID *This, OSPHYSMEMHDL hMem, CTDWORD dwMode)
{
    switch (dwMode) {
        case CTOS_CACHE_INVALIDATE:
            // TODO: How to implement this?
            break;
        case CTOS_CACHE_WRITEBACK:
            flush_cache_all(); 
            // Consider switching to flush_cache_range for finer granularity
            break;
        default:
            break;
    }
}


///////////////////////////
// Some weird Windows stuff
///////////////////////////

HRESULT 
sysMapFlat16BitPointer(CTVOID *This, CTDWORD p16, CTDWORD *pdwFlat)
{
    return E_FAIL; // Specific to Windows - should fail
}


///////////////////////////////////
// File path manipulation functions
///////////////////////////////////

CTULONG 
sysGetPathLength(CTVOID *This, CTPATH *pszString)
{
    if (!pszString) {
        PDEBUG("WARNING: sysGetPathLength() encountered NULL string\n");
        return 0;
    }
    
    return strlen(pszString);
}

CTBOOL 
sysIsFullPath(CTVOID *This, CTPATH * pszString)
{
    CTASSERT(pszString);
    if (strlen(pszString) > 0 && pszString[0] == '/') 
        return TRUE;
    if (strlen(pszString) > 4 && pszString[4] == ':') 
        return TRUE;
    
    return FALSE;
}

CTPATH *
sysGetSystemStylePathName(CTVOID *This, CTPATH* pszString, PCTDWORD pBufSize)
{
    CTCHAR pPath[MAX_PATH], *pChar;

    if (!pszString || strlen(pszString) >= MAX_PATH) {
        CTASSERT(0);
        return pszString;
    }

    if (strnicmp(pszString, "file://", 7) == 0) { // URL style
        pChar = pszString + 7;
        while (strncmp(pChar, "//", 2) == 0) 
            pChar++;
        strcpy(pPath, pChar);	
        strcpy(pszString, pPath);
    }
    
    return pszString;
}

CTBOOL 
sysArePathsIdentical(CTVOID *This, CTPATH * pszString1, CTPATH * pszString2)
{
    CTCHAR pPath1[MAX_PATH], pPath2[MAX_PATH];
    CTDWORD dwSize1 = sizeof(pPath1), dwSize2 = sizeof(pPath2);

    if (!pszString1 || !pszString2 || strlen(pszString1) >= MAX_PATH || strlen(pszString2) 
            >= MAX_PATH) {
        CTASSERT(0);
        return FALSE;
    }

    strcpy(pPath1, pszString1);
    strcpy(pPath2, pszString2);

    return (strcmp(sysGetSystemStylePathName(This, pPath1, &dwSize1), 
                sysGetSystemStylePathName(This, pPath2, &dwSize2)) == 0);
}

HRESULT 
sysGetDriverDataDirectory(CTVOID *This, CTPATH* pszString, CTDWORD bufSize )
{
    if (bufSize < (strlen("/etc/creative") + 1))
        return E_FAIL;
    strcpy(pszString, "/etc/creative");

    return S_OK;
}

HRESULT 
sysGetDriverPreferenceDirectory(CTVOID *This, CTPATH* pszString, CTDWORD bufSize )
{
    return sysGetDriverDataDirectory(This, pszString, bufSize);
}

CTPATH 
sysGetPathDelimiter(CTVOID *This)
{
    return '/';
}

CTBOOL 
sysIsFileSystemCaseSensitive(CTVOID *This)
{
    return TRUE;
}

CTBOOL 
sysIsFileSystemWideChar(CTVOID *This)
{

    return FALSE;
}

HRESULT 
sysASCIIPathToWideChar(CTVOID *This, CTUSHORT *wideCharPath, const CTCHAR *asciiPath, 
        PCTDWORD pWideCharPathBufferSize, CTBOOL bDoSystemStyle)
{
    CTASSERT(0);

    return E_FAIL;
}

HRESULT 
sysWideCharPathtoASCII(CTVOID *This, CTCHAR *asciiPath, const CTUSHORT *wideCharPath, 
        PCTDWORD pAsciiPathBufferSize, CTBOOL bDoSystemStyle)
{
    CTASSERT(0);

    return E_FAIL;
}

HRESULT 
sysASCIIPathToCTPATH(CTVOID *This, CTPATH *ctPath, const CTCHAR *asciiPath, 
        PCTDWORD pCtPathBufferSize, CTBOOL bDoSystemStyle)
{
    if (!ctPath || !asciiPath || !pCtPathBufferSize) {
        CTASSERT(0);
        return E_INVALIDARG;
    }

    if (*pCtPathBufferSize < strlen(asciiPath) + 1) {
        ctPath[0] = '\0';
        *pCtPathBufferSize = strlen(asciiPath) + 1;
        return E_FAIL;
    }
    
    strcpy(ctPath, asciiPath);

    if (bDoSystemStyle)
        sysGetSystemStylePathName(This, ctPath, pCtPathBufferSize);

    return S_OK;
}

HRESULT 
sysWideCharPathToCTPATH(CTVOID *This, CTPATH *ctPath, const CTUSHORT *wideCharPath, 
        PCTDWORD pCtPathBufferSize, CTBOOL bDoSystemStyle)
{
    CTASSERT(0);
 
    return E_FAIL;
}


/////////////////////
// Spinlock functions
/////////////////////

typedef struct {
    spinlock_t Lock; // Per-instance lock
    unsigned long ulFlags;
} tSpinLock;

HRESULT 
sysCreateSpinLock(CTVOID *This, CTHANDLE *phLock, interface ICtDevice *pDevice)
{
    tSpinLock *pLock;
    int priority;

    //CTDPF("sysCreateSpinLock is called");
    if (!phLock) 
        CTRETURN(E_INVALIDARG);
    *phLock = CTHANDLE_NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    priority = in_interrupt() ? GFP_ATOMIC : GFP_KERNEL;
#else
    priority = (in_interrupt() || in_atomic() || irqs_disabled()) ? GFP_ATOMIC : GFP_KERNEL;
#endif

    pLock = kmalloc(sizeof(tSpinLock), priority);
    if (!pLock) 
        CTRETURN(E_OUTOFMEMORY);
    memset(pLock, 0, sizeof(tSpinLock));

    pLock->Lock = SPIN_LOCK_UNLOCKED;
    pLock->ulFlags = 0;

    *phLock = (CTHANDLE) pLock;

    return S_OK;
}

HRESULT 
sysDestroySpinLock(CTVOID *This, CTHANDLE hLock)
{
    tSpinLock *pLock = (tSpinLock *) hLock;

    //CTDPF("sysDestroySpinLock is called");
    if (!pLock) 
        CTRETURN(E_INVALIDARG);
    kfree(pLock);

    return S_OK;
}

HRESULT 
sysAcquireSpinLock(CTVOID *This, CTHANDLE hLock)
{
    tSpinLock *pLock = (tSpinLock *) hLock;

    //PDEBUG("sysAcquireSpinLock called\n");
    if (!pLock) 
        CTRETURN(E_INVALIDARG);
    spin_lock_irqsave(&pLock->Lock, pLock->ulFlags);

    return S_OK;
}

HRESULT 
sysReleaseSpinLock(CTVOID *This, CTHANDLE hLock)
{
    tSpinLock *pLock = (tSpinLock *) hLock;

    //PDEBUG("sysReleaseSpinLock called\n");
    if (!pLock) 
        CTRETURN(E_INVALIDARG);
    spin_unlock_irqrestore(&pLock->Lock, pLock->ulFlags);

    return S_OK;
}


//////////////////
// Mutex functions
//////////////////

typedef struct {
    struct semaphore tSem;
    pid_t pidOwner;
    CTDWORD dwRef;
} tMutex;

HRESULT 
sysCreateMutex(CTVOID *This, OSMUTEXHDL *phMutex)
{
    tMutex *pMutex;

    //PDEBUG("sysCreateMutex called\n");
    if (!phMutex) 
        CTRETURN(E_INVALIDARG);
    *phMutex = CTHANDLE_NULL;

    pMutex = kmalloc(sizeof(tMutex), GFP_KERNEL);
    if (!pMutex) 
        CTRETURN(E_OUTOFMEMORY);
    memset(pMutex, 0, sizeof(tMutex));

    init_MUTEX(&pMutex->tSem);
    pMutex->pidOwner = 0; // Assume app pid will be nonzero
    pMutex->dwRef = 0;

    *phMutex = (OSMUTEXHDL) pMutex;

    return S_OK;
}

HRESULT 
sysAcquireMutex(CTVOID *This, OSMUTEXHDL hMutex)
{
    tMutex *pMutex = (tMutex *) hMutex;

    //PDEBUG("sysAcquireMutex called\n");
    if (!pMutex) 
        CTRETURN(E_INVALIDARG);
    if (pMutex->pidOwner != current->pid) {
        if (down_trylock(&pMutex->tSem) != 0) 
            return E_FAIL;
        pMutex->pidOwner = current->pid;
    }
    pMutex->dwRef++;

    return S_OK;
}

HRESULT 
sysReleaseMutex(CTVOID *This, OSMUTEXHDL hMutex)
{
    tMutex *pMutex = (tMutex *) hMutex;

    //PDEBUG("sysReleaseMutex called\n");
    if (!pMutex) 
        CTRETURN(E_INVALIDARG);
    if (pMutex->pidOwner != current->pid // Current process not the owner
            || pMutex->dwRef == 0) // Mutex not acquired
        return E_FAIL; 
    
    if (--pMutex->dwRef == 0) {
        pMutex->pidOwner = 0; // No owner
        up(&pMutex->tSem);
    }

    return S_OK;
}

HRESULT 
sysWaitMutex(CTVOID *This, OSMUTEXHDL hMutex)
{
    tMutex *pMutex = (tMutex *) hMutex;

    //PDEBUG("sysWaitMutex called\n");
    if (!pMutex) 
        CTRETURN(E_INVALIDARG);
    if (pMutex->pidOwner != current->pid) {
        down_interruptible(&pMutex->tSem);
        if (signal_pending(current)) {
            down_trylock(&pMutex->tSem);
            up(&pMutex->tSem);
            CTRETURN(E_FAIL);
        }
        pMutex->pidOwner = current->pid;
    }
    pMutex->dwRef++;

    return S_OK;
}

HRESULT 
sysDeleteMutex(CTVOID *This, OSMUTEXHDL hMutex)
{
    tMutex *pMutex = (tMutex *) hMutex;
    
    //PDEBUG("sysDeleteMutex called\n");
    if (!pMutex) 
        CTRETURN(E_INVALIDARG);
    kfree(pMutex);
    
    return S_OK;
}


//////////////////////
// Semaphore functions
//////////////////////

typedef struct {
    struct semaphore tSem; // This must always be the first element
    CTBOOL bAbort;
} stSemHandle;

HRESULT 
sysCreateSemaphore(CTVOID *This, OSSEMAHDL *phNewSema, CTDWORD dwInitialCount)
{
    stSemHandle *pSemHdl;
    int priority;

    if (!phNewSema) CTRETURN(E_INVALIDARG);
    *phNewSema = CTHANDLE_NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    priority = in_interrupt() ? GFP_ATOMIC : GFP_KERNEL;
#else
    priority = (in_interrupt() || in_atomic() || irqs_disabled()) ? GFP_ATOMIC : GFP_KERNEL;
#endif

    pSemHdl = kmalloc(sizeof(stSemHandle), priority);
    if (!pSemHdl) 
        CTRETURN(E_OUTOFMEMORY);
    memset(pSemHdl, 0, sizeof(stSemHandle));

    sema_init(&pSemHdl->tSem, dwInitialCount);
    
    *phNewSema = pSemHdl;
    return S_OK;
}

HRESULT 
sysWaitSemaphore(CTVOID *This, OSSEMAHDL hSema)
{
    stSemHandle *pSemHdl = hSema;

    if (!pSemHdl) 
        CTRETURN(E_INVALIDARG);
    down_interruptible(&pSemHdl->tSem);
    if (signal_pending(current))
        CTRETURN(E_FAIL);
    
    return pSemHdl->bAbort ? E_FAIL : S_OK;
}

HRESULT 
sysSignalSemaphore(CTVOID *This, OSSEMAHDL hSema)
{
    stSemHandle *pSemHdl = hSema;

    if (!pSemHdl) 
        CTRETURN(E_INVALIDARG);
    up(&pSemHdl->tSem);

    return S_OK;
}

HRESULT 
sysDestroySemaphore(CTVOID *This, OSSEMAHDL hSema)
{
    stSemHandle *pSemHdl = hSema;

    if (!pSemHdl) 
        CTRETURN(E_INVALIDARG);
    // Free all blocked threads 
    while (down_trylock(&pSemHdl->tSem) != 0) {
        pSemHdl->bAbort = TRUE;
        up(&pSemHdl->tSem);
    }
    up(&pSemHdl->tSem); // To balance successfully locked semaphore
    kfree(pSemHdl);

    return S_OK;
}


/////////////////////
// File I/O functions
/////////////////////

// The following is defined in /usr/include/unistd.h and /usr/include/stdio.h
// Unfortunately, neither file is included here, so we have to define our own
#ifndef SEEK_SET
#   define SEEK_SET        0       /* Seek from beginning of file.  */
#endif
#ifndef SEEK_CUR
#   define SEEK_CUR        1       /* Seek from current position.  */
#endif
#ifndef SEEK_END
#   define SEEK_END        2       /* Seek from end of file.  */
#endif

int errno __attribute__ ((nocommon)); /* Needs to be declared for file I/O */

typedef struct {
    int nFlags;
    int nMode;
    char szPath[MAX_PATH_LEN];
} stFileHdl;

HRESULT 
sysOpenCreateFile(CTVOID *This, CTWORD wAttribFlags, CTPATH *szFileName, CTHANDLE *phFile)
{
    int nFlags, nMode = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    int nFD;
#else
    struct file* filp;
#endif
    mm_segment_t fs = get_fs();    
    stFileHdl *pFileHdl;

    //PDEBUG("sysOpenCreateFile called: FileName = %s\n", szFileName);
    if (!phFile) 
        CTRETURN(E_INVALIDARG);
    *phFile = CTHANDLE_NULL;
    
    switch (wAttribFlags) {
        case (CT_FILE_READ_ACCESS | CT_FILE_WRITE_ACCESS):
            PDEBUG("Read/Write access requested: szFileName=%s\n", szFileName);
            nFlags = O_RDWR | O_CREAT;
            nMode = S_IRUGO | S_IWUGO;
            break;
        case CT_FILE_READ_ACCESS:
            PDEBUG("Read access requested: szFileName=%s\n", szFileName);
            nFlags = O_RDONLY;
            break;
        case CT_FILE_WRITE_ACCESS:
            PDEBUG("Write access requested: szFileName=%s\n", szFileName);
            nFlags = O_WRONLY | O_CREAT;
            nMode = S_IRUGO | S_IWUGO;
            break;
        default:
            PDEBUG("Error: sysOpenCreateFile encountered unknown attrib flag: szFileName=%s\n", szFileName);
            return E_INVALIDARG;
    }
    
    set_fs(get_ds());
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    nFD = open(szFileName, nFlags, nMode);
    if (nFD != -1) {
        close(nFD);
    }
#else
    filp = filp_open(szFileName, nFlags, nMode);
    if (!IS_ERR(filp)) {
        filp_close(filp, current->files);
    }
#endif
    set_fs(fs);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    if (nFD == -1) {
#else
    if (IS_ERR(filp)) {
#endif
        PDEBUG("sysOpenCreateFile() failed, name=%s\n", szFileName);
        return E_FAIL;
    }

    pFileHdl = kmalloc(sizeof(stFileHdl), GFP_KERNEL);
    if (!pFileHdl) 
        CTRETURN(E_OUTOFMEMORY);
    memset(pFileHdl, 0, sizeof(stFileHdl));
    
    pFileHdl->nFlags = nFlags;
    pFileHdl->nMode = nMode;
    strcpy(pFileHdl->szPath, szFileName);

    *phFile = (CTHANDLE) pFileHdl;

    return S_OK;
}

HRESULT 
sysGetFileSize(CTVOID *This, CTDWORD *pdwSize, CTHANDLE hFile)
{
    stFileHdl *pFileHdl = (stFileHdl *) hFile;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    int nFD;
    mm_segment_t fs = get_fs();
#else
    struct file* filp;
#endif
    off_t offset = -1;

    /*
    PDEBUG("sysGetFileSize called: hfile=0x%x\n", (CTDWORD) hFile);
    CTDPF("Calling open(), path=%s, flags=%d, mode=%d ", pFileHdl->szPath, pFileHdl->nFlags, pFileHdl->nMode);
    */
    
    if (!pFileHdl) 
        CTRETURN(E_INVALIDARG);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    set_fs(get_ds());
    nFD = open(pFileHdl->szPath, pFileHdl->nFlags, pFileHdl->nMode);
    if (nFD != -1) {
        offset = lseek(nFD, 0, SEEK_END);
        close(nFD);
    }
    set_fs(fs);
#else
    filp = filp_open(pFileHdl->szPath, pFileHdl->nFlags, pFileHdl->nMode);
    if (!IS_ERR(filp)) {
        offset = filp->f_op->llseek(filp, 0, SEEK_END);
        filp_close(filp, current->files);
    }
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    if (nFD == -1 || offset == -1) {
#else
    if (IS_ERR(filp) || offset == -1) {
#endif
        PDEBUG("sysGetFileSize failed\n");
        return E_FAIL;
    }

    *pdwSize = (CTDWORD) offset;

    return S_OK;
}

HRESULT 
sysWriteFile(CTVOID *This, CTDWORD dwCount, CTDWORD dwPos, const CTBYTE* pBuffer, CTHANDLE hFile, 
        CTDWORD *pdwBytesWritten)
{
    stFileHdl *pFileHdl = (stFileHdl *) hFile;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    int nFD;
    mm_segment_t fs = get_fs();
#else
    struct file* filp;
#endif
    int nBytesWritten = -1;
    loff_t offset = -1;

    //PDEBUG("sysWriteFile called\n");
    if (!pFileHdl) 
        CTRETURN(E_INVALIDARG);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    set_fs(get_ds());
    nFD = open(pFileHdl->szPath, pFileHdl->nFlags, pFileHdl->nMode);
    if (nFD != -1) {
        offset = lseek(nFD, dwPos, SEEK_SET);
        nBytesWritten = write(nFD, pBuffer, dwCount);
        close(nFD);
    }
    set_fs(fs);
#else
    filp = filp_open(pFileHdl->szPath, pFileHdl->nFlags, pFileHdl->nMode);
    if (!IS_ERR(filp)) {
        offset = filp->f_op->llseek(filp, dwPos, SEEK_SET);
        nBytesWritten = filp->f_op->write(filp, pBuffer, dwCount, &filp->f_pos);
        filp_close(filp, current->files);
    }
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    if (nFD == -1 || offset == -1 || nBytesWritten == -1) {
#else
    if (IS_ERR(filp) || offset == -1 || nBytesWritten == -1) {
#endif
        //PDEBUG("sysWriteFile failed\n");
        return E_FAIL;
    }

    *pdwBytesWritten = (CTDWORD) nBytesWritten;
    
    return S_OK;
}

HRESULT 
sysReadFile(CTVOID *This, CTDWORD dwCount, CTDWORD dwPos, CTBYTE *pBuffer, CTHANDLE hFile, 
        CTDWORD *pdwBytesRead)
{
    stFileHdl *pFileHdl = (stFileHdl *) hFile;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    int nFD;
#else
    struct file* filp;
#endif
    CTDWORD dwBytesRead=-1;
    off_t offset=-1;
    mm_segment_t fs = get_fs();
 
    //PDEBUG("sysReadFile called: count=%u, pos=%u, hfile=0x%x\n", dwCount, dwPos, (CTDWORD) hFile);
    if (!pFileHdl) 
        CTRETURN(E_INVALIDARG);

    set_fs(get_ds());
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    nFD = open(pFileHdl->szPath, pFileHdl->nFlags, pFileHdl->nMode);
    if (nFD != -1) {
        offset = lseek(nFD, dwPos, SEEK_SET);
        dwBytesRead = read(nFD, pBuffer, dwCount);
        close(nFD);
    }
#else
    filp = filp_open(pFileHdl->szPath, pFileHdl->nFlags, pFileHdl->nMode);
    if (!IS_ERR(filp)) {
        offset = filp->f_op->llseek(filp, dwPos, SEEK_SET);
        dwBytesRead = filp->f_op->read(filp, pBuffer, dwCount, &filp->f_pos);
        filp_close(filp, current->files);
    }
#endif
    set_fs(fs);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    if (nFD == -1 || offset == -1 || dwBytesRead == -1) {
#else
    if (IS_ERR(filp) || offset == -1 || dwBytesRead == -1) {
#endif
        PDEBUG("sysReadFile failed\n");
        return E_FAIL;
    }

    *pdwBytesRead = (CTDWORD) dwBytesRead;
    
    return S_OK;
}

HRESULT 
sysCloseFile(CTVOID *This, CTHANDLE hFile)
{
    stFileHdl *pFileHdl = (stFileHdl *) hFile;

    //PDEBUG("sysCloseFile called\n");
    if (!pFileHdl) 
        CTRETURN(E_INVALIDARG);

    kfree(pFileHdl);

    return S_OK;
}


////////////////////
// FP functions
////////////////////

double 
sys_acos(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(acos)(dbArg);
#endif
}

double 
sys_asin(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(asin)(dbArg);
#endif
}

double 
sys_atan(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(atan)(dbArg);
#endif
}

double 
sys_atan2(CTVOID *This, double dbArg1, double dbArg2)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(atan2)(dbArg1, dbArg2);
#endif
}

double 
sys_cos(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(cos)(dbArg);
#endif
}

double 
sys_cosh(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(cosh)(dbArg);
#endif
}

double 
sys_exp(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(exp)(dbArg);
#endif
}

double 
sys_fabs(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(fabs)(dbArg);
#endif
}

double 
sys_fmod(CTVOID *This, double dbArg1, double dbArg2)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(fmod)(dbArg1, dbArg2);
#endif
}

double 
sys_log(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(log)(dbArg);
#endif
}

double 
sys_log10(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(log10)(dbArg);
#endif
}

double 
sys_pow(CTVOID *This, double dbArg1, double dbArg2)
{
#ifdef NO_MATH
return 1.0;
#else
    return FPCALL(pow)(dbArg1, dbArg2);
#endif
}

double 
sys_sin(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(sin)(dbArg);
#endif
}

double 
sys_sinh(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(sinh)(dbArg);
#endif
}

double 
sys_tan(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(tan)(dbArg);
#endif
}

double
sys_tanh(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(tanh)(dbArg);
#endif
}

double
sys_sqrt(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 1.0;
#else
    return FPCALL(sqrt)(dbArg);
#endif
}

double 
sys_ceil(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(ceil)(dbArg);
#endif
}

double 
sys_floor(CTVOID *This, double dbArg)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(floor)(dbArg);
#endif
}

double 
sys_frexp(CTVOID *This, double dbArg1, int *pnArg2)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(frexp)(dbArg1, pnArg2);
#endif
}

double 
sys_ldexp(CTVOID *This, double dbArg1, int nArg2)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(ldexp)(dbArg1, nArg2);
#endif
}

double 
sys_modf(CTVOID *This, double dbArg1, double *pdbArg2)
{
#ifdef NO_MATH
return 0.0;
#else
    return FPCALL(modf)(dbArg1, pdbArg2);
#endif
}

double 
sysGetHugeNum(CTVOID *This)
{
    return HUGE_VAL;
}


CTBOOL 
sys_finite(CTVOID *This, double x)
{
    int32_t hx;

    GET_HIGH_WORD(hx,x);

    return (int)((u_int32_t)((hx&0x7fffffff)-0x7ff00000)>>31);
}

CTBOOL
sys_isnan(CTVOID *This, double x)
{
    int32_t hx,lx;

    EXTRACT_WORDS(hx,lx,x);
    hx &= 0x7fffffff;
    hx |= (u_int32_t)(lx|(-lx))>>31;
    hx = 0x7ff00000 - hx;

    return (int)((u_int32_t)(hx))>>31;
}

HRESULT
sysGetFPUBufSize(CTVOID *This, PCTDWORD pdwBytes)
{
#if defined(i386) || defined(__x86_64__)
    *pdwBytes = FPUBUFSIZE;
    
    return S_OK;
#else
#warning sysGetFPUBufSize not defined under this platform
    return E_FAIL;
#endif
}

HRESULT 
sysSaveFPU(CTVOID *This, PCTVOID pvFPUBuffer)
{
#ifdef NO_MATH
    return S_OK;
#elif defined(i386) || defined(__x86_64__)
  #ifdef USE_NATIVE_MATH
    char *fpu_save = pvFPUBuffer;
    unsigned long flags;

    #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    save_flags(flags);
    cli();
    #else
    //local_save_flags(flags);
    spinlock_t mr_lock = SPIN_LOCK_UNLOCKED;
    spin_lock_irqsave(&mr_lock, flags);
    #endif
    if(!(current->flags & PF_USEDFPU))
        __asm__ __volatile__ ( "clts;\n");
    __asm__ __volatile__ ("fsave %0; fwait\n"::"m"(fpu_save[0]));
    #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    restore_flags(flags);
    #else
    //local_irq_restore(flags);
    spin_unlock_irqrestore(&mr_lock, flags);
    #endif
    
    return S_OK;
  #else
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
    get_cpu();
    #endif

    fp_SaveFPU(pvFPUBuffer);
    
    return S_OK;
  #endif // USE_NATIVE_MATH
#else
#   warning sysSaveFPU not defined under this platform

    return E_FAIL;
#endif
}

HRESULT
sysRestoreFPU(CTVOID *This, PCTVOID pvFPUBuffer)
{
#ifdef NO_MATH
    return S_OK;
#elif defined(i386) || defined(__x86_64__)
  #ifdef USE_NATIVE_MATH
    char *fpu_save = pvFPUBuffer;
    unsigned long flags;

    #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    save_flags(flags);
    cli();
    #else
    //local_save_flags(flags);
    spinlock_t mr_lock = SPIN_LOCK_UNLOCKED;
    spin_lock_irqsave(&mr_lock, flags);
    #endif
    __asm__ __volatile__ ("frstor %0\n"::"m"(fpu_save[0]));
    if(!(current->flags & PF_USEDFPU))
        stts();
    #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    restore_flags(flags);
    #else
    //local_irq_restore(flags);
    spin_unlock_irqrestore(&mr_lock, flags);
    #endif
    return S_OK;
  #else
    
    fp_RestoreFPU(pvFPUBuffer);
    
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
    put_cpu();
    #endif
    
    return S_OK;
  #endif // USE_NATIVE_MATH
#else
#   warning sysRestoreFPU not defined under this platform

    return E_FAIL;
#endif
}


////////////////////////////////////////
// Persistent settings manager functions
////////////////////////////////////////

IasDevReg *g_pIasDevReg = NULL;

CTDWORD
uda_to_win(ISYSASSET_TYPE iType)
{
    switch(iType) {
        case ISYSASSET_DWORD:  
            return REG_DWORD;
        case ISYSASSET_SZ:
            return REG_SZ;
        case ISYSASSET_BINARY: 
            return REG_BINARY;    
        default:
            CTASSERT(0);
            return REG_DWORD;
    }
}

ISYSASSET_TYPE
win_to_uda(CTDWORD dwType)
{
    switch(dwType) {
        case REG_DWORD:  
            return ISYSASSET_DWORD;
        case REG_SZ:     
            return ISYSASSET_SZ;
        case REG_BINARY: 
            return ISYSASSET_BINARY;
        default:
            CTASSERT(0);
            return ISYSASSET_DWORD;
    }
}

HRESULT
sysOpenKey(CTVOID *This, CTHANDLE hTopKey, const PCTCHAR szRelative, PCTHANDLE phKey)
{
    CTDWORD dwResult;

    if (!g_pIasDevReg) 
        return E_FAIL;

    //PDEBUG("sysOpenKey called: szRelative=%s, hTopKey = %x, phKey = %x\n", szRelative, hTopKey, phKey);
    if (hTopKey == HSYSCTKEY_ROOT || hTopKey == CTHANDLE_NULL) {
        hTopKey = (CTHANDLE) COMCALL(IasDevReg, g_pIasDevReg, GetRootHKEY)(g_pIasDevReg);    
    }
    
    dwResult = COMCALL(IasDevReg, g_pIasDevReg, RegOpenKey)(g_pIasDevReg, (__HKEY) hTopKey, szRelative, (__PHKEY) phKey);

    if (dwResult != ERROR_SUCCESS) {
        PDEBUG("IasDevReg::RegOpenKey failed: szRelative=%s\n", szRelative);
        return E_FAIL;
    }

    return S_OK;
}

HRESULT
sysEnumKey(CTVOID *This, CTHANDLE hTopKey, CTDWORD dwIterator, PCTCHAR szRelative, CTDWORD dwSize)
{
    CTDWORD dwResult;

    if (!g_pIasDevReg) 
        return E_FAIL;

    //PDEBUG("sysEnumKey called: szRelative=%s\n", szRelative);
    if (hTopKey == HSYSCTKEY_ROOT || hTopKey == CTHANDLE_NULL) 
        hTopKey = (CTHANDLE) COMCALL(IasDevReg, g_pIasDevReg, GetRootHKEY)(g_pIasDevReg);

    dwResult = COMCALL(IasDevReg, g_pIasDevReg, RegEnumKey)(g_pIasDevReg, (__HKEY) hTopKey, 
            dwIterator, szRelative, dwSize);

    if (dwResult != ERROR_SUCCESS) {
        PDEBUG("IasDevReg::RegEnumKey failed: szRelative=%s\n", szRelative);
        return E_FAIL;
    }

    return S_OK;
}

HRESULT 
sysEnumValue(CTVOID *This, CTHANDLE hKey, CTDWORD dwIterator, PCTCHAR szValue, PCTDWORD pdwNameSize,
        ISYSASSET_TYPE *pType, PCTVOID pBuffer, PCTDWORD pdwDataSize)
{
    CTDWORD dwResult, dwType;

    if (!g_pIasDevReg) 
        return E_FAIL;

    //PDEBUG("sysEnumValue called\n");
    if (hKey == HSYSCTKEY_ROOT || hKey == CTHANDLE_NULL) 
    hKey = (CTHANDLE) COMCALL(IasDevReg, g_pIasDevReg, GetRootHKEY)(g_pIasDevReg);

    dwResult = COMCALL(IasDevReg, g_pIasDevReg, RegEnumValue)(g_pIasDevReg, (__HKEY) hKey, 
            dwIterator, szValue, pdwNameSize, NULL, &dwType, pBuffer, pdwDataSize); 

    if (dwResult != ERROR_SUCCESS) {
        PDEBUG("IasDevReg::RegEnumValue failed: szValue=%s\n", szValue);
        return E_FAIL;
    }

    *pType = win_to_uda(dwType);
    return S_OK;
}

HRESULT
sysQueryValue(CTVOID *This, CTHANDLE hKey, PCTCHAR szValName, ISYSASSET_TYPE *pType, 
        PCTVOID pBuffer, PCTDWORD pdwBufSize)
{
    CTDWORD dwResult, dwType;

    if (!g_pIasDevReg) 
        return E_FAIL;

    //PDEBUG("sysQueryValue called: szValName=%s\n", szValName);
    if (hKey == HSYSCTKEY_ROOT || hKey == CTHANDLE_NULL) {
        hKey = (CTHANDLE) COMCALL(IasDevReg, g_pIasDevReg, GetRootHKEY)(g_pIasDevReg);
    }

    dwResult = COMCALL(IasDevReg, g_pIasDevReg, RegQueryValueEx)
            (g_pIasDevReg, (__HKEY) hKey, szValName, NULL, &dwType, pBuffer, pdwBufSize);

    if (dwResult != ERROR_SUCCESS) {
        PDEBUG("IasDevReg::RegQueryValueEx failed: szValName=%s\n", szValName);
        return E_FAIL;
    }

    *pType = win_to_uda(dwType);       
  
    return S_OK;
}

HRESULT 
sysCloseKey(CTVOID *This, CTHANDLE hKey)
{
    CTDWORD dwResult;

    if (!g_pIasDevReg) 
        return E_FAIL;

    //PDEBUG("sysCloseKey called, hKey=0x%08x\n", (CTDWORD) hKey);
    if (hKey == HSYSCTKEY_ROOT || hKey == CTHANDLE_NULL) 
        return S_OK; 

    dwResult = COMCALL(IasDevReg, g_pIasDevReg, RegCloseKey)(g_pIasDevReg, (__HKEY) hKey);

    if (dwResult != ERROR_SUCCESS) {
        PDEBUG("IasDevReg::RegCloseKey failed\n");
        return E_FAIL;
    }

    return S_OK;
}

HRESULT 
sysCreateKey(CTVOID *This, CTHANDLE hTopKey, const PCTCHAR szRelative, PCTHANDLE phKey)
{
    CTDWORD dwResult;

    if (!g_pIasDevReg) 
        return E_FAIL;

    //PDEBUG("sysCreateKey called: szRelative=%s\n", szRelative);
    if (hTopKey == HSYSCTKEY_ROOT || hTopKey == CTHANDLE_NULL) {
        hTopKey = (CTHANDLE) COMCALL(IasDevReg, g_pIasDevReg, GetRootHKEY)(g_pIasDevReg);    
    }
    
    dwResult = COMCALL(IasDevReg, g_pIasDevReg, RegCreateKey)
	(g_pIasDevReg, (__HKEY) hTopKey, szRelative, (__PHKEY) phKey);

    if (dwResult != ERROR_SUCCESS) {
        PDEBUG("IasDevReg::RegCreateKey failed\n");
        return E_FAIL;
    }

    return S_OK;
}

HRESULT 
sysSetValue(CTVOID *This, CTHANDLE hTopKey, const PCTCHAR szValName, ISYSASSET_TYPE type, 
        PCTVOID pBuffer, CTDWORD dwBufSize)
{
    CTDWORD dwResult;

    if (!g_pIasDevReg) 
        return E_FAIL;

    //PDEBUG("sysSetValue called: szValName=%s\n", szValName);
    if (hTopKey == HSYSCTKEY_ROOT || hTopKey == CTHANDLE_NULL) 
        hTopKey = (CTHANDLE) COMCALL(IasDevReg, g_pIasDevReg, GetRootHKEY)(g_pIasDevReg);

    if (type == ISYSASSET_SZ) {
        dwResult = COMCALL(IasDevReg, g_pIasDevReg, RegSetValue)(g_pIasDevReg, (__HKEY) hTopKey, 
                szValName, uda_to_win(type), pBuffer, dwBufSize);
    } else {
        dwResult = COMCALL(IasDevReg, g_pIasDevReg, RegSetValueEx)(g_pIasDevReg, (__HKEY) hTopKey, 
                szValName, 0, uda_to_win(type), pBuffer, dwBufSize);
    }

    if (dwResult != ERROR_SUCCESS) {
        PDEBUG("IasDevReg::RegSetValueEx failed\n");
        return E_FAIL;
    }

    return S_OK;
}


//////////////////////////////
// Module management functions
//////////////////////////////
  
HRESULT 
sysLoadModule(CTVOID *This, CTPATH *szPath, PCTHANDLE phModule)
{
    char *pch;
    PLUGINNODE *pCurrNode;
    char szModName[MAX_FILE_LEN];

    PDEBUG("sysLoadModule called: szPath=%s\n", szPath);
    if (!phModule) 
        CTRETURN(E_INVALIDARG);
    *phModule = CTHANDLE_NULL;

    // Extract module name
    pch = strrchr(szPath, '/');
    if (pch) 
        pch++;
    else 
        pch = szPath;
    strcpy(szModName, pch);
    pch = strrchr(szModName, '.');
    if (pch) 
        *pch = '\0';

    // Search for module
    pCurrNode = g_PluginList;
    while (pCurrNode) {
        if (strcmp(pCurrNode->szName, szModName) == 0) {
            pCurrNode->pfnPluginAddRef();
            *phModule = (CTHANDLE) pCurrNode;
            return S_OK;
        } 
        pCurrNode = pCurrNode->pNext;
    }

    // Not found, try to load it
    if (request_module(szPath)) {
        PDEBUG("sysLoadModule: request_module failed\n");
        return E_FAIL; 
    }

    // Add module handle to list
    pCurrNode = g_PluginList;
    while (pCurrNode) {
        if (strcmp(pCurrNode->szName, szModName) == 0) {
            pCurrNode->pfnPluginAddRef();
            strcpy(pCurrNode->szPath, szPath);
            break;
        }
        pCurrNode = pCurrNode->pNext;
    }

    CTASSERT(pCurrNode);
    *phModule = (CTHANDLE) pCurrNode;
  
    return S_OK;
}

HRESULT 
sysUnloadModule(CTVOID *This, CTHANDLE hModule)
{
    mm_segment_t fs = get_fs();    
    HRESULT hr = E_FAIL;
    uid_t uid = current->uid, euid = current->euid, fsuid = current->fsuid;
    CTBOOL bInhClear = cap_isclear(current->cap_inheritable);
    CTBOOL bEffClear = cap_isclear(current->cap_effective);
    PLUGINNODE *pNode = (PLUGINNODE *) hModule;

    PDEBUG("sysUnloadModule called\n");
    if (!pNode) 
        CTRETURN(E_INVALIDARG);

    if (pNode->pfnPluginRelease() > 0) 
        return S_OK;

    // Ref count is zero, unload plugin
    current->uid = current->euid = current->fsuid = 0;
    cap_set_full(current->cap_inheritable);
    cap_set_full(current->cap_effective);

    set_fs(get_ds());
    /*
    if (delete_module(pNode->szName) == 0) 
        hr = S_OK;
    */
    set_fs(fs);

    current->uid = uid;
    current->euid = euid;
    current->fsuid = fsuid;
    if (bInhClear) 
        cap_clear(current->cap_inheritable);
    if (bEffClear) 
        cap_clear(current->cap_effective);

    //PDEBUG("sysUnloadModules returning\n");
    return hr;
}

HRESULT 
sysGetProcAddress(CTVOID *This, CTHANDLE hModule, PCTCHAR szFnName, PPCTVOID ppfnAddr)
{
    PLUGINNODE *pNode = (PLUGINNODE *) hModule;

    if (!ppfnAddr || !pNode) 
        CTRETURN(E_INVALIDARG)
    *ppfnAddr = NULL;

    PDEBUG("sysGetProcAddress called, szModName = %s, szFnName = %s\n", pNode->szName, szFnName);

    if (strcmp(szFnName, "DllCanUnloadNow") == 0) 
        *ppfnAddr = pNode->pfnDllCanUnloadNow;
    else if (strcmp(szFnName, "DllAdviseIfCanUnload") == 0)
        *ppfnAddr = pNode->pfnDllAdviseIfCanUnload;
    else if (strcmp(szFnName, "DllGetClassObject") == 0)
        *ppfnAddr = pNode->pfnDllGetClassObject;
    else if (strcmp(szFnName, "DllEnumCLSIDs") == 0)
        *ppfnAddr = pNode->pfnDllEnumCLSIDs;
    else if (strcmp(szFnName, "DllEnumDefaultStrings") == 0)
        *ppfnAddr = pNode->pfnDllEnumDefaultStrings;
    else if (strcmp(szFnName, "DllEnumDefaultUnits") == 0)
        *ppfnAddr = pNode->pfnDllEnumDefaultUnits;
    else { // Function not found
        PDEBUG("sysGetProcAddress: Function not found\n");
        return E_FAIL;
    }

    return S_OK;
}


/////////////////////////////
// Global interrupt functions
/////////////////////////////

spinlock_t inter_lock = SPIN_LOCK_UNLOCKED;

CTHANDLE 
sysDisableAllInterrupts(CTVOID *This)
{
    unsigned long ulFlags;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    save_flags(ulFlags);
    cli();
#else
    //local_save_flags(ulFlags);
    spin_lock_irqsave(&inter_lock, ulFlags);
#endif

    return (CTHANDLE)ulFlags;
}

CTVOID 
sysEnableAllInterrupts(CTVOID *This, CTHANDLE hIntr)
{
    unsigned long ulFlags = (unsigned long) hIntr;
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    restore_flags(ulFlags);
#else 
    //local_irq_restore(ulFlags);
    spin_unlock_irqrestore(&inter_lock, ulFlags);
#endif
    inter_lock = SPIN_LOCK_UNLOCKED;
}

CTLONG 
sysInterlockedIncrement(CTVOID *This, CTLONG *plRef)
{
    return InterlockedIncrement(plRef);
}

CTLONG 
sysInterlockedDecrement(CTVOID *This, CTLONG *plRef)
{
    return InterlockedDecrement(plRef);
}

HRESULT 
extract_GUID_from_string(CTCHAR * pGuidString, CLSID *pGuid)
{
    CTDWORD i, dwData;
    CTCHAR *str = pGuidString;

    CTDPF("extract_GUID_from_string() is called");

    memset(pGuid, 0, sizeof(CLSID));

    if (*str++ != '{') 
        return E_FAIL;
    
    for (i = 0; i < 8; i++) {
        pGuid->Data1 <<= 4;
        dwData = CTCHARTOHEX(*str);
        if (dwData > 16) 
            return E_FAIL;
        pGuid->Data1 += (CTULONG)dwData;
        str++;
    }

    if (*str++ != '-')
        return E_FAIL;
    
    for (i = 0; i < 4; i++) {
        pGuid->Data2 <<= 4;
        dwData = CTCHARTOHEX(*str);
        if (dwData > 16)
            return E_FAIL;
        pGuid->Data2 += (CTUSHORT)dwData;
        str++;
    }

    if (*str++ != '-')
        return E_FAIL;
    
    for (i = 0; i<4; i++) {
        pGuid->Data3 <<= 4;
        dwData = CTCHARTOHEX(*str);
        if (dwData > 16)
            return E_FAIL;
        pGuid->Data3 += (CTUSHORT)dwData;
        str++;
    }

    if (*str++ != '-')
        return E_FAIL;

    for (i = 0; i < 2; i++) {
        dwData = CTCHARTOHEX(*str);
        if (dwData > 16)
            return E_FAIL;
        pGuid->Data4[i] = (CTUCHAR)dwData;
        str++;
        dwData = CTCHARTOHEX(*str);
        if (dwData > 16)
            return E_FAIL;
        pGuid->Data4[i] <<= 4;
        pGuid->Data4[i] += (CTUCHAR)dwData;
        str++;
    }

    if (*str++ != '-')
        return E_FAIL;
    for (i = 2; i < 8; i++) {
        dwData = CTCHARTOHEX(*str);
        if (dwData > 16)
            return E_FAIL;
        pGuid->Data4[i] = (CTUCHAR)dwData;
        str++;
        dwData = CTCHARTOHEX(*str);
        if (dwData > 16)
            return E_FAIL;
        pGuid->Data4[i] <<= 4;
        pGuid->Data4[i] += (CTUCHAR)dwData;
        str++;
    }

    if (*str != '}')
        return E_FAIL;

    return S_OK;
}

HRESULT
sysGetHardwareInfo2(CTVOID *This, CTHANDLE hDevice, stHardwareInfo2 *pInfo2)
{
    HRESULT hr = E_FAIL;
    stHardwareInfo Info;
    CTHANDLE hTopKey = HSYSCTKEY_ROOT;
    PCTCHAR szRelative = heap_alloc(ALLOCARG(MAX_REGISTRYPATH_LEN, TRUE));
    CTHANDLE hKey;
    PCTHANDLE phKey = &hKey;
    ISYSASSET_TYPE Type = ISYSASSET_SZ;

    CTDPF("sysGetHardwareInfo2() is called");

    PCTCHAR strValue=heap_alloc(ALLOCARG(MAX_REGVALUENAME_LEN, TRUE));
    CTDWORD dwSize = MAX_REGVALUENAME_LEN * sizeof(CTCHAR);

    memset(szRelative, 0, MAX_REGISTRYPATH_LEN);
    memset(strValue, 0, dwSize);

    if(!hDevice) {
        heap_free(szRelative);
        heap_free(strValue);
        return E_HANDLE;
    }
    
    if(!pInfo2) {
        heap_free(szRelative);
        heap_free(strValue);
        return E_POINTER;
    }
    
    if(pInfo2->cbSize != sizeof(stHardwareInfo2)) {
        heap_free(szRelative);
        heap_free(strValue);
        return E_INVALIDARG;
    }

    hr = sysGetHardwareInfo(This, hDevice, &Info);
    if(FAILED(hr)) {
        heap_free(szRelative);
        heap_free(strValue);
        return hr;
    }

    pInfo2->busType    = Info.busType;
    pInfo2->deviceType = 0;
    pInfo2->busNum     = Info.busNum;
    pInfo2->funcNum    = Info.funcNum;
    pInfo2->instance   = Info.instance;
    pInfo2->vendorID   = Info.vendorID;
    pInfo2->deviceID   = Info.deviceID;
    pInfo2->subsys     = Info.subsys;
    pInfo2->rev        = Info.rev;

    if (pInfo2->busType == HWINFO_BUS_TYPE_PCI) {
        sprintf(szRelative,"HardwareID\\PCI&VEN_%04lX&DEV_%04lX", (unsigned long)pInfo2->vendorID,
                        (unsigned long)pInfo2->deviceID);
    } else if (pInfo2->busType == HWINFO_BUS_TYPE_USB) {
        sprintf(szRelative,"HardwareID\\USB&VID_%04lX&PID_%04lX", (unsigned long)pInfo2->vendorID,
                        (unsigned long)pInfo2->deviceID);
    }
 
    // Open "HardwareID\\xxx_%04lX&xxx_%04lX" entry
    hTopKey = HSYSCTKEY_ROOT;
    hr = E_FAIL;
    hr = sysOpenKey(This, hTopKey, szRelative, phKey);
    if(FAILED(hr)) {
        heap_free(szRelative);
        heap_free(strValue);
        return hr;
    }

    // Query "HWCLSID" string value.
    hr = E_FAIL;
    Type = ISYSASSET_SZ;
    hr = sysQueryValue(This, hKey, "HWCLSID", &Type, strValue, &dwSize);
    if(SUCCEEDED(hr)) {
        // Convert GUID_STRING to GUID_VALUE.
        // GUID_STRING should format like this:"{xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}".
        extract_GUID_from_string(strValue, &pInfo2->HWClassID);
    }

    // Query "DeviceType" string value.
    hr = E_FAIL;
    Type = ISYSASSET_SZ;
    hr = sysQueryValue(This, hKey, "DeviceType", &Type, strValue, &dwSize);
    if(SUCCEEDED(hr)) {
        if (0 == strcmp(strValue, "AUDIO")) {
            pInfo2->deviceType = HWINFO_DEVICE_TYPE_AUDIO;
        } else if (0 == strcmp(strValue, "JOYSTICK")) {
            pInfo2->deviceType = HWINFO_DEVICE_TYPE_JOYSTICK;
        }
    }
    
    // temp added by tristan 
    pInfo2->deviceType = HWINFO_DEVICE_TYPE_AUDIO;
    hr = S_OK;
    // end

    //Close the opened entry
    sysCloseKey(This, hKey);
    heap_free(szRelative);
    heap_free(strValue);

    return hr;
}

HRESULT
sysLoadDriver(CTVOID *This, REFCLSID DriverClsid, FEATURESCALE featureScale)
{
    HRESULT hr = E_FAIL;

    CTHANDLE hTopKey = HSYSCTKEY_ROOT;
    PCTCHAR szRelative = heap_alloc(ALLOCARG(MAX_REGISTRYPATH_LEN, TRUE));
    CTHANDLE hKey;
    PCTHANDLE phKey = &hKey;
    ISYSASSET_TYPE Type;

    CTDWORD dwServiceType;
    CTDWORD dwServiceTypeSize = sizeof(CTDWORD);
    PCTCHAR ServiceName=heap_alloc(ALLOCARG(MAX_REGVALUENAME_LEN, TRUE));
    CTDWORD dwServiceNameSize = MAX_REGVALUENAME_LEN * sizeof(CTCHAR);

    CTDPF("sysLoadDriver() is called");
    
    //
    memset(szRelative, 0, MAX_REGISTRYPATH_LEN);
    memset(ServiceName, 0, dwServiceNameSize);

    //Creat registry path "BinaryClassID\\{%08lX-%04lX-%04lX-%02X%02X-%02X%02X%02X%02X%02X%02X}"
    sprintf(szRelative, "BinaryClassID\\{%08lX-%04lX-%04lX-%02X%02X-%02X%02X%02X%02X%02X%02X}",
            (unsigned long)DriverClsid->Data1, (unsigned long)DriverClsid->Data2, (unsigned long)DriverClsid->Data3,
            DriverClsid->Data4[0], DriverClsid->Data4[1], DriverClsid->Data4[2], DriverClsid->Data4[3], 
            DriverClsid->Data4[4], DriverClsid->Data4[5], DriverClsid->Data4[6], DriverClsid->Data4[7]);

    //Open "BinaryClassID\\{%08lX-%04lX-%04lX-%02X%02X-%02X%02X%02X%02X%02X%02X}" entry
    hTopKey = HSYSCTKEY_ROOT;
    hr = E_FAIL;
    hr = sysOpenKey(This, hTopKey, szRelative, phKey);
    if(FAILED(hr)) {
        heap_free(szRelative);
        heap_free(ServiceName);
        return hr;
    }

    //Query "ServiceType" DWORD value
    hr = E_FAIL;
    Type = ISYSASSET_DWORD;
    hr = sysQueryValue(This, hKey, "ServiceType", &Type, &dwServiceType, &dwServiceTypeSize);
    if(FAILED(hr)) {
        //Close the opened entry
        sysCloseKey(This, hKey);
        heap_free(szRelative);
        heap_free(ServiceName);
        return hr;
    } else if((1<<featureScale) & dwServiceType) {
        // Query "ServiceName" string value.
        hr = E_FAIL;
        Type = ISYSASSET_SZ;
        hr = sysQueryValue(This, hKey, "ServiceName", &Type, ServiceName, &dwServiceNameSize);
        if(FAILED(hr))    {
            //Close the opened entry
            sysCloseKey(This, hKey);
            heap_free(szRelative);
            heap_free(ServiceName);
            return hr;
        }

        //TODO:add code for searching for module to see if the dirver is aready loaded 

        // Driver not loaded yet... so load it now.
        if (request_module(ServiceName)) {
            PDEBUG("sysLoadDriver: request_module failed\n");
            //Close the opened entry
            sysCloseKey(This, hKey);
            heap_free(szRelative);
            heap_free(ServiceName);
            return E_FAIL;
        }

        //TODO:add code for searching for module to ensure loading it successfully
    }

    //Close the opened entry
    sysCloseKey(This, hKey);
    heap_free(szRelative);
    heap_free(ServiceName);

    return S_OK;
}

HRESULT
sysUnloadDriver(CTVOID *This, REFCLSID DriverClsid)
{
    HRESULT hr = E_FAIL;

    CTHANDLE hTopKey = HSYSCTKEY_ROOT;
    PCTCHAR szRelative=heap_alloc(ALLOCARG(MAX_REGISTRYPATH_LEN, TRUE));
    CTHANDLE hKey;
    PCTHANDLE phKey = &hKey;
    ISYSASSET_TYPE Type;

    PCTCHAR ServiceName=heap_alloc(ALLOCARG(MAX_REGVALUENAME_LEN, TRUE));
    CTDWORD dwServiceNameSize = MAX_REGVALUENAME_LEN * sizeof(CTCHAR);

    CTDPF("sysUnloadDriver() is called");

    mm_segment_t fs = get_fs();    
    uid_t uid = current->uid, euid = current->euid, fsuid = current->fsuid;
   	CTBOOL bInhClear = cap_isclear(current->cap_inheritable), bEffClear = cap_isclear(current->cap_effective);

    //
    memset(szRelative, 0, MAX_REGISTRYPATH_LEN);
    memset(ServiceName, 0, dwServiceNameSize);

    //Creat registry path "BinaryClassID\\{%08lX-%04lX-%04lX-%02X%02X-%02X%02X%02X%02X%02X%02X}"
    sprintf( szRelative, "BinaryClassID\\{%08lX-%04lX-%04lX-%02X%02X-%02X%02X%02X%02X%02X%02X}", 
            (unsigned long)DriverClsid->Data1, (unsigned long)DriverClsid->Data2, (unsigned long)DriverClsid->Data3,
            DriverClsid->Data4[0], DriverClsid->Data4[1], DriverClsid->Data4[2], DriverClsid->Data4[3],
            DriverClsid->Data4[4], DriverClsid->Data4[5], DriverClsid->Data4[6], DriverClsid->Data4[7]);

    //Open "BinaryClassID\\{%08lX-%04lX-%04lX-%02X%02X-%02X%02X%02X%02X%02X%02X}" entry
    hTopKey = HSYSCTKEY_ROOT;
    hr = E_FAIL;
    hr = sysOpenKey(This, hTopKey, szRelative, phKey);
    if(FAILED(hr)) {
        heap_free(szRelative);
        heap_free(ServiceName);
        return hr;
    }
    
    // Query "ServiceName" string value.
    hr = E_FAIL;
    Type = ISYSASSET_SZ;
    hr = sysQueryValue(This, hKey, "ServiceName", &Type, ServiceName, &dwServiceNameSize);
    if(FAILED(hr)) {
        //Close the opened entry
        sysCloseKey(This, hKey);
        heap_free(szRelative);
        heap_free(ServiceName);
        return hr;
    }

    PDEBUG("sysUnloadDriver called\n");

    //TODO:add code for searching for module to see if the dirver exists
    // Unload it
    current->uid = current->euid = current->fsuid = 0;
    cap_set_full(current->cap_inheritable);
    cap_set_full(current->cap_effective);

    set_fs(get_ds());
    /*
    if (delete_module(ServiceName) == 0) 
        hr = S_OK;
    */
    set_fs(fs);

    current->uid = uid;
    current->euid = euid;
    current->fsuid = fsuid;
    if (bInhClear)
        cap_clear(current->cap_inheritable);
    if (bEffClear) 
        cap_clear(current->cap_effective);

    //TODO:add code for searching for module to ensure unloading it successfully
    //PDEBUG("sysUnloadDriver returning\n");

    //Close the opened entry
    sysCloseKey(This, hKey);
    heap_free(szRelative);
    heap_free(ServiceName);
    
    return hr;
}

    
// IMP_IasSystem3

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
static void nonCriticalTimeout_callback(void *pData)
{
    stTimerCbkHandle *phTimerCbk = (stTimerCbkHandle *)pData;
#else
static void nonCriticalTimeout_callback(struct work_struct *work)
{
    stTimerCbkHandle *phTimerCbk = container_of(work, stTimerCbkHandle, SchedulerQueue);
#endif
    //PDEBUG("nonCriticalTimeout_callback called\n");
    if (!phTimerCbk || !phTimerCbk->fpCallback) {
        PDEBUG("Error: Invalid timer callback handle!\n");
        CTASSERT(0);
        return;
    }

    phTimerCbk->fpCallback(phTimerCbk->hUser);
}

HRESULT
sysCreateNonCriticalTimeout(CTVOID *This, OSTIMEOUTHANDLE *phTmo)
{
    HRESULT hr = E_FAIL; 
    stTimerCbkHandle *phTimerCbk;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    struct tq_struct *pWork;
#else
    struct work_struct *pWork;
#endif

    CTDPF("CreateNonCriticalTimeout called");
    hr = sysCreateTimeout(This, phTmo);
    if (FAILED(hr))
        return hr;

    phTimerCbk = *phTmo;
    phTimerCbk->bNonCriticalTimeoutWork = TRUE;

    pWork = &phTimerCbk->SchedulerQueue;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    pWork->routine = nonCriticalTimeout_callback;
    pWork->data = (void *)phTimerCbk;
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
    INIT_WORK(pWork, (void *)nonCriticalTimeout_callback, phTimerCbk);
#else
    INIT_WORK(pWork, nonCriticalTimeout_callback);
#endif

    return hr;
}
                 
HRESULT
sysDestroyNonCriticalTimeout(CTVOID *This, OSTIMEOUTHANDLE hTmo)
{
    stTimerCbkHandle *phTimerCbk = (stTimerCbkHandle *)hTmo;
    
    CTDPF("DestroyNonCriticalTimeout called");
    if (!phTimerCbk)
        CTRETURN(E_INVALIDARG);
	 
    return sysDestroyTimeout(This, phTimerCbk);
}
    
HRESULT
sysScheduleNonCriticalTimeout(CTVOID *This, OSTIMEOUTHANDLE hTimeOut,  PFCALLBACK pfCallback, CTHANDLE hUser, CTDWORD dwMillisecs)
{
    //CTDPF("ScheduleNonCriticalTimeout called");

    return sysScheduleTimeout(This, hTimeOut, pfCallback, hUser, dwMillisecs);
}
    
HRESULT
sysQuerySystemInformation(CTVOID *This, CTDWORD dwType, CTBYTE *pData, CTULONG *pcbSize)
{
    CTDPF("QuerySystemInformation called");
    if (!pcbSize ||!pData) {
        CTASSERT(0)
        return E_INVALIDARG;
    }

    switch (dwType) {
        case CT_SYSINFO_PHYSPAGESIZE:
        case CT_SYSINFO_NUMPHYSPAGES:
        case CT_SYSINFO_ACTIVEPROCESSORS:
                if (*pcbSize < sizeof(CTPTRSIZE)) {
                    *pcbSize = sizeof(CTPTRSIZE);
                    return E_INVALIDARG;
                }
                break;
        
        case CT_SYSINFO_NUMPROCESSORS:
                if (*pcbSize < sizeof(CTULONG)) {
                    *pcbSize = sizeof(CTULONG);
                    return E_INVALIDARG;
                }
                break;
        
        default:
            return E_NOTIMPL; // Unrecognized system information type.
    }

    //TODO: Hard-coded here, need to collect from actual system
    switch (dwType) {
        case CT_SYSINFO_PHYSPAGESIZE:
            *((CTPTRSIZE *)pData) =0x1000;
            *pcbSize = sizeof(CTPTRSIZE);
            break;
        
        case CT_SYSINFO_NUMPHYSPAGES:
            *((CTPTRSIZE *)pData) = 0x8000;
            *pcbSize = sizeof(CTPTRSIZE);
            break;

        case CT_SYSINFO_NUMPROCESSORS:
            *((CTULONG *)pData) = 1;
            *pcbSize = sizeof(CTULONG);
            break;

        case CT_SYSINFO_ACTIVEPROCESSORS:
            *((CTPTRSIZE *)pData) = 1;
            *pcbSize = sizeof(CTPTRSIZE);
            break;

        default:
            break;
	}

	return S_OK;
}
                 

// IMP_IasSystem4

OSPHYSMEMHDL 
sysMapDeviceMemory(CTVOID *This, interface ICtDevice *pDevice, CTPTRSIZE baseVirtualAddr, CTPTRSIZE baseBusAddr, 
        CTDWORD offset, CTDWORD numBytes, CTBOOL bCacheable)
{
    CTHANDLE hHandle = NULL;

    CTDPF("MapDeviceMemory called");

    return hHandle;
}
           
CTULONG
sysSetTimerResolution(CTVOID *This, CTULONG ulTimeIn100nsUnits, CTBOOL bSetResolution)
{
    CTDPF("SetTimerResolution called");
    
    return (CTULONG)0;
}
    
CTLONGLONG
sysQueryPerformanceCounter(CTVOID *This, PCTLONGLONG pllPerformanceFrequency)
{
    CTDPF("QueryPerformanceCounter called");
    
    return (CTLONGLONG)0;
}
                    
HRESULT
sysDelayExecutionThread(CTVOID *This, CTLONGLONG llDelayIntervalIn100nsUnits, CTBOOL bRelAbs)
{
    HRESULT hr = E_FAIL;

    CTDPF("DelayExecutionThread called");
    CTULARGE val = (CTULARGE) llDelayIntervalIn100nsUnits * 100;
    ndelay(val); // nano seconds

    hr = S_OK;
    
    return hr;
}
    
OSPHYSMEMHDL
sysMapSystemMemory(CTVOID *This, CTVOID *pVirtualAddress, CTDWORD dwSize, CTBOOL bKernelAddress, CTBOOL bAccessMode, CTDWORD fAccessRights)
{
    CTHANDLE hHandle = NULL;

    CTDPF("MapSystemMemory called");

    return hHandle;
}


// IMP_IasSystem5
CTHANDLE
sysGetCurrentThreadId(CTVOID *This)
{
    CTHANDLE hHandle = NULL;

    //CTDPF("CreateThreadSpinLock called");

    return hHandle;
}
            
HRESULT
sysCreateThreadSpinLock(CTVOID *This, OSSPINLOCKHDL *phSpinlock, interface ICtDevice *device)
{
    HRESULT hr = E_FAIL;

    //CTDPF("CreateThreadSpinLock called");
    hr = S_OK;
    
    return hr;
}

HRESULT
sysDestroyThreadSpinLock(CTVOID *This, OSSPINLOCKHDL hSpinlock)
{
    HRESULT hr = E_FAIL;

    //CTDPF("DestroyThreadSpinLock called");
    hr = S_OK;
    
    return hr;
}

HRESULT
sysAcquireThreadSpinLock(CTVOID *This, OSSPINLOCKHDL hSpinlock)
{
    HRESULT hr = E_FAIL;

    //CTDPF("AcquireThreadSpinLock called");
    hr = S_OK;
    
    return hr;
}

HRESULT
sysReleaseThreadSpinLock(CTVOID *This, OSSPINLOCKHDL hSpinlock)
{
    HRESULT hr = E_FAIL;

    //CTDPF("ReleaseThreadSpinLock called");
    hr = S_OK;
    
    return hr;
}

    
static const tSysVtbl g_SysVtbl = 
{
    // IMP_IasSystem
    .vtbl1 =
    {
        // IMP_DECL_ICTNONDEL
        sysNondelegatingQueryInterface,
        sysNondelegatingAddRef,
        sysNondelegatingRelease,
        sysInit,
        // IMP_DECL_ICTUNK
        sysQueryInterface,
        sysAddRef,
        sysRelease,
        // IMP_IasSystem
        sysPostBootInit,
        sysGetSystemTime,
        sysCreateTimeout,
        sysDestroyTimeout,
        sysScheduleTimeout,
        sysCreateInterruptTimeout,
        sysDestroyInterruptTimeout,
        sysScheduleInterruptTimeout,
        sysCreateDpc,
        sysDestroyDpc,
        sysScheduleDpc,
        sysRegisterInterrupt,
        sysUnregisterInterrupt,
        sysStartInterrupt,
        sysStopInterrupt,
        sysAckInterrupt,
        sysGetHardwareInfo,
        sysHeapAllocate,
        sysLockedHeapAllocate,
        sysHeapFree,
        sysLockedHeapFree,
        sysAllocPages,
        sysAllocContiguousPages,
        sysFreePages,
        sysLockPages,
        sysUnlockPages,
        sysGetPageVirtAddr,
        sysGetPageUserModeVirtAddr,
        sysGetPagePhysAddr,
        sysGetPageBusAddr,
        sysGetPageRangeSize,
        sysFlushPageCache,
        sysMapFlat16BitPointer,
        sysCreateSpinLock,
        sysDestroySpinLock,
        sysAcquireSpinLock,
        sysReleaseSpinLock,
        sysCreateMutex,
        sysAcquireMutex,
        sysReleaseMutex,
        sysWaitMutex,
        sysDeleteMutex,
        sysCreateSemaphore, 
        sysWaitSemaphore,
        sysSignalSemaphore,
        sysDestroySemaphore,
        sysOpenCreateFile,
        sysGetFileSize,
        sysWriteFile,
        sysReadFile,
        sysCloseFile,
        sysGetPathLength,
        sysIsFullPath,
        sysArePathsIdentical,
        sysGetSystemStylePathName,
        sysGetDriverDataDirectory,
        sysGetDriverPreferenceDirectory,
        sysGetPathDelimiter,
        sysIsFileSystemCaseSensitive,
        sysIsFileSystemWideChar,
        sysASCIIPathToWideChar,
        sysWideCharPathtoASCII,
        sysASCIIPathToCTPATH,
        sysWideCharPathToCTPATH,
        sys_acos,
        sys_asin,
        sys_atan,
        sys_atan2,
        sys_cos,
        sys_cosh,
        sys_exp,
        sys_fabs,
        sys_fmod,
        sys_log,
        sys_log10,
        sys_pow,
        sys_sin,
        sys_sinh,
        sys_tan,
        sys_tanh,
        sys_sqrt,
        sys_ceil,
        sys_floor,
        sys_frexp,
        sys_ldexp,
        sys_modf,
        sysGetHugeNum,
        sys_finite,
        sys_isnan,
        sysGetFPUBufSize,
        sysSaveFPU,
        sysRestoreFPU,
        sysOpenKey,
        sysEnumKey,
        sysEnumValue,
        sysQueryValue,
        sysCloseKey,
        sysCreateKey,
        sysSetValue,
        sysLoadModule,
        sysUnloadModule,
        sysGetProcAddress,
        sysDisableAllInterrupts,
        sysEnableAllInterrupts,
        sysInterlockedIncrement,
        sysInterlockedDecrement
    },
    // IMP_IasSystem2
    .vtbl2 =
    {
        sysGetHardwareInfo2,
        sysLoadDriver,
        sysUnloadDriver
    },
    // IMP_IasSystem3
    .vtbl3 =
    {
        sysCreateNonCriticalTimeout,
        sysDestroyNonCriticalTimeout,
        sysScheduleNonCriticalTimeout,
        sysQuerySystemInformation
    },
    // IMP_IasSystem4
    .vtbl4 =
    {
        sysMapDeviceMemory,
        sysSetTimerResolution,
        sysQueryPerformanceCounter,
        sysDelayExecutionThread,
        sysMapSystemMemory
    },
    // IMP_IasSystem5
    .vtbl5 =
    {
        sysGetCurrentThreadId,
        sysCreateThreadSpinLock,
        sysDestroyThreadSpinLock,
        sysAcquireThreadSpinLock,
        sysReleaseThreadSpinLock
    }
};


// Constructor/Destructor
CTVOID 
sysConstruct(CTVOID *This, IasOSSrv *pIasOSSrv)
{
    PSYSOBJ pObj = This;
    
    if (!pObj) {
        CTASSERT(0);
        return;
    }

    pObj->lpVtbl = kmalloc(sizeof(tSysVtbl), GFP_KERNEL);
    if (!pObj->lpVtbl) {
        CTASSERT(0);
        return;
    }

    memcpy(pObj->lpVtbl, &g_SysVtbl, sizeof(tSysVtbl));
    pObj->m_cRef = 1;
    pObj->m_pIasOSSrv = pIasOSSrv;
    pObj->m_pIUnknownOuter = This;    
}

CTVOID 
sysDestruct(CTVOID *This)
{
    PSYSOBJ pObj = This;

    if (pObj->lpVtbl) {
        kfree(pObj->lpVtbl);
        pObj->lpVtbl = NULL;
    }
}

#ifdef __USE_LIBM_A__
//======================================================================
// Floating Point support
// This is needed when using floating point in linux modules linked
// with libm.a.
//======================================================================

FILE *stderr = NULL; // The floating point library needs this when error occurs
//int errno=0;
extern void __assert_fail (const char *__assertion, const char *__file,
               unsigned int __line, __const char *__function);

extern int __isnan(double f)
{
    return f == NAN;
}

extern int fputs (const char *s, FILE *stream)
{
    PDEBUG("fputs is called with string=%s !!!!!!!!!!!!",s);
    return 0;
}

extern size_t fwrite (const void *ptr, size_t size,size_t n, FILE *s)
{
    PDEBUG("fwrite is called!!!!!!!!!!!!!!!!!!in");
    return n;
}

extern int *__errno_location (void)
{
    return &errno;
}

extern void __assert_fail (const char *__assertion, const char *__file,
               unsigned int __line, __const char *__function)
{
     printk(KERN_ERR "%s %s LINE%d %s\n",__assertion,__file,__line,__function);
     BUG();
}
#endif

/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

