/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
Functions exported by ctossrv
Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#define __NO_VERSION__
#define EXPORT_SYMTAB
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "ctossrv.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#   include <linux/wrapper.h>
#else
#   include <linux/init.h>
#   include <linux/kernel.h>
#   include <linux/pagemap.h>
#   include <linux/preempt.h>
#   include <linux/rmap.h>
#   include <linux/swap.h>
#   include <linux/swapops.h>
#   include <asm/tlbflush.h>
#   ifndef EXPORT_SYMBOL_NOVERS
#       define EXPORT_SYMBOL_NOVERS EXPORT_SYMBOL
#   endif
#endif

////////////////////
// Plugin management 
////////////////////

PLUGINNODE *g_PluginList = NULL;

// Registers a loading plugin module
HRESULT 
register_plugin(PLUGINNODE * pPluginNode)
{
    PLUGINNODE *pNewNode;

    if (!pPluginNode)
        return E_INVALIDARG;

    pNewNode = kmalloc(sizeof(PLUGINNODE), GFP_KERNEL);
    if (!pNewNode)
        return E_OUTOFMEMORY;

    memcpy(pNewNode, pPluginNode, sizeof(PLUGINNODE));
    pNewNode->pNext = g_PluginList;
    g_PluginList = pNewNode;

    PDEBUG("register_plugin registered %s\n", pPluginNode->szName);

	return S_OK;
}
EXPORT_SYMBOL_NOVERS(register_plugin);

// Unregisters an unloading plugin module
HRESULT 
unregister_plugin(char *szName)
{
    PLUGINNODE *pCurrNode, *pPrevNode;

    pCurrNode = g_PluginList;
    pPrevNode = NULL;

    while (pCurrNode) {
        if (strcmp(szName, pCurrNode->szName) == 0) {
            if (pPrevNode)
                pPrevNode->pNext = pCurrNode->pNext;
            else
                g_PluginList = pCurrNode->pNext;
            kfree(pCurrNode);
            return S_OK;
        }
        pPrevNode = pCurrNode;
        pCurrNode = pCurrNode->pNext;
    }

    return E_FAIL;
}
EXPORT_SYMBOL_NOVERS(unregister_plugin);

// This has to sync with LinuxSys.c
extern HRESULT sysUnloadModule(CTVOID *This, CTHANDLE hModule);

// This is a "garbage collection" function that unloads all remnant plugin modules
// It should only be called when emupia is unloaded, or Bad Things will happen
void 
unload_all_plugins(void)
{
    while (g_PluginList) {
        // Assumes this ptr is not used
        // Assumes typeof(hModule) == PLUGINNODE*
        sysUnloadModule(NULL, g_PluginList);
    }
}
EXPORT_SYMBOL_NOVERS(unload_all_plugins);

////////////////////////
// Interface acquisition
////////////////////////

extern IasOSSrv *g__pIasOSSrv;

CTHANDLE 
get_ossrv(void)
{
    return ((CTHANDLE) g__pIasOSSrv);
}
EXPORT_SYMBOL_NOVERS(get_ossrv);

CTHANDLE 
get_devreg(void)
{
    return ((CTHANDLE) g_pIasDevReg);
}
EXPORT_SYMBOL_NOVERS(get_devreg);

#define MAX_CONTIG (PAGE_SIZE * 8)

////////////////////
// Memory management
////////////////////

void *
heap_alloc(ALLOC_ARGLIST)
{
    void *pv;
    int priority;

    if (size == 0)
        size = 1;
   
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    priority = in_interrupt() ? GFP_ATOMIC : GFP_KERNEL;
#else
    priority = (in_interrupt() || in_atomic() || irqs_disabled()) ? GFP_ATOMIC : GFP_KERNEL;
#endif

    if ( (priority != GFP_ATOMIC) && (size >= MAX_CONTIG) )
    {
        pv = vmalloc(size);
    }
    else
    {
        if (priority == GFP_ATOMIC)
        {
            //CTDPF("heap_alloc: ATOMIC. size=%d, preempt_count=%x, interrupt=%d, atomic=%d, irqs_disabled=%d",
            //            (int)size, preempt_count(), (int)in_interrupt(), in_atomic(), irqs_disabled());
           
            /* 
            static int dumpstack=0;
            if (0 && dumpstack < 15)
                dump_stack();
            dumpstack++;
            */
        }

        pv = kmalloc(size, priority);
        if (!pv)
        {
            #if 0
            CTDPF("heap_alloc: kmalloc failed. Try vmalloc. size=%d, preempt_count=%x, interrupt=%d, atomic=%d, irqs_disabled=%d\n\n\n", 
                        (int)size, preempt_count(), (int)in_interrupt(), in_atomic(), irqs_disabled());
            int preempt_cnt;
            preempt_cnt = preempt_count();
            preempt_count() = 0;
            pv = vmalloc(size);
            preempt_count() = preempt_cnt;
            #else
            pv = vmalloc(size);
            #endif
        }
    }
    
    if (pv && bZero)
        memset(pv, 0, size);

    CTASSERT(pv);
#if DEBUG
    if (!pv)
    {
        CTDPF("heap_alloc failed!!!!! size=%d\n", (int)size);
    }
#endif

    return pv;
}
EXPORT_SYMBOL_NOVERS(heap_alloc);

void 
heap_free(void *pv)
{
    if (pv == NULL) {
        CTASSERT(0);
        return;
    }

    if (pv >= (void *) VMALLOC_START)
        vfree(pv);
    else if (pv)
        kfree(pv);
}
EXPORT_SYMBOL_NOVERS(heap_free);

///////////////////////////////////////
// Memory mapping utilities
///////////////////////////////////////

// Terminology:
// uvirt - User-mode virtual address        (eg. returned by malloc()  )
// kvirt - Kernel-mode high virtual address (eg. returned by vmalloc() )
// kva   - Kernel-mode low virtual address  (eg. returned by kmalloc() )

#if 0 /* no use actually */
// convert *virt(uvirt/kvirt) address to kva address
static inline unsigned long 
_virt_to_kva(pgd_t *p_pgd, unsigned long _virt)
{
    pmd_t *p_pmd;
    pte_t *p_pte;
    unsigned long kva = 0;

    //CTDPF("_virt_to_kva: _virt = 0x%lx ", _virt);
    if (!pgd_none(*p_pgd)) {
        p_pmd = pmd_offset(p_pgd, _virt);
        if (!pmd_none(*p_pmd)) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
            p_pte = pte_offset(p_pmd, _virt);
#else
            p_pte = pte_offset_map(p_pmd, _virt);
#endif
            if(pte_present(*p_pte))
                kva = (unsigned long)page_address(pte_page(*p_pte)) | (_virt&(PAGE_SIZE-1));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
            pte_unmap(p_pte);
#endif
        }
    }

    return kva;
}

unsigned long 
uvirt_to_phys(unsigned long uvirt)
{
    unsigned long kva;

    kva = _virt_to_kva(pgd_offset(current->mm, uvirt), uvirt);

    return virt_to_phys((void *)kva);
}

unsigned long 
uvirt_to_bus(unsigned long uvirt)
{
    unsigned long kva;

    kva = _virt_to_kva(pgd_offset(current->mm, uvirt), uvirt);

    return virt_to_bus((void *)kva);
}

unsigned long 
kvirt_to_phys(unsigned long kvirt)
{
    unsigned long kva;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,10)
    kva = _virt_to_kva(pgd_offset_k(kvirt), kvirt);
#else
    kva = _virt_to_kva(pgd_offset(current->mm, kvirt), kvirt);
#endif

    return virt_to_phys((void *)kva);
}

unsigned long 
kvirt_to_bus(unsigned long kvirt)
{
    unsigned long kva;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,10)
    kva = _virt_to_kva(pgd_offset_k(kvirt), kvirt);
#else
    kva = _virt_to_kva(pgd_offset(current->mm, kvirt), kvirt);
#endif

    return virt_to_bus((void *)kva);
}

#endif

#ifndef pte_offset_map
    #define pte_offset_map  pte_offset
#endif

static inline struct page * 
kvirt_to_page(unsigned long kvirt)
{
    pgd_t *p_pgd;
#if defined(__x86_64__)
    #ifdef pud_offset
    pud_t *p_pud;
    #endif
#endif
    pmd_t *p_pmd;
    pte_t *p_pte;
    struct page *p_page = NULL;
    
    //CTDPF("kvirt_to_kva: kvirt = 0x%lx ", kvirt);
    p_pgd = pgd_offset_k(kvirt);
    if (!pgd_none(*p_pgd)) {
#if defined(__x86_64__)
    #ifdef pud_offset 
       p_pud = pud_offset(p_pgd, kvirt);
       p_pmd = pmd_offset(p_pud, kvirt);
    #else
       p_pmd = pmd_offset(p_pgd, kvirt);
    #endif
#else
        p_pmd = pmd_offset(p_pgd, kvirt);
#endif
        //p_pmd = pmd_offset(p_pgd, kvirt);
        if (!pmd_none(*p_pmd)) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
            p_pte = pte_offset_map(p_pmd, kvirt);
#else
            //p_pte = pte_offset_map(p_pmd, kvirt);
            p_pte = pte_offset_kernel(p_pmd, kvirt);
#endif
            if(pte_present(*p_pte))
                p_page = pte_page(*p_pte);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
            //pte_unmap(p_pte);
#endif
        }
    }

    return p_page;
}

#if !defined(page_to_pfn) && !defined(CONFIG_OUT_OF_LINE_PFN_TO_PAGE)
#   define page_to_pfn(page) ((unsigned long)((page)-mem_map))
#endif

unsigned long 
kvirt_to_phys(unsigned long kvirt)
{
    unsigned long offset;
    struct page *p_page;
    
    p_page = kvirt_to_page(kvirt);
    offset = (unsigned long)kvirt&~PAGE_MASK;

    return page_to_pfn(p_page)*PAGE_SIZE+offset;
}

unsigned long 
kvirt_to_bus(unsigned long kvirt)
{
    unsigned long offset;
    struct page *p_page;
    
    p_page = kvirt_to_page(kvirt);
    offset = (unsigned long)kvirt&~PAGE_MASK;

    return page_to_pfn(p_page)*PAGE_SIZE+offset;
}

HRESULT 
reserve_remap_pages(unsigned long ulUserAddr, stMemHandle *pMem, unsigned long ulSize, 
        pgprot_t prot)
{
    struct page *p_page;
    int nSize = ulSize;
    unsigned long phys_pos, kvirt_pos, uvirt_pos = ulUserAddr;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0) || LINUX_VERSION_CODE == KERNEL_VERSION(2, 4, 21)
    struct vm_area_struct *p_vma;
#endif

    if (!pMem)
        CTRETURN(E_INVALIDARG);
    if (ulSize > PAGE_ALIGN(pMem->dwPageRange))
        CTRETURN(E_FAIL);
    if (!(pMem->MemType == MEMTYPE_PAGE || pMem->MemType == MEMTYPE_VMALLOC))
        CTRETURN(E_FAIL);
    
    kvirt_pos = (unsigned long)pMem->pVirtAddr;
    pMem->bReserved = TRUE;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0) || LINUX_VERSION_CODE == KERNEL_VERSION(2, 4, 21)
    p_vma = find_vma(current->mm, uvirt_pos);
#endif

    while (nSize > 0) {
        if (pMem->MemType == MEMTYPE_VMALLOC) {
            phys_pos = kvirt_to_phys(kvirt_pos);
            p_page = kvirt_to_page(kvirt_pos);
        } else { // pMem->MemType == MEMTYPE_PAGE
            phys_pos = virt_to_phys((void *)kvirt_pos);
            p_page = virt_to_page(kvirt_pos);
        }
       
        SetPageReserved(p_page);

        down_write(&current->mm->mmap_sem);
#if LINUX_VERSION_CODE == KERNEL_VERSION(2, 4, 21)
        if (remap_page_range(p_vma, uvirt_pos, phys_pos, PAGE_SIZE, prot)) {
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
        if (remap_page_range(uvirt_pos, phys_pos, PAGE_SIZE, prot)) {
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 10)
        if (remap_pfn_range(p_vma, uvirt_pos, page_to_pfn(p_page), PAGE_SIZE, prot)) {
#else
        if (remap_page_range(p_vma, uvirt_pos, phys_pos, PAGE_SIZE, prot)) {
#endif
            up_write(&current->mm->mmap_sem);
            pMem->bReserved = FALSE;
            CTDPF("fetal error: remap_page_range failed...");
            CTRETURN(E_FAIL);
        }
        up_write(&current->mm->mmap_sem);

        // do sth for the loop
        uvirt_pos += PAGE_SIZE;
        kvirt_pos += PAGE_SIZE;
        nSize -= PAGE_SIZE;
    }
    

    return S_OK;
}
EXPORT_SYMBOL_NOVERS(reserve_remap_pages);

HRESULT 
unreserve_pages(stMemHandle *pMem, unsigned long ulSize)
{
    int nSize;
    struct page *p_page; 
    unsigned long kvirt_pos;

    if (!pMem)
        CTRETURN(E_INVALIDARG);
    if (ulSize > PAGE_ALIGN(pMem->dwPageRange))
        CTRETURN(E_FAIL);

    nSize = ulSize;
    kvirt_pos = (unsigned long)pMem->pVirtAddr;
    
    CTDPF("unreserve_pages(): addr=0x%lx, size=%ld ", (unsigned long)pMem->pVirtAddr, ulSize);
    if (pMem->bReserved) {
        // to unmap the allocated user mode address range. 
        if (pMem->bMapped && pMem->ulUserAddr && pMem->MemType!=MEMTYPE_UMALLOC) {
            if (current->pid == pMem->pid) {
                down_write(&current->mm->mmap_sem);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 21)
                do_munmap(current->mm, pMem->ulUserAddr, pMem->dwPageRange, 1);
#else
                do_munmap(current->mm, pMem->ulUserAddr, pMem->dwPageRange);
#endif
                up_write(&current->mm->mmap_sem);
            }
            pMem->bMapped = FALSE;
        }

        // to clear page-reserved flag.
        while (nSize > 0) {
            if (pMem->MemType == MEMTYPE_VMALLOC || pMem->MemType == MEMTYPE_UMALLOC) 
                p_page = kvirt_to_page(kvirt_pos);
            else // Mem->MemType == MEMTYPE_PAGE
                p_page = virt_to_page(kvirt_pos);

            ClearPageReserved(p_page);

            kvirt_pos += PAGE_SIZE;
            nSize -= PAGE_SIZE;
        }

        pMem->bReserved = FALSE;
    }

    return S_OK;
} 

OSPHYSMEMHDL 
reserve_r3_buffer(unsigned long ulUserAddr, unsigned long ulSize) 
{
    stMemHandle *phMem;
    struct page **pp_page_pool;
    unsigned int nPage, nSize = ulSize;
    unsigned long uvirt_pos = ulUserAddr;
    void *kvirt_pos;
    int res;
    
    CTDPF("reserve_r3_buffer called: ulSize=%lu\n", ulSize);

    if (!ulUserAddr) {
        CTDPF("Error: user mode address is NULL.\n");
        return NULL;
    }
    
    // to calculate how many pages are going to be used.
    nPage = (unsigned int)(nSize/PAGE_SIZE) + ((nSize%PAGE_SIZE) ? 1 : 0);
    if ((pp_page_pool = (struct page **)kmalloc(nPage*sizeof(struct page *), GFP_ATOMIC)) == NULL)
        return NULL;
    
    down_read(&current->mm->mmap_sem); 
    res = get_user_pages(current, current->mm, uvirt_pos, nPage, 1, 0, pp_page_pool, NULL);
    up_read(&current->mm->mmap_sem); 
    if (res < nPage) {
        kfree(pp_page_pool);
        CTDPF("Error: get_user_pages failed.\n");
        return NULL;
    }
    
    // vmap the pages in the pp_page_pool array.
    kvirt_pos = vmap(pp_page_pool, nPage, VM_ALLOC, PAGE_KERNEL);

    kfree(pp_page_pool);

    // allocate and fill in the phMem descriptor. 
    phMem = kmalloc(sizeof(stMemHandle), GFP_KERNEL);
    if (!phMem) {
        PDEBUG("Error: sysAllocPages could not allocate memory\n");
        return NULL;
    }
    memset(phMem, 0, sizeof(stMemHandle));
    phMem->ulOrder = bytes_to_order(ulSize); 
    phMem->pVirtAddr = kvirt_pos;
    if (!phMem->pVirtAddr) {
        CTASSERT(0);
        kfree(phMem);
        return NULL;
    }
    phMem->MemType = MEMTYPE_UMALLOC;

    phMem->dwPageRange = ulSize;
    phMem->pid = current->pid;
    phMem->ulUserAddr = ulUserAddr;
    phMem->bReserved = TRUE;
    phMem->bMapped = TRUE;
    
    return ((OSPHYSMEMHDL)phMem);
}
EXPORT_SYMBOL_NOVERS(reserve_r3_buffer);

unsigned long 
bytes_to_shift(CTDWORD dwBytes)
{
    unsigned long ulShift = 0;

    while ((1 << ulShift) < dwBytes)	
        ulShift++;

    return ulShift;
}
EXPORT_SYMBOL_NOVERS(bytes_to_shift);

unsigned long 
bytes_to_order(CTDWORD dwBytes)
{
    unsigned long ulShift;

    ulShift = bytes_to_shift(dwBytes);

    return ulShift >= PAGE_SHIFT ? ulShift - PAGE_SHIFT : 0;
}
EXPORT_SYMBOL_NOVERS(bytes_to_order);

/////////////////
// ioctl handler
/////////////////

int 
ioctl_dispatch(unsigned long ulArg, FPDIOCDISPATCH pfnDiocDispatch)
{
    CTDIOCPARAMETERS kdioc, udioc;
    HRESULT hr;
    CTDWORD dwBytesRet;
    int nRet = 0;

    CTASSERT(ulArg && pfnDiocDispatch);
    memset(&kdioc, 0, sizeof(CTDIOCPARAMETERS));
    
    if (copy_from_user(&udioc, (void *) ulArg, sizeof(CTDIOCPARAMETERS))) {
        PDEBUG("Copying dioc from userspace failed\n");
        nRet = -EFAULT;
        goto free_mem;
    }
   
    /*
    PDEBUG("DEVICEIOCONTROL called: cbInBuf=%d, cbOutBuf=%d\n", udioc.dioc_cbInBuf, 
            udioc.dioc_cbOutBuf);
    */
    
    kdioc.dioc_IOCtlCode = udioc.dioc_IOCtlCode;

    // Copy input buffer from user to kernel space
    kdioc.dioc_cbInBuf = udioc.dioc_cbInBuf;
    if (udioc.dioc_cbInBuf) {
        kdioc.dioc_InBuf = vmalloc(udioc.dioc_cbInBuf);
        if (!kdioc.dioc_InBuf) {
            nRet = -ENOMEM;
            goto free_mem;
        }
        if (copy_from_user(kdioc.dioc_InBuf, udioc.dioc_InBuf, udioc.dioc_cbInBuf)) {
            PDEBUG("Copying inbuf from userspace failed\n");
            nRet = -EFAULT;
            goto free_mem;
        }
    }

    // Allocate output buffer
    kdioc.dioc_cbOutBuf = udioc.dioc_cbOutBuf;
    if (udioc.dioc_cbOutBuf) {
        kdioc.dioc_OutBuf = vmalloc(udioc.dioc_cbOutBuf);
        if (!kdioc.dioc_OutBuf) {
            nRet = -ENOMEM;
            goto free_mem;
        }
        memset(kdioc.dioc_OutBuf, 0, udioc.dioc_cbOutBuf);
    }

    // Allocate bytes returned
    dwBytesRet = 0;
    kdioc.dioc_bytesret = &dwBytesRet;
	
    // Zero bytesret first
    if (copy_to_user(udioc.dioc_bytesret, kdioc.dioc_bytesret, sizeof(CTDWORD))) {
        PDEBUG("Copying bytesret to userspace failed\n");
        nRet = -EFAULT;
        goto free_mem;
    }

    // Make dispatch call
    hr = pfnDiocDispatch(&kdioc);
    
    // Copy output buffer from kernel to user space
    if (udioc.dioc_cbOutBuf && dwBytesRet) {
        if (!access_ok(VERIFY_WRITE, udioc.dioc_OutBuf, udioc.dioc_cbOutBuf) 
            || copy_to_user(udioc.dioc_OutBuf, kdioc.dioc_OutBuf, udioc.dioc_cbOutBuf)) {
            PDEBUG("Copying outbuf to userspace failed\n");
            nRet = -EFAULT;
            goto free_mem;
        }
    }

    // Copy bytes returned from kernel to user space
    if (dwBytesRet) {
        if (copy_to_user
            (udioc.dioc_bytesret, kdioc.dioc_bytesret, sizeof(CTDWORD))) {
            PDEBUG("Copying bytesret to userspace failed\n");
            nRet = -EFAULT;
            goto free_mem;
        }
    }

free_mem:
    if (kdioc.dioc_InBuf)
        vfree(kdioc.dioc_InBuf);
    if (kdioc.dioc_OutBuf)
        vfree(kdioc.dioc_OutBuf);

    return nRet;
}
EXPORT_SYMBOL_NOVERS(ioctl_dispatch);

///////////////
// Atomic stuff
///////////////

__attribute__ ((unused))
static spinlock_t g_lockCount = SPIN_LOCK_UNLOCKED;

CTLONG 
InterlockedIncrement(CTLONG *plRef) 
{
    CTLONG ret;
    unsigned long flags;

    //CTDPF("InterlockedIncrement called");
    CTASSERT(plRef);
    spin_lock_irqsave(&g_lockCount, flags);
    ret = ++(*plRef);
    spin_unlock_irqrestore(&g_lockCount, flags);

    return ret;
}
EXPORT_SYMBOL_NOVERS(InterlockedIncrement);

CTLONG 
InterlockedDecrement(CTLONG *plRef) 
{
    CTLONG ret;
    unsigned long flags;

    //CTDPF("InterlockedDecrement called");
    CTASSERT(plRef);
    spin_lock_irqsave(&g_lockCount, flags);
    ret = --(*plRef);
    spin_unlock_irqrestore(&g_lockCount, flags);

    return ret;
}
EXPORT_SYMBOL_NOVERS(InterlockedDecrement);

/////////////////////////
// devfs opertation stuff
/////////////////////////
/* Legacy
#define __NR_ct_sys_mknod __NR_mknod
_syscall3(long, ct_sys_mknod, const char *, filename, int, mode, dev_t, dev)
#define __NR_ct_sys_unlink __NR_unlink
_syscall1(long, ct_sys_unlink, const char *, pathname)
EXPORT_SYMBOL_NOVERS(ct_sys_mknod);
EXPORT_SYMBOL_NOVERS(ct_sys_unlink);
 */
 
///////////////////////
// Monochrome debugging
///////////////////////

/* FIXME : Not using at the moment
EXPORT_SYMBOL_NOVERS(monoOutValue);
EXPORT_SYMBOL_NOVERS(monoInKey);
EXPORT_SYMBOL_NOVERS(monoDebug);
EXPORT_SYMBOL_NOVERS(monoSetDisplay);
*/


void *
stack_alloc(size_t size)
{
    void *pv;

    if (size == 0)
        size = 1;

    pv = kmalloc(size, GFP_ATOMIC);

    if (pv)
        memset(pv, 0, size);

    CTASSERT(pv);

    return pv;
}
EXPORT_SYMBOL_NOVERS(stack_alloc);

void
stack_free(void *pv)
{
    if (pv == NULL) {
        CTASSERT(0);
        return;
    }

    kfree(pv);
}
EXPORT_SYMBOL_NOVERS(stack_free);

unsigned long stack_alloc_page(void)
{
    unsigned long l;
    l = __get_free_pages(GFP_ATOMIC, 1);
    PDEBUG("get_free_pages %lx,  %p", l, (void*)l);
    return l;
}
EXPORT_SYMBOL_NOVERS(stack_alloc_page);

void stack_free_page(void *pv)
{
    PDEBUG("free_pages %lx,  %p", (unsigned long)pv, pv);
    free_pages((unsigned long)pv, 1);
}
EXPORT_SYMBOL_NOVERS(stack_free_page);

void myDelay(unsigned long n)
{
	mdelay(n);
}
EXPORT_SYMBOL_NOVERS(myDelay);

int my_is_atomic(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    return 0;
#else
    return (in_atomic());
#endif
}
EXPORT_SYMBOL_NOVERS(my_is_atomic);


