/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
Main code module for ctsfman
Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/

#define EXPORT_SYMTAB
#include <linux/version.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/sound.h>

#include "osutils.h"
//#include "isfmscore.h"
#include "sfmain.h"

int Sfms_InitDriver(void);
int Sfms_ExitDriver(void);

////////////////////////
// SFMan file operations
////////////////////////


int
emu10kx_sfman_ioctl(struct inode *pInode, struct file *pFile, unsigned int uCmd, unsigned long ulArg)
{
    return ioctl_dispatch(ulArg, SfmsCore_Dispatch);
}

int
emu10kx_sfman_open(struct inode *pInode, struct file *pFile)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    CT_MOD_INC_USE_COUNT;
#endif
    return 0;
}

int
emu10kx_sfman_release(struct inode *pInode, struct file *pFile)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    CT_MOD_DEC_USE_COUNT;
#endif
    return 0;
}

struct file_operations emu10kx_sfman_fops = {
    ioctl:   emu10kx_sfman_ioctl,
    open:    emu10kx_sfman_open,
    release: emu10kx_sfman_release,
};


////////////////////////
// Entry and exit points
////////////////////////

// Global vars
INSERT_VERSION_INFO;

static int disable = 0;
static int g_nDevSFMan = -1;

int
ctsfman_init(void)
{
    CTDPF("ctsfman_init");

    if (disable)
        return 0;

    __do_global_ctors_aux();

    if (Sfms_InitDriver()) {
        CTDPF("init driver failed\n");
        goto err;
    }
    g_nDevSFMan = register_sound_special(&emu10kx_sfman_fops, SFMAN_MINOR);
    if (g_nDevSFMan < 0) {
        CTDPF("cannot register device\n");
        goto err;
    }
    
    return 0;

err:
    Sfms_ExitDriver();
    return -ENODEV;
}

void
ctsfman_exit(void)
{
    CTDPF("ctsfman_exit");

    if (disable)
        return;

    if (g_nDevSFMan >= 0)
        unregister_sound_special(g_nDevSFMan);

    Sfms_ExitDriver();
}


module_init(ctsfman_init);
module_exit(ctsfman_exit);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
MODULE_PARM(disable, "i");
#else
module_param(disable, int, 0);
#endif
MODULE_PARM_DESC(disable, "Sets module to disabled state");

MODULE_AUTHOR("Copyright (C) 1999-2008 Creative Technology Ltd");
MODULE_DESCRIPTION("SFMAN driver version 1.18.0000");
MODULE_LICENSE("Creative Proprietary License");

