/**
*******************************************************************************
Confidential & Proprietary
Private & Confidential
Creative Confidential
*******************************************************************************
*/
/**
*******************************************************************************
This application is to convert an ALSA version ID string to an numeric for ALSA version
comparison on preprocessing stage of building process.

Copyright (C) 2008, Creative Technology Ltd. All Rights Reserved.
*******************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define FILE_NAME "sound/version.h"
#define ALSA_VERSION_FIELDS 3

int main(void) 
{
	unsigned long code;
	size_t size;
	unsigned char alsainc;
	char buf[512];
	char *str;
	char *token;
	FILE *f_ver;
	int i;

	/* Get the path to sound/version.h which contains ALSA version ID string
	 * from environment variable. */
	alsainc = 0;
	size = sizeof(buf) - strlen(FILE_NAME) - 2;
	if ((str = getenv("ALSAINC"))) {
		alsainc = 1;
	} else if ((str = getenv("KSRC"))) {
		size -= strlen("include/");
	} else {
		fprintf(stderr, "Can't find the path to %s\n", FILE_NAME);
		exit(1);
	}

	/* Extract the ALSA version string from sound/version.h */
	/* First get the path to the file to open. */
	strncpy(buf, str, size);
	if (buf[strlen(buf) - 1] != '/') {
		strcat(buf, "/");
	}
	if (!alsainc) {
		strcat(buf, "include/");
	}
	strcat(buf, FILE_NAME);
	
	/* Open the file and extract version ID string. */
	if (!(f_ver = fopen(buf, "r"))) {
		fprintf(stderr, "Can't open file %s\n", buf);
		exit(1);
	} 

	token = NULL;
	while (fgets(buf, sizeof(buf), f_ver)) {
		str = buf + strspn(buf, " \t\n\r");
		if ('/' == *str) {
			continue; /* Ignore comment lines. */
		}

		while ((token = strsep(&str, " \t"))) {
			if (!strcmp(token, "CONFIG_SND_VERSION")) {
				str += strspn(str, " \t\"");
				token = strsep(&str, "\"");
				break;
			}
		}
		
		if (token) {
			break;
		}
	}

	fclose(f_ver);

	if (!token) {
		fprintf(stderr, "Can't find verion ID.\n");
		exit(1);
	}
	
	/* Convert the version ID string to a numeric. */
	code = 0;
	i = 0;
	while ((*token != '\0') && (i < ALSA_VERSION_FIELDS)) {
		code <<= 8;
		code |= strtoul(token, &token, 10);
		i++;
		if (*token != '\0') {
			token++;
		}
	}

	fprintf(stdout, "%lu\n", code);

	return 0;
}

